package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import java.awt.*; 
import javax.swing.*;
import java.awt.image.*;
import java.util.*;
import java.lang.*;
import java.io.*;

class ArchiveMarks extends JFrame{
	RandomAccessFile inout=null; // global variable so i don't have to pass it as an argument 
	String mostRecentLineRead="";			
 	String filename=null;
 
 	public void readWriteProjectionsPanelData(boolean ifWrite, Stereo3DReconstruction projections) {
		int i;
	  	if(ifWrite) { // write data to disk file...
	  		if((inout = openWritingFile("name of Stereo3DReconstruction Panel Data file to write?"))==null) {
	  			talk("failed to open file. try again");
	  			return;
	  			}
				projections.archiveParameters(this,ifWrite);
				closeFile();
				projections.progressReport("Stereo3DReconstruction Panel Data written to disk file successfully");
			}
	  	else { //read data from disk file...
	  		if((inout = openReadingFile("name of file to read colors from?"))==null) {
	  			talk("failed to open file. try again");
	  			return;
	  			}
	  		    double badval = -7777.777;
				projections.archiveParameters(this,ifWrite);
				projections.progressReport("Stereo3DReconstruction Panel Data recovered from disk file successfully");
				closeFile();
			}
 		}
 		
    public void writeCentrosomeList(Centrosomes centsJPanel, int [] Xshifts,
    		int [] Yshifts, int [] Zshifts) { 
		int i;
		String stg;
		if((inout = openWritingFile("name of file to write?"))==null) {
	  			talk("failed to open file. try again");
	  			return;
	  			}
		writeAString("  Put this in red channel boost file");
		writeAString("1.0    1.0   intensity boosts at first and last section");
		writeNum(2*Zshifts[0], "  = Zshift (red channel) in half sections");
		writeNum(-Xshifts[0], "  = xshift (red channel)");
		writeNum(Yshifts[0], "  = yshift (red channel)");
		writeAString("  ");
		
		writeAString("  Put this in green channel boost file");
		writeAString("1.0    1.0   intensity boosts at first and last section");
		writeNum(2*Zshifts[1], "  = Zshift (green channel) in half sections");
		writeNum(-Xshifts[1], "  = xshift (green channel)");
		writeNum(Yshifts[1], "  = yshift (green channel)");
		writeAString("  ");
		
		writeAString("  Put this in blue channel boost file");
		writeAString("1.0    1.0   intensity boosts at first and last section");
		writeNum(2*Zshifts[2], "  = Zshift (blue channel) in half sections");
		writeNum(-Xshifts[2], "  = xshift (blue channel)");
		writeNum(Yshifts[2], "  = yshift (blue channel)");
		writeAString("  ");
		
		writeAString("  The centrosome list follows,  where the Z axis ");
		writeAString("  numbers are actual section numbers... ");
		centsJPanel.writeCentrosomeListToDisk(this);
		closeFile();
		talk("Centrosome list and shifts written to disk file successfully");
	}

    public void writeHistogram(Histogram Hst) {
	  	// write data to disk file...
	  	if((inout = openWritingFile("name of Histogram file to write?"))==null) {
	  			talk("failed to open file. try again");
	  			return;
	  			}
    	Hst.writeToDisk(this);
 		closeFile();
		talk("Histogram written to disk file successfully");
   	}

    public void writeZCore(Histogram Hst) {
	  	// write data to disk file...
	  	if((inout = openWritingFile("name of z-core data file to write?"))==null) {
	  			talk("failed to open file. try again");
	  			return;
	  			}
    	Hst.writeZCoreToDisk(this);
 		closeFile();
		talk("z-core data written to disk file successfully");
   	}

    public void writePointSpreadData(writePointSpreadFile wpsf,int iwh) {
	  	// write data to disk file...
	  	if(iwh==0 && 
	  	    (inout = openWritingFile("Create text file to write point spread data to"))==null) {
	  			talk("failed to open file. try again");
	  			return;
	  			}
    	wpsf.writeToDisk(this,iwh);
 		if(iwh==0) {
 			closeFile();
			talk("Point spread data written to disk file successfully");
			} 
   	}

    public boolean readPointSpreadProfiles(writePointSpreadFile psfw) {
	  	// read point spread profiles from a disk file...
	  	if((inout = openReadingFile("name of file to read point spread profiles from"))==null) {
	  		talk("failed to open file. try again");
	  		return false;
	  		}
    	psfw.readFromDisk(this);
 		closeFile();
		talk("Point spread profiles successfully read from disk file ");
	  	return true;
   		}

    public void readHistogram(Histogram Hst) {
	  	// read old histogram specifications from a disk file...
	  	if((inout = openReadingFile("name of Histogram file to write?"))==null) {
	  		talk("failed to open file. try again");
	  		Hst.ifDiskReadWorked =  false;
	  		}
    	Hst.readFromDisk(this);
 		closeFile();
		talk("Histogram read from disk file successfully");
	  	Hst.ifDiskReadWorked =  true;
   		}

   public void readWriteColors(boolean ifWrite, PickColors colorPicker) { 
		int i;
	  	if(ifWrite) { // write data to disk file...
	  		if((inout = openWritingFile("name of Colors file to write?"))==null) {
	  			talk("failed to open file. try again");
	  			return;
	  			}
				colorPicker.archiveColors(this,ifWrite);
				closeFile();
				talk("colors data written to disk file successfully");
			}
	  	else { //read colors data from disk file...
	  		if((inout = openReadingFile("name of file to read colors from?"))==null) {
	  			talk("failed to open file. try again");
	  			return;
	  			}
	  		    double badval = -7777.777;
				colorPicker.archiveColors(this,ifWrite);
				talk("colors data recovered from disk file successfully");
				closeFile();
			}
	}

   public void readWriteSegmenterFile(boolean ifWrite, Segmenter seg) { 
		int i;
	  	if(ifWrite) { // write segmentation data to disk file...
	  		if((inout = openWritingFile("name of segmentation file to write?"))==null) {
	  			talk("failed to open file. try again");
	  			return;
	  			}
				seg.archiveSegmentationData(this,ifWrite);
				closeFile();
				talk("segmentation data written to disk file successfully");
			}
	  	else { //read segmentation data from disk file...
	  		if((inout = openReadingFile("name of file to read segmentation data from?"))==null) {
	  			talk("failed to open file. try again");
	  			return;
	  			}
	  		    double badval = -7777.777;
				seg.archiveSegmentationData(this,ifWrite);
				talk("segmentation data recovered from disk file successfully");
				closeFile();
			}
	}

   public void readWriteMarks(boolean ifWrite,
    		RGBMarker [] marks, PickColors colorPicker, 
				Centrosomes centsJPanel, BlackHoles holesJPanel, GoToHotSpots hotSpotJPanel,
				 int [] Xshifts, int [] Yshifts, int [] Zshifts) { 
		int i;
	  	if(ifWrite) { // write data to disk file...
	  		if((inout = openWritingFile("name of file to write?"))==null) {
	  			talk("failed to open file. try again");
	  			return;
	  			}
				for(i=0;i<3;i++) writeNum(Xshifts[i],(" = Xshifts["+i+"]"));
				for(i=0;i<3;i++) writeNum(Yshifts[i],(" = Yshifts["+i+"]"));
				for(i=0;i<3;i++) writeNum(Zshifts[i],(" = Zshifts["+i+"]"));
				colorPicker.archiveColors(this,ifWrite);
				for(i=1;i<11;i++)
					marks[i].archiveMarks(this,ifWrite,i);
				hotSpotJPanel.archiveHotSpots(this,ifWrite);
				centsJPanel.archiveCents(this,ifWrite);
				writeNum(-7777," = flag that ends centrosomes list");
				holesJPanel.archiveHoles(this,ifWrite);
				writeNum(-7777," = flag that ends holes list");
				closeFile();
				talk("marks, shifts, colors data written to disk file successfully");
			}
	  	else { //read data from disk file...
	  		if((inout = openReadingFile("name of file to read from?"))==null) {
	  			talk("failed to open file. try again");
	  			return;
	  			}
	  		     double badval = -7777.777;
				for(i=0;i<3;i++) Xshifts[i] = (int)getNextFileDouble(badval);
				for(i=0;i<3;i++) Yshifts[i] = (int)getNextFileDouble(badval);
				for(i=0;i<3;i++) Zshifts[i] = (int)getNextFileDouble(badval);
				colorPicker.archiveColors(this,ifWrite);
				for(i=1;i<11;i++)
					marks[i].archiveMarks(this,ifWrite,i);
				hotSpotJPanel.archiveHotSpots(this,ifWrite);
				centsJPanel.archiveCents(this,ifWrite); 
				holesJPanel.archiveHoles(this,ifWrite);				
				talk("marks, shifts, colors data recovered from disk file successfully");
				closeFile();
			}
	}

	public String readOneLine() {
		byte [] buf = new byte[129];
		byte [] inbyte = {0};
		int i=0,igot=0;
		while(i<128) {
			try {igot=inout.read(inbyte);}
			catch(IOException iex) {
				talk("IOException was: "+iex);
				return null;
				}
			if(igot == 1 && inbyte[0] != 10)
				buf[i++]=inbyte[0];
			else {
				if(i>0) 
					return (new String(buf,(byte)0,0,i));
				else return null;
				}
			}
		return ("overflow line");
		}
		
	public void read16Ints(int [] LUT, int k)  {
		byte [] intbuf = new byte[256];
		byte [] inbyte = {0};
		Double D;
		int ngot=0;
		int i=0,in=0, igot;
		for(i=0;i<16;i++) LUT[k+i]=k+i;  // in case we dont read proper data...
		i=0;
		while(i<255) {
		    // search for first non-white-space
		    try {igot=inout.read(inbyte);}
		    catch(IOException iex) {
			    talk("IOException was: "+iex);
			    return;
			    }
		    if(igot != 1) {
			talk("read16ints igot = "+igot+"when ngot="+ngot);
			return;
			}
		    if(inbyte[0]==10) {
			talk("read16ints hit 10=EOL when ngot="+ngot);
			return;
			}
		    if(in==0) { //still searching for first non-blank;
			if(inbyte[0] != ' ' && inbyte[0] != '\t')
			    intbuf[in++]=(byte)inbyte[0];
			++i;
			} 
		    else { 
		      if(inbyte[0] == ' ' || inbyte[0]=='\t') { // just came to end of a number
			intbuf[in]=(byte)' ';
			String stg = new String(intbuf,(byte)0,0,in);
			try{D = new Double(stg);}
			catch (NumberFormatException NEF) {
			    talk("bad conversion of one of 16 ints??? "+NEF);
			    readToEOL();
			    return;
			    }
			if((in=(int)(D.doubleValue())) >255) in=255;
			if(in<0)in=0;
			LUT[k+ngot] = in;
			if(++ngot>=16) {
			    readToEOL();
			    return;
			    }
			in =0; // restart search for next int.
			}
		      else {
			intbuf[in++] = inbyte[0];
			++i;
			}
		      }
		   }
	    readToEOL();
	    return;
	   }
			
	public void read12Ints(int [] LUT)  {
		byte [] intbuf = new byte[256];
		byte [] inbyte = {0};
		Double D;
		int ngot=0;
		int k=0; // this is copied from read16ints() and we just always
		// use the offset k=0
		int i=0,in=0, igot;
		for(i=0;i<12;i++) LUT[k+i]=-1;  // in case we dont read proper data...
		i=0;
		while(i<255) {
		    // search for first non-white-space
		    try {igot=inout.read(inbyte);}
		    catch(IOException iex) {
			    talk("IOException was: "+iex);
			    return;
			    }
		    if(igot != 1) {
			talk("read12ints igot = "+igot+"when ngot="+ngot);
			return;
			}
		    if(inbyte[0]==10) {
			talk("read12ints hit 10=EOL when ngot="+ngot);
			return;
			}
		    if(in==0) { //still searching for first non-blank;
			if(inbyte[0] != ' ' && inbyte[0] != '\t')
			    intbuf[in++]=(byte)inbyte[0];
			++i;
			} 
		    else { 
		      if(inbyte[0] == ' ' || inbyte[0]=='\t') { // just came to end of a number
			intbuf[in]=(byte)' ';
			String stg = new String(intbuf,(byte)0,0,in);
			try{D = new Double(stg);}
			catch (NumberFormatException NEF) {
			    talk("bad conversion of one of 12 ints??? "+NEF);
			    readToEOL();
			    return;
			    }
			in=(int)(D.doubleValue());
			LUT[k+ngot] = in;
			if(++ngot>=12) {
			    readToEOL();
			    return;
			    }
			in =0; // restart search for next int.
			}
		      else {
			intbuf[in++] = inbyte[0];
			++i;
			}
		      }
		   }
	    readToEOL();
	    return;
	   }
			
	private void readToEOL() {
	    int igot;
	    byte [] inbyte={0};
	    while(inbyte[0] != 10) {
		try {igot=inout.read(inbyte);}
		catch(IOException iex) {
		    talk("IOException was: "+iex);
		    return;
		    }
		}
	    return;
	    }
	    		
	public double getNextFileDouble(double badval) {
		return getNextFileDouble(badval,false);
		}
		
	public double getNextFileDouble(double badval, boolean ifSaveEntireLine) {
		byte [] buf = new byte[129];
		byte [] inbyte = {0};
		Double D;
		int i=0,igot=0,firstwhitespaceat=-1;
		while(i<128) {
			try {igot=inout.read(inbyte);}  // we probably start at the byte just after the number on the previous line
			catch(IOException iex) {
				talk("IOException was: "+iex);
				return badval;
				}
			if(igot == 1 && inbyte[0] != 10) { // we have not yet found the 10 that ends the previous line...
				if(firstwhitespaceat<0 && (inbyte[0]==' ' || inbyte[0]=='\t'))
					 firstwhitespaceat=i; // hit first white space
				buf[i++]=inbyte[0]; // packs entire line into the buf array.
				} 
			else { // land here when the first 10 is hit.
				if(i>0) {
					if(firstwhitespaceat<0) firstwhitespaceat=i;
					String stg = new String(buf,(byte)0,0,firstwhitespaceat);
					if(ifSaveEntireLine) mostRecentLineRead =  new String(buf,(byte)0,0,i);
					if(stg.length() >=1) {
						try{D = new Double(stg);}
						catch (NumberFormatException NEF) {
							talk("bad conversion of string <"+
								new String(buf,(byte)0,0,i)+"> to Double: "+NEF);
							return badval;
							}
						return (D.doubleValue());
						}
					else return badval;						
					}
				else return badval;
				}
			}
		talk("TooLongLine encountered = <"+new String(buf,(byte)0,0,i)+">");
		return (badval);
		}
		
	public void writeNum(int n, String stg) {
		String outline=""+n+"    "+stg;
		writeAString(outline);
		}	
		
	public void writeNum(boolean b, String stg) {
		String outline;
		if(b) outline="1  "+stg;
		else outline = "0  "+stg;
		writeAString(outline);
		}	
		
	public void writeNum(double n, String stg) {
		String outline=""+n+"    "+stg;
		writeAString(outline);
		}	
	
	public void write16Ints(int [] LUT, int k,  String lineStg) {
	    String stg = ""+LUT[k]+" "+LUT[k+1]+" "+LUT[k+2]+" "+LUT[k+3]+" "+
	    LUT[k+4]+" "+LUT[k+5]+" "+LUT[k+6]+" "+LUT[k+7]+" "+
	    LUT[k+8]+" "+LUT[k+9]+" "+LUT[k+10]+" "+LUT[k+11]+" "+
	    LUT[k+12]+" "+LUT[k+13]+" "+LUT[k+14]+" "+LUT[k+15]+" "+lineStg;
	    writeAString(stg);
	    }

	public void write12Ints(int [] LUT, String lineStg) {
	    String stg = ""+LUT[0]+" "+LUT[1]+" "+LUT[2]+" "+LUT[3]+" "+
	    LUT[4]+" "+LUT[5]+" "+LUT[6]+" "+LUT[7]+" "+
	    LUT[8]+" "+LUT[9]+" "+LUT[10]+" "+LUT[11]+" "+lineStg;
	    writeAString(stg);
	    }


	public void writeAString(String stg) {
		byte [] buf = new byte[stg.length()+1];
		stg.getBytes(0,stg.length(),buf,0);
		buf[stg.length()] = 10;
			try {
			inout.write(buf);
			}
		catch (IOException ioex) {
			talk("writeAString to file failed: "+ioex);
			}
		}
		
	public RandomAccessFile openWritingFile(String prompt) {
		RandomAccessFile outfile;
		FileDialog getFileToWrite = 
		   new FileDialog(this,prompt, FileDialog.SAVE);
		getFileToWrite.show();
		filename = getFileToWrite.getDirectory()+"/"+getFileToWrite.getFile();
		if(filename == null) return null;
		try{
			outfile = new RandomAccessFile(filename,"rw");
			}
		catch (IOException e) {
			talk("openWritingFile("+prompt+") failed");
			return null;
			}
		try{outfile.seek((long)0);}
		catch(IOException e) {};
		return outfile;
		}
	
	public void closeFile() {
		if(inout !=null)
			closeFile(inout);
		inout=null;
		}
		
	public void closeFile(RandomAccessFile file) {
		try{ 
		   	file.close();
			}
		catch (IOException exc){
		   	talk("Trouble closing file:" + exc);
			}
		}	
		
	public RandomAccessFile openReadingFile(String prompt) {
		RandomAccessFile infile;
		FileDialog getFileToWrite = 
		   new FileDialog(this,prompt, FileDialog.LOAD);
		getFileToWrite.show();
		filename = getFileToWrite.getDirectory()+"/"+getFileToWrite.getFile();
		if(filename == null) return null;
		try{
			infile = new RandomAccessFile(filename,"r");
			}
		catch (FileNotFoundException e) {
			talk("openReadingFile("+prompt+") failed -- file not found");
			return null;
			}
		catch (IOException e) {
			talk("openReadingFile("+prompt+") failed IOexception" + e);
			return null;
			}
		try{infile.seek((long)0);}
		catch(IOException e) {};
		return infile;
		}
		
public RandomAccessFile openReadWriteFile(String prompt) {
		RandomAccessFile infile;
		FileDialog getFileToWrite = 
		   new FileDialog(this,prompt, FileDialog.LOAD);
		getFileToWrite.show();
		filename = getFileToWrite.getDirectory()+"/"+getFileToWrite.getFile();
		if(filename == null) return null;
		try{
			infile = new RandomAccessFile(filename,"rw");
			}
		catch (FileNotFoundException e) {
			talk("openReadingFile("+prompt+") failed -- file not found");
			return null;
			}
		catch (IOException e) {
			talk("openReadingFile("+prompt+") failed IOexception" + e);
			return null;
			}
		try{infile.seek((long)0);}
		catch(IOException e) {};
		return infile;
		}
		
	public void changeJPanelSize(double ratio) {
		int pwidth = getSize().width;
		int pheight = getSize().height;
		pwidth = (int)(ratio*pwidth +.4999);
		pheight = (int)(ratio*pheight +.4999);
		setSize(pwidth,pheight);
		setVisible(true);
		}

	private byte [] booleanArrayToByteArray(boolean [] bol)	{
		byte [] bitMask = new byte [] {1,2,4,8,16,32,64,(byte)128};
		int nbools = bol.length;
		int nby = 1 + nbools/8;
		byte [] by = new byte[nby];		
		int i,j=0,k;
		for(i=0;i<nby;i++)
			by[i]=0;
			
		for(i=0;i<nbools;i++) 
			if(bol[i]) {
				k=i/8;
				j=i - 8*k;
				by[k] = (byte)(by[k] | bitMask[j]);
				}
		return by;
		}
		
	private boolean [] byteArrayToBooleanArray(byte [] by, int nbools)	{
		byte [] bitMask = new byte [] {1,2,4,8,16,32,64,(byte)128};
		boolean [] bol = new boolean[nbools];		
		int i,j=0,k;
			
		for(i=0;i<nbools;i++) {
			k=i/8;
			j=i - 8*k;
			bol[i] = ((byte)(by[k] & bitMask[j])) == 0? false: true;
			}
		return bol;
		}
		
	public void rewindInoutFile() {
		try{inout.seek((long)0);}
		catch(IOException e) {
			talk("rewind exception = "+e);
			};	
		}	
		
	public boolean writeBitMask(BitMask theBitMask) {
		   if(inout == null && (inout = openWritingFile("name of bit mask file to write"))==null) {
	  			talk("failed to open file. try again");
	  			return false;
	  			}
	  	// write  the mask file with all the header information.
	  	rewindInoutFile();		
		writeNum(theBitMask.NCOLS,"  = NCOLS");
		writeNum(theBitMask.NROWS,"  = NROWS");
		writeNum(theBitMask.NSECTS,"  = NSECTS");
		byte [] xytodo = booleanArrayToByteArray(theBitMask.toDoXYsects);
		byte [] xztodo = booleanArrayToByteArray(theBitMask.toDoXZsects);
		byte [] zytodo = booleanArrayToByteArray(theBitMask.toDoZYsects);
		writeNum(xytodo.length,"  = number of bytes in xytodo[] array");		
		writeNum(xztodo.length,"  = number of bytes in xztodo[] array");		
		writeNum(zytodo.length,"  = number of bytes in zytodo[] array");		
		writeNum(theBitMask.nbytes,"  = number of bytes in bytes[] array");		
		// write the byte array...
		try {
			inout.write(xytodo);
			inout.write(xztodo);
			inout.write(zytodo);
			inout.write(theBitMask.bytes);
			}
		catch (IOException ioex) {
			talk("writing byte arrays to file failed: "+ioex);
			return false;
			}
		xytodo=null;
		xztodo=null;
		zytodo=null;
		return true;
		}
	
	public boolean readBitMask(BitMask theBitMask) {
	  	if((inout = openReadWriteFile("name of BitMask file to read?"))==null) {
	  		talk("failed to open file. try again");
	  		return false;
	  		}
 		double badval = -7777.777;
 		int nbxys,nbxzs,nbzys;
 		byte [] bys;
		theBitMask.NCOLS = (int)getNextFileDouble(badval);		
		theBitMask.NROWS = (int)getNextFileDouble(badval);		
		theBitMask.NSECTS = (int)getNextFileDouble(badval);		
		
		nbxys = (int)getNextFileDouble(badval); // number of bytes in todoxy byte array
		nbxzs = (int)getNextFileDouble(badval); // number of bytes in todoxz byte array
		nbzys = (int)getNextFileDouble(badval); // number of bytes in todozy byte array
		theBitMask.nbytes = (int)getNextFileDouble(badval);				

		bys = new byte [nbxys]; // read the todoxy byte array
		try {inout.read(bys);}
		catch(IOException iex) {
			talk("IOException reading todoxy array was: "+iex);
			closeFile();
			return false;
			}
		theBitMask.toDoXYsects = byteArrayToBooleanArray(bys, theBitMask.NSECTS);
		
		bys = new byte [nbxzs]; // read the todoxz byte array
		try {inout.read(bys);}
		catch(IOException iex) {
			talk("IOException reading todoxz array was: "+iex);
			closeFile();
			return false;
			}
		theBitMask.toDoXZsects = byteArrayToBooleanArray(bys, theBitMask.NROWS);
		
		bys = new byte [nbzys]; // read the todozy byte array
		try {inout.read(bys);}
		catch(IOException iex) {
			talk("IOException reading todozy array was: "+iex);
			closeFile();
			return false;
			}
		theBitMask.toDoZYsects = byteArrayToBooleanArray(bys, theBitMask.NCOLS);
		
		theBitMask.bytes = new byte[theBitMask.nbytes];
		try {inout.read(theBitMask.bytes);}
		catch(IOException iex) {
			talk("IOException reading bytes[] was: "+iex);
			closeFile();
			return false;
			}

		talk("BitMask with NCOLS = "+theBitMask.NCOLS+", NROWS = "+theBitMask.NROWS+
			", NSECTS = "+theBitMask.NSECTS);
		talk("occupying "+theBitMask.nbytes+" bytes read from disk");
		theBitMask.tellHowManyRemainToDo(1);
		theBitMask.tellHowManyRemainToDo(2);
		theBitMask.tellHowManyRemainToDo(3);
		talk("The mask file remains open for future writing");
		rewindInoutFile();
		return true;
		}

	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}
