package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */




/**
 * Ball3D.java
 *
 * @author Created by G Odell june 2006
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  
import javax.swing.*;

class Ball3D { 
int ix, iy, iz;// coordinates of center of ball
int radius, maxrad;
double rimErrorFrac, xctr, yctr, rad1;
boolean iftalk=false;

public double disttocenter(int x, int y) {
double dist;
	x = x-ix;
	y = y-iy;
    return(Math.sqrt(x*x + y*y));		  
	}

public double disttoball(int x, int y) {
double dist;
	x = x-ix;
	y = y-iy;
	if((dist = disttocenter(x, y) - radius) < 0)
		return 0;
    else
    	return(dist);		  
	}

public double [] catarrays(double [] aray, double [] bray) {
	if(aray==null || bray==null)
		return null;
	double [] retaray = new double[aray.length + bray.length];
	int i;
	for(i=0;i<aray.length;i++)
		retaray[i]=aray[i];
	for(i=0;i<bray.length;i++)
		retaray[i+aray.length]=bray[i];
	return retaray;
	}
	
/* use this like a constructor. Send it a SwarmOfPoints. If it returns
false then that SwarmOfPoints does not fit in a ball of radius maxrad. 
If it returns true, then that SwarmOfPoints does fit in such a ball
and this Ball3D will have the best fit coefficients stored.
*/

public double [] ballQ(SwarmOfPoints swarm, int maxrad, double goodEnoughScore, boolean ifCirclesOnly) {
	this.maxrad=maxrad;
	// use EllipsoiddConvolutionKernel() to do this.
	int i;
	int maxextent=swarm.getLocations(); // returns the diameter of the swarm
	if(maxextent > 2*maxrad)
		return null;
	int npts= swarm.numPoints;
	SwarmPoint apoint;
	double cgx=swarm.locations[0][2]; // x coordinate of center of mass
	double cgy=swarm.locations[1][2]; // y coordinate of center of mass
	// nptsFullBallWhouldWontain = (int)(.49 + Math.PI*rad1*rad1);
	double minRadIfSphere = Math.sqrt(npts/Math.PI);
	double explorerad,r1 = ifCirclesOnly ? minRadIfSphere - 2 : minRadIfSphere/2;
	if(r1 < 2)
		r1=2;
	
	EllipsoidConvolutionKernel trialKernel;
	double [] aConvolvedResult=null,  kpars=null;; 
	double  [] returnArray=null;
	if(ifCirclesOnly) {
		while(r1 < minRadIfSphere+2 && (returnArray==null || returnArray[0] < goodEnoughScore)) {
			if((explorerad = r1/2) < 2)
				explorerad=2;		
			trialKernel = EllipsoidConvolutionKernel.getEllipsoidalKernel((int) r1, (int)r1);
			aConvolvedResult = trialKernel.fullSpinAndExploreCenterConvolution((int)(cgx+.49), (int)(cgy+.49), explorerad); 
			if(returnArray==null || aConvolvedResult[0] > returnArray[0]) 
				returnArray = catarrays(aConvolvedResult,trialKernel.kpars);
			++r1;
			}
		}
	else {
		double r2; // use ellipsoids 
		double r2lim;
		while(r1 < minRadIfSphere+2 && (returnArray==null || returnArray[0] < goodEnoughScore)) {
		// start with small balls, then increase radius. Quit when score is good enough
		// approx area = pi*r1*r2,
			r2lim = 2+ (npts+1)/(Math.PI*r1);
			for(r2=r1; r2<r2lim;r2++) {
				if((explorerad = r1/2) < 2)
					explorerad=2;		
				if(r2/2 > explorerad)
					explorerad=r2/2;		
				trialKernel = EllipsoidConvolutionKernel.getEllipsoidalKernel((int) r1, (int) r2);
				aConvolvedResult = trialKernel.fullSpinAndExploreCenterConvolution((int)(cgx+.49), (int)(cgy+.49), explorerad); 
			if(returnArray==null || aConvolvedResult[0] > returnArray[0]) 
				returnArray = catarrays(aConvolvedResult,trialKernel.kpars);
				}
				++r1;
			}
		}
	// on return, bestConvolutionResult = {score, xcen,ycen,angle}
	// decide if the score is good enough.bestConvolutionResult[0] is score/(# interior points)
	if(returnArray != null && returnArray[0] >= goodEnoughScore) { 
		talk("Ball3D.ballQ(goodEnoughScore="+goodEnoughScore+") returns with bestConvolutionResult: score="+returnArray[0]+", xcen="+returnArray[1]+
			", ycen="+returnArray[2]+", angle="+returnArray[3]);
		return returnArray;
		}
		
	else {
		if(returnArray != null)
			talk("----------failed bestConvolutionResult = "+returnArray[0]+" < "+goodEnoughScore);
		return null;
		}
	}
	
	
	private int inbounds(int xy,int NRC) {
		int i=xy;
		if(i<0) i=0;
		if(i>=NRC) i=NRC-1;
		return i;
		}
		
public int paintOnOutRas(short [] outras,int NCOLS,int NROWS, double [] arrayDefiningEllipsoid)  {
	// arrayDefiningEllipsoid={score,xcen,ycen,angle,rad1,rad2,...}
	talk("paintOnOutRas() : score="+arrayDefiningEllipsoid[0]+", xcen="+arrayDefiningEllipsoid[1]+", ycen="+arrayDefiningEllipsoid[2]+
		", angle="+arrayDefiningEllipsoid[3]+", r1="+arrayDefiningEllipsoid[4]+", r2="+arrayDefiningEllipsoid[5]);
	EllipsoidConvolutionKernel theKernel = EllipsoidConvolutionKernel.getEllipsoidalKernel((int)arrayDefiningEllipsoid[4], (int)arrayDefiningEllipsoid[5]);
	theKernel.rotateKernel(arrayDefiningEllipsoid[3]);
	int ipt,xpt,ypt,non = 0;
	for(ipt=0;ipt < theKernel.nInteriorPoints; ipt++) {// paint the interior white
		xpt = (int)(arrayDefiningEllipsoid[1] + theKernel.rotatedInteriorXYoffsets[ipt][0]);
		if(xpt>=0 && xpt < NCOLS) {
			ypt = (int)(arrayDefiningEllipsoid[2] + theKernel.rotatedInteriorXYoffsets[ipt][1]);
			if(ypt>=0 && ypt < NROWS) {
				outras[xpt + ypt*NCOLS]=255;
				++non;
				}
			}
		}
	for(ipt=0;ipt < theKernel.nBoundaryPoints; ipt++) { // black out the rim
		xpt = (int)(arrayDefiningEllipsoid[1] + theKernel.rotatedBoundaryOffsets[ipt][0]);
		
		if(xpt>=0 && xpt < NCOLS) {
			ypt = (int)(arrayDefiningEllipsoid[2] + theKernel.rotatedBoundaryOffsets[ipt][1]);
			if(ypt>=0 && ypt < NROWS) {
				outras[xpt + ypt*NCOLS] = 0;
				++non;
				}
			}
		 }
// paints this ball onto outras and returns the number of pixels turned on.
// It turns on only pixels that were off.
	return (non);
	}
	
static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}
