package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.lang.*;
import java.util.*;
import java.awt.*;  
import javax.swing.*;
import infoCCD.*;

class BioradFile extends RandomAccessFile   {
	ShowPicJPanel parent=null;

	int trailerLength=0;
	byte [] theTrailer=null; // the trailer holds the biorad info about lens,scanrate, etc.
	String goodChars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ.= _-+,/;1234567890";

	int		width, height, numJFrames, num_pics, merged, magic,width_height=0;
	double dxy=-1., dz=-.1;
	BioradStackConfig myConfig = null;
	int		interlace = -1;		// Assume no interlace
	int ifRGB=0;
	public	String	filename,filesName;		// Filename includes full path, filesName doesn't
	byte []	rawdata = null;
	byte []	temp16bitinras = null;

	RasterCache rascache = null;
	int maxCache = 7;
	boolean ifdiags=false;
	boolean ifValidFile=false;
	boolean ifMainStackFile=false;
	static final int bioradHeaderLength = 76;
	byte [] header = new byte[bioradHeaderLength]; // holds the header.
	boolean ifDataFromWeb = false;
	sectionConsumerClient webStack;
	int sectionType=0;
	int pixel_bit_depth = 8;
	byte [] squash16to8 = null;
	boolean ifUseSqrtTransformFor16BitData = false;

	
	public void useSqrtTransform(boolean ifTrue) {
		if(ifUseSqrtTransformFor16BitData != ifTrue)
			cachePurge(); // so that all sections are shown with the new setting of ifUseSqrtTransformFor16BitData
		ifUseSqrtTransformFor16BitData = ifTrue;
		}
		
	private void extractHighBytes(byte [] inras, byte [] raster, int npix) {
		int i,low,high,val16,j=0; 
		  for(i=0;i<npix;i++) {
			j++; // bypass the low byte
			if( (high = inras[j++]) < 0)
				raster[i] = (byte)(high + 256);
			else
				raster[i] = (byte)high;
			}
		}	


	private void extractSqrtTransformedBytes(byte [] inras, byte [] raster, int npix) {
		int i,low,high,val16,j=0; 
		talk("here we go");
		  for(i=0;i<npix;i++) {
			if((low = inras[j++]) < 0)
				low += 256;
			if( (high = inras[j++]) < 0)
				high = 256*(high+256) + low;
			else
				high = 256*high + low;
			raster[i] = squash16to8[high];
			}
		}	

	public void webCacheLevel(int lev) {
		if(webStack != null) {
			webStack.webCacheLevel(lev);
			}
		}
						
	public void tellName(String chanName) {
		talk(chanName+" channel shows: "+filename);
		}
			
	public BioradFile(ShowPicJPanel parent, int maxCache, int ifRGB, String filename,SectionNavigator navigator,
			int sectionType) throws Exception{
		super(filename, "r"); // cannot open "rw" or else each file can be opened only once only one 
		// apparently, I must open some actual file, but we will not use it;
	try {close();}
		catch(IOException e) {
			talk("Could not close BioradFile = "+filename);
			} 
	//talk("In bioradfile constructor for webStacks. Tried to open");
	//talk(filename);
		this.sectionType = sectionType;
		// open a stack taking picture data from a web server
		talk("BioradFIle.Constuctor for web files");
		this.ifRGB=ifRGB;
		this.maxCache = maxCache;
		this.parent = parent;
		ifMainStackFile=true;
		int theRedPort=7601,theGreenPort=7602,theBluePort=7603;
		int portNum=theRedPort;
		ifDataFromWeb=true;
		if(ifRGB==2) {
			this.filename = "webBlue";
			portNum=theBluePort;
			portNum=theBluePort;
			}
		else if(ifRGB==1) {
			this.filename = "webGreen";
			portNum=theGreenPort;
			}
		else { // this is the default
			this.filename = "webRed";
			portNum=theRedPort;
			}
		// null for hostname in next line means search this subnet for any server.
		try{
			webStack = new sectionConsumerClient(null,portNum,navigator,ifRGB);			
			}
		catch (Exception e) {
			talk("failed to create new sectionConsumerClient, threw exception "+e);
			webStack=null;
			}
		//talk("try getBioradHeaderFromWeb() ");
		ifValidFile=getBioradHeaderFromWeb(webStack);
		if(ifValidFile) {
		    rascache = new RasterCache(num_pics,maxCache); 
		    }
		else talk("getBioradHeaderFromWeb() failed");
		// last arg is max number of frames to cache
		//talk("try BioradStackConfig() ");
		myConfig = new BioradStackConfig(ifRGB,parent,this);
		switch (ifRGB) {
				case 0: parent.parent.exhibitRedChanLUT.canvas.myMaster=myConfig; break;
				case 1: parent.parent.exhibitGreenChanLUT.canvas.myMaster=myConfig; break;
				case 2: parent.parent.exhibitBlueChanLUT.canvas.myMaster=myConfig; break;
				}
		//talk("try myConfig");
		myConfig.init();
		}
	
	public BioradFile(ShowPicJPanel parent, String filename, int maxCache, int ifRGB) throws IOException {
		super(filename, "r"); // cannot open "rw" or else each file can be opened only once only one 
		this.filename = filename;
		this.ifRGB=ifRGB;
		this.maxCache = maxCache;
		this.parent = parent;
		ifMainStackFile=true;
		ifDataFromWeb=false;
		filesName = (new File(filename)).getName();
		talk("Great, we successfully opened filename = "+filename);
		talk("In contructor II, calling readBioradHeader() ");
			// Read in the header
		ifValidFile=readBioradHeader(false);
		if(ifValidFile) {
		    rascache = new RasterCache(num_pics,maxCache); 
		    }
		else talk("readBioradHeader() failed");
		// last arg is max number of frames to cache
		myConfig = new BioradStackConfig(ifRGB,parent,this);
		switch (ifRGB) {
				case 0: parent.parent.exhibitRedChanLUT.canvas.myMaster=myConfig; break;
				case 1: parent.parent.exhibitGreenChanLUT.canvas.myMaster=myConfig; break;
				case 2: parent.parent.exhibitBlueChanLUT.canvas.myMaster=myConfig; break;
				}
		myConfig.init();
		}
	
	public BioradFile(String filename) throws IOException {
		super(filename, "r");
		this.filename = filename;
		this.maxCache = 0;
		filesName = (new File(filename)).getName();
		talk("In contructor III, calling readBioradHeader() ");
		// Read in the header
		ifValidFile=readBioradHeader(false);
		}
		
	public BioradFile(String filename, boolean ifAnnounceHeader) throws IOException {
		super(filename, "r");
		this.filename = filename;
		this.maxCache = 0;
		filesName = (new File(filename)).getName();
		talk("In contructor IV, calling readBioradHeader() ");
		ifValidFile=readBioradHeader(ifAnnounceHeader);
		}
	
	public void sepuku() {
		empty();
		try {close();}
		catch(IOException e) {
			talk("Could not close BioradFile = "+filename);
			}
		if(myConfig != null) 
			myConfig.sepuku();
		if(webStack != null)
			webStack.sepuku();
		if(ifMainStackFile) switch (ifRGB) {
				case 0: parent.parent.exhibitRedChanLUT.canvas.myMaster=null; break;
				case 1: parent.parent.exhibitGreenChanLUT.canvas.myMaster=null; break;
				case 2: parent.parent.exhibitBlueChanLUT.canvas.myMaster=null; break;
				default : break;
				}
		myConfig=null;
		rascache=null;
		parent=null;
		}
		
	public void finalize() {
		talk("                                                                 BioradFile with ifRGB="+ifRGB+" snuffs");
		}
		
	public boolean isValid() {
	    return ifValidFile;
	    }
		
	private boolean getBioradHeaderFromWeb(sectionConsumerClient webStack) {
		int [] headerDat = webStack.getStackSize();
	    merged = headerDat[3];
	    magic = headerDat[4];
	    switch (sectionType) {
	    	default:
	    	case 0:
	    		width = headerDat[0]; // actual X-width
	    		height = headerDat[1]; // actual Y-height
	    		num_pics = headerDat[2]; // actual number of sections (including interpolated ones)
	    		break;
	    	
	    	case 1: // XZ sections
	    		width = headerDat[0]; // actual X-width
	    		height = headerDat[2]; // actual number of sections (including interpolated ones)
	    		num_pics = headerDat[1]; // actual Y-height
	    		break;
	    	
	    	case 2: // ZY sections
	    		width = headerDat[1]; // actual Y-height
	    		height = headerDat[2]; // actual number of sections (including interpolated ones)
	    		num_pics = headerDat[0]; // actual X-width
	    		break;
	    	}
	    	
	    //talk("Biorad webStack header indicates: width="+width+", height="+height+", with "+num_pics+" sections, merged="+merged);	    	
	    width_height = width * height;
	    if((numJFrames = num_pics) <= 0 || !(webStack.ifOnLine)) {
	    	talk(numJFrames+" = numJFrames negative");
			return false;
			}
		else 
	    	return true;		
	    }

	private boolean readBioradHeader(boolean ifTalk) throws IOException {
		talk("Start in readBioradHeader()");
		if(read(header) != bioradHeaderLength) {
			talk("readBioradHeader() failed by not getting "+bioradHeaderLength+" bytes");
			return false; // flag invalid or nonexistant file...
			}
		 talk("We are in : readBioradHeader()");
		// Convert header carefully; java only has signed bytes
	    if(header[0] < 0) width = 256 + header[0] + 256 * header[1];
	    else width = header[0] + 256 * header[1];
	    if(header[2] < 0) height = 256 + header[2] + 256 * header[3];
	    else height = header[2] + 256 * header[3];
	    if(header[4] < 0) num_pics = 256 + header[4] + 256 * header[5];
	    else num_pics = header[4] + 256 * header[5];
	    merged = header[50] + 256 * header[51];	// I don't use these, so ignore subtleties
	    magic = header[54] + 256 * header[55];
	    talk("Biorad File header indicates: width="+width+", height="+height+", with "+num_pics+" sections, merged="+merged);	    	
		talk("header[4]="+header[4]+", header[5]="+header[5]+", decodes to num_pics="+num_pics);
	    width_height = width * height;
		talk("length() = "+length()+", and width_height*num_pics = "+(width_height*num_pics));
	    if((numJFrames = num_pics) <= 0) {
	    	talk(numJFrames+" = numJFrames negative");
			return false;
			}
		else {
			// read and store the trailer, if any;
			if(length() > 2*width_height*num_pics) {  //  these are 16 bit data
				talk("It's 16-bit data...");
				pixel_bit_depth = 16;
				seek(2*width_height*num_pics + bioradHeaderLength-1); // anything after this is trailer
				}
			else {
				seek(width_height*num_pics + bioradHeaderLength-1); // anything after this is trailer
				talk("It's 8-bit data");
				pixel_bit_depth = 8;
				}
			byte [] trail = new byte[1000000];
			int got,templen=10000;
			byte [] tempBytes = new byte[1];
			boolean hitEnd=false;
			trailerLength=0;
			while(trailerLength < 10000000 && !hitEnd) {
				try { 
					got=read(tempBytes);
					for(int j=0;j<got;j++)
						trail[trailerLength+j]=tempBytes[j];
					trailerLength += got;
					if(got==tempBytes.length)
						hitEnd=false;
					else
						hitEnd=true;
					}
				catch(IOException e) {
					talk("the trailer is "+trailerLength+" long, but got IOException.");
					hitEnd=true;
					}
				}
			talk("Trailer length = "+trailerLength);
			theTrailer = new byte[trailerLength];
			for(int i=0;i<trailerLength;i++)
				theTrailer[i]=trail[i];
			trail = null;
			parseTrailer();										
	    	return true;
			}
	    }

	public void empty() {
		if(rascache != null)
		    rascache.empty();
		rascache=null;
		}
	
	public void cachePurge() {
		if(rascache != null)
		    rascache.empty();
		}
	
	public int adjustCacheSize(int cs) {
		if(rascache != null) 
			return(rascache.adjustCacheSize(cs));
		else return 0;
		}
		
	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
			
	private void gainDepthCorrection(int frame, byte[] raster) {
		int i,k;
		byte [] sectLUT = myConfig.GainBleachCorrectionLUT_ForSect(frame);
		if(raster != null) 
		  for(i=0;i<width_height;i++) {
			k= 256 + raster[i];
			raster[i] = sectLUT[k];
			}
		}

	public byte [] readShiftedBioradRaster(int frame) {
		// just like readBioradRaster() but incorporates the Xshifts,Yshifts,Zshifts
		// BioradCanvas curStack
		int xshift = parent.curStack.Xshifts[ifRGB];
		int yshift = parent.curStack.Yshifts[ifRGB];
		int zshift = parent.curStack.Zshifts[ifRGB];
		byte [] raster = readBioradRaster(frame-zshift);
		if(raster==null || (xshift==0 && yshift==0))
			return raster;
		// do the x and/or y shift(s) in a new raster so as not to foul up the cached frame
		// because in future, xshift and/or yshift may change.
		
		switch (ifRGB) {
			case 0 :
				talk("return shifted red frame "+frame+" with zshift="+zshift+", xshift="+xshift+
					", and yshift="+yshift);
				break;
				
			case 1 :
				talk("return shifted green frame "+frame+" with zshift="+zshift+", xshift="+xshift+
					", and yshift="+yshift);
				break;
				
			case 2 :
				talk("return shifted blue frame "+frame+" with zshift="+zshift+", xshift="+xshift+
					", and yshift="+yshift);
				break;
			default :
				talk("OOPS, ifRGB= "+ifRGB+" in BioradFile.readShiftedBioradRaster");
				break;
			}
				
		byte [] shiftedRaster = new byte[width_height];
		int row,col,rowoff,shrow,shcol,shrowoff;
		for(row=0;row<height;row++) {
			rowoff = row*width;
			shrow = row+yshift;
			if(shrow>=0 && shrow<height) {
				shrowoff = shrow*width;
				for(col=0;col<width;col++) {
					shcol = col+xshift;
					if(shcol>=0 && shcol<width)
						shiftedRaster[rowoff+col]=raster[shrowoff+shcol];
					else					
						shiftedRaster[rowoff+col]=0;
					}
				}
			else  {
				for(col=0;col<width;col++) 
					shiftedRaster[rowoff+col]=0;
				}
			}				
		return (shiftedRaster);
		}
		
	private float [] getBiorad16bitFloatRaster(int frame) {
		/* this is used only by the Deconvolver class. Do not cache it. This BIo-rad file does have 16-bit data,
		so use it.*/
		int i,low,high,j=0; 
		float [] floatras= new float[width_height];
		double eents = 1/256.0;
		
		if(temp16bitinras==null || temp16bitinras.length != 2*width_height) 
		   temp16bitinras = new byte[2*width_height]; 
		   
		talk("read 16-bit data into float array for frame "+frame);	
		   
		try {
		    seek(width_height*frame*2 + bioradHeaderLength);	// Get to beginning of frame
		    talk("reading 16-bit data from section number "+frame);
		    read(temp16bitinras);
		    }
		catch(IOException e) {
			talk("Got exception "+e+" while reading 16-bit data for section number "+frame);
			return null;
			}
		  
		for(i=0;i<width_height;i++) {
			if((low = temp16bitinras[j++]) < 0)
				low += 256;
			if( (high = temp16bitinras[j++]) < 0)
				floatras[i] =   (float)(high + 256.00 + eents*low);
			else
				floatras[i] = (float)(high + eents*low);
			}
		return floatras;   		 	
		}
			
	public byte [] readBioradRawByteRaster(int frame) {
		/* this is used only by the Deconvolver class. Do not cache it. If this BIo-rad file has 16-bit data,
		return two-bytes-per-pixel data. Else just return one-byte-per-pixel data. */
		int bytecount = (pixel_bit_depth>8)? 2*width_height : width_height;
		if(temp16bitinras==null || temp16bitinras.length != bytecount) 
		   temp16bitinras = new byte[bytecount]; 
		try {
		    seek(bytecount*frame + bioradHeaderLength);	// Get to beginning of frame
		    }
		catch(IOException e) {
		    talk("On seek "+(bytecount*frame + bioradHeaderLength)+" got exception "+e);
			return null;
		    }
		    
		try { // read the bute raster...
		    	read(temp16bitinras);
		    	talk("BioradFile.readBioradRawByteRaster("+frame+") read "+bytecount+" byte raster starting at loction "+
		    				(bytecount*frame + bioradHeaderLength));
		    	}
		catch(IOException e) {
    		talk("On read of raster with "+bytecount+" bytes, got exception "+e);
			return null;
		    }
		return temp16bitinras;
		}

	public float [] readBioradFloatRaster(int frame) {
		/* this is used only by the Deconvolver class. Do not cache it. If this BIo-rad file has 16-bit data,
		use it. Else just use 8-bit data. */

		if(pixel_bit_depth>8) // it's 16-bit data.
			return getBiorad16bitFloatRaster(frame);
		
		talk("read 8-bit data into float array for frame "+frame);	
		if(temp16bitinras==null || temp16bitinras.length != width_height) 
		   temp16bitinras = new byte[width_height]; 

		byte [] raster = temp16bitinras;
		int pix;
		   try {
		    seek(width_height*frame + bioradHeaderLength);	// Get to beginning of frame
		    }
		catch(IOException e) {
		    talk("On seek "+(width_height*frame + bioradHeaderLength)+" got exception "+e);
			return null;
		    }
		    
		  try { // read the bute raster...
		    	read(temp16bitinras);
		    	}
		  catch(IOException e) {
    			talk("On read of raster with "+width_height+" bytes, got exception "+e);
				return null;
		    	}
		   // float the byte raster.
		float [] floatras= new float[width_height];
		for(int i=0;i<width_height;i++) {
			if((pix=temp16bitinras[i])<0) 
				floatras[i] = pix + 256;
			else
				floatras[i]= pix;
			}
		return floatras;   		 	
		}
		
		
	public byte [] readBioradRaster(int frame) {
		byte	[] raster;
		int i,j,ioff,k=0;
		if(frame<0 || frame >= numJFrames || !ifValidFile) {
			//talk("readBioradRaster("+frame+") returns null. ifValidFile="+ifValidFile);
			return null;
			}
		
		if(rascache==null || (raster=rascache.fetch(frame)) == null) {
 		  if(ifdiags) talk("raster cache miss; read section "+frame);
		  if((raster =  readInBioradRaster(frame)) == null){ 
				talk("Error while loading frame " + frame + " of " + filename);
				raster = null;
				return(null);
				}
		  if(ifdiags) talk("bioradfile read the raw data.");
		  if(rascache != null) {
		  	gainDepthCorrection(frame, raster);
		  	rascache.add(frame,raster);
		  	}
		  return(raster);
		  }
		
		else {
		    if(ifdiags) talk("a cache hit");
			gainDepthCorrection(frame, null); // to draw the LUT...
		    return(raster); 
		    }
		}

	public byte [] readGmoHeader() {
		byte	[] raster;
		  if((raster = readInBioradRaster(0)) == null) 
			talk("Error while reading the header ");
		return raster;
		}

	private byte []  readInBioradRaster(int frame){
		if(ifDataFromWeb) {
			switch (sectionType) {
				default :
				case 0:
					return(webStack.getXYsect(frame));
				case 1:
					return(webStack.getXZsect(frame));
				case 2:
					return(webStack.getZYsect(frame));
					}
				}
		byte [] raster = new byte[width_height];
		readInBioradRaster(frame, raster);
		return raster;
		}
		
	private void make16to8LUT() {	
		int b, bmax = 256*256;
		double toodim=16;
		double scalar = 255./Math.sqrt((double)(bmax-toodim)/toodim);
		if(squash16to8 == null) {
			squash16to8 = new byte[256*256];
			for(b=0;b<bmax;b++) {
				if(b<=toodim) 
					squash16to8[b] = 0;
				else 
					squash16to8[b] = (byte)(.499 + scalar*Math.sqrt((double)(b-toodim)/toodim));
				}
			}
		}
	
	public void readInBioradRaster16bit(int frame, byte [] raster) throws Exception {
		make16to8LUT();
		if(interlace >= 0) {
		    if(rawdata == null) rawdata = new byte[width_height*2];
		    seek(width_height*2*frame + bioradHeaderLength);
		    read(rawdata);
		    
		    for(int i = 0; i < width_height; i++) {
				if(interlace == 0) 
					raster[i] = rawdata[i*2];
				else raster[i] = rawdata[i*2+1];
		    	}
		    rawdata=null;
			}
		else {
		    seek(width_height*frame*2 + bioradHeaderLength);	// Get to beginning of frame
		    if(temp16bitinras==null || temp16bitinras.length != 2*raster.length) {
		    		temp16bitinras = new byte[2*raster.length]; 
		    		}
		    talk("ifUseSqrtTransformFor16BitData = "+ifUseSqrtTransformFor16BitData);
		    read(temp16bitinras);
			if(ifUseSqrtTransformFor16BitData) {
		    	talk("read the file, use sqrt() transform.");
				extractSqrtTransformedBytes(temp16bitinras, raster, width_height);
				}
			else {
		    	talk("read the file, high byte only.");	
		    	extractHighBytes(temp16bitinras,raster,width_height); // divide by 256
		    	}
		    }
		}
	
	private void readInBioradRaster(int frame, byte [] raster) {
		boolean ifGotGoodRaster=true;
		if(ifDataFromWeb) {
			switch (sectionType) {
				default :
				case 0:
					webStack.getXYsect(frame,raster);
					break;
				case 1:
					webStack.getXZsect(frame,raster);
					break;
				case 2:
					webStack.getZYsect(frame,raster);
					break;
					}
			}
		else {
		if(interlace >= 0) {
		    if(rawdata == null) rawdata = new byte[width_height*2];
		    try {
		    	seek(width_height*2*frame + bioradHeaderLength);
		    	}
		    catch(IOException e) {
		    	talk("On seek "+(width_height*2*frame + bioradHeaderLength)+" got exception "+e);
				ifGotGoodRaster=false;
		    	}
		    try {
		    	read(rawdata);
		    	}
		    catch(IOException e) {
    			talk("On read of interlaced raster with "+width_height*2+" bytes, got exception "+e);
				ifGotGoodRaster=false;
		    	}
		    if(ifGotGoodRaster)
		      for(int i = 0; i < width_height; i++) {
				if(interlace == 0) raster[i] = rawdata[i*2];
				else raster[i] = rawdata[i*2+1];
		    	}
		    rawdata=null;
			}
		else {
			if(pixel_bit_depth>8) {
				parent.makeUseSqrtTransformButton();
				try {readInBioradRaster16bit(frame, raster);}
				catch(Exception e) {
					talk("Got exception "+e+" trying to read section "+frame+" from 16 bit stack");
					ifGotGoodRaster=false;
		    		}					
				}
			else {
		    	try {
		    		seek(width_height*frame + bioradHeaderLength);	// Get to beginning of frame
		    		}
		    	catch(IOException e) {
		    		talk("On seek "+(width_height*frame + bioradHeaderLength)+" got exception "+e);
					ifGotGoodRaster=false;
		    		}
		    	try {
		    		read(raster);
		    		}
		    	catch(IOException e) {
    				talk("On read of raster with "+width_height+" bytes, got exception "+e);
					ifGotGoodRaster=false;
		    		}
		    	}
		    }
		}
		if(ifGotGoodRaster && myConfig != null) 
			myConfig.blotOutHoles(raster, width, height, frame);
	}
	
	
	// These two functions set and get the value of interlace, which 
	// specifies whether its a one (= -1) or two channel stack,and if 
	// two channel, whether to read the first (= 0) or second (= 1) channel.
	public void setInterlace(int val) { // never used
		interlace = val; 
		}
		
	public int getInterlace() {
		 return(interlace); 
		}

	/** Below here are methods added in august 2001 to implement writing the info in the header
	and in the trailer of biorad files to an html file. */

	public void writeTrailer(RandomAccessFile outfile, int numsects) {
		int trailerLocation = pixel_bit_depth>8? 2*width_height*numsects  : width_height*numsects;
		trailerLocation += (bioradHeaderLength-1);
		
		if(trailerLength<=0 || outfile==null) {
			talk("There was no trailer.");
			return;
			}
		try{
			outfile.seek(trailerLocation); // anything after this is trailer
			outfile.write(theTrailer);
			talk("Wrote trailer of "+trailerLength+" bytes");
			}
		catch(IOException e) {
			talk("something went wrong while writing the trailer");
			}
		}    
	    	    	
	private boolean isNormalCharacter(byte [] oneByte) {
		String stg = new String(oneByte);
		if(goodChars.indexOf(stg) >=0)
			return true;
		else return false;
		}
	
	private void pluckInterestingParams(String lcstring) {
					/* pick out and store interesting parameters...
axis_0 004 0.000000e+000 1.000000e+000 intensity<br>
axis_2 001 0.000000e+000 1.983333e-001 microns<br>  this is the dxy parameter
axis_3 001 0.000000e+000 1.983333e-001 microns<br>  this is the dxy parameter
axis_4 001 7.000000e+000 2.000000e-001 microns<br>  this is the dz parameter
      i1  i2            i3            i4  */
		if(lcstring.indexOf("axis_2 ") >=0) {
			//talk("axis_2 in  pluckInterestingParams("+lcstring+")");
			dxy = fourthFieldToDouble(lcstring);
			}
		if(lcstring.indexOf("axis_3 ") >=0) {
			//talk("axis_3 in  pluckInterestingParams("+lcstring+")");
			dxy = fourthFieldToDouble(lcstring);
			}
		if(lcstring.indexOf("axis_4 ") >=0) {
			//talk("axis_4 in  pluckInterestingParams("+lcstring+")");
			dz = fourthFieldToDouble(lcstring);
			}
		if(lcstring.indexOf("pixel_bit_depth") >=0) {
			talk("pixel_bit_depth in  pluckInterestingParams("+lcstring+")");
			int i1 = lcstring.indexOf("=") +1;
			if(i1>0) {
				pixel_bit_depth = (int)(Double.valueOf(lcstring.substring(i1+1)).doubleValue());
				talk("-------------------------------> pixel_bit_depth = "+pixel_bit_depth);
				}
			}
		//talk(" pluckInterestingParams("+lcstring+") returns");
		}
				
	private double fourthFieldToDouble(String lcstring) {
		int len=lcstring.length();
		int i1  = lcstring.indexOf(" ");
		
		if(i1<0 || i1>len-5) return 0.;
		int i2 = lcstring.indexOf(" ",i1+1);
		
		if(i2<0 || i2>len-5) return 0.;
		int i3 = lcstring.indexOf(" ",i2+1);
		
		if(i3<0 || i3>len-5) return 0.;
		int i4 = lcstring.indexOf(" ",i3+1);
		if(i4<0) i4=len;
		
		double val = Double.valueOf( lcstring.substring(i3+1,i4) ).doubleValue();
		//talk("fourthFieldToDouble() returns value "+val);
		return val;
		}
		
					
	/** writes the trailer to a text file...*/
	public void writeTrailerToHTMLFile(int irgb) {
		String [] htmlHead = {"<!DOCTYPE html public \"-//w3c//dtd html 4.0 transitional//en\">",
			"<HTML>",
			"<HEAD>",
		  	"<META NAME=\"GENERATOR\" CONTENT=\"Adobe PageMill 3.0 Mac\">",
		  	"<META HTTP-EQUIV=\"Content-Type\" CONTENT=\"text/html; charset=iso-8859-1\">",
		  	"<META NAME=\"GENERATOR\" CONTENT=\"Mozilla/4.7C-CCK-MCD {C-UDP; EBM-APPLE} (Macintosh; I; PPC) [Netscape]\">",
		  	"<TITLE>bio-rad settings</TITLE>",
		  	"</HEAD>",
		  	"<BODY TEXT=\"#ffa83f\" BGCOLOR=\"#000000\" LINK=\"#00ffff\" VLINK=\"#ff00ff\"",
		  	"ALINK=\"#0000ff\">",
		  	" ",
		  	"<FONT SIZE=\"-1\" FACE=\"Monaco\">"};
		  	
		String htmlEOL = "<BR>";
		byte [] htmlEOLbuf = new byte[4];
		htmlEOL.getBytes(0,htmlEOL.length(),htmlEOLbuf,0);
		String [] htmlEnd = {"</FONT>", " ", "</BODY>", "</HTML>"};
		String lcstring;
		RandomAccessFile trailfile=null;
		boolean justWroteEOL=false;
		byte [] oneByte = new byte[1];
		int i=0,k,j=0;
		String oneLine;
		byte [] lineBytes = new byte[256];
		int inline=blankLine(lineBytes);
		boolean hitJunk=false;
		String endFlag = "+,-./0123456789";
		k=filename.indexOf(".pic");
		if(k<0) k=filename.indexOf(".PIC");
		String trailerFileName;
		if(k>0) trailerFileName = filename.substring(0,k);
		else trailerFileName=new String(filename);
 		trailerFileName = trailerFileName+"_info.html";

		try {
			trailfile = new RandomAccessFile(trailerFileName,"rw");
			for(i=0;i<htmlHead.length;i++) // write the html header
				writeFileString(htmlHead[i], trailfile);
			}
		catch(IOException e) {
			talk("got exception writing start of trailer file = "+e);
			return;
			}
		// write stuff extracted earlier from the header
		k = filename.indexOf("dataFiles0/");
		if(k<0) k=0; // don't print the path to get to dataFIles0 because it differs on
					// different computers. 
		writeFileString(("name = "+(filename.substring(k))+htmlEOL), trailfile);
		writeFileString(("-------from file header------>"+htmlEOL), trailfile);
		writeFileString(("raster width = "+width+htmlEOL), trailfile);
		writeFileString(("raster height = "+height+htmlEOL), trailfile);
		writeFileString(("number of sections = "+num_pics+htmlEOL), trailfile);
		writeFileString(("--------from file trailer--------->"+htmlEOL), trailfile);
		i=0; // i counts through the array holding the trailer
		while(i<trailerLength && !hitJunk) {
			oneByte[0] = theTrailer[i];
			if(isNormalCharacter(oneByte)) {
					lineBytes[inline++] = theTrailer[i];
					if(inline>249) {
						inline=249;
						//talk("inline rose to "+inline);
						}
					justWroteEOL=false;
					}
			else if(!justWroteEOL) { // this line just ended with some unknown junk
				for(k=0;k<4;k++)
					lineBytes[inline++] = htmlEOLbuf[k];
				//talk("Line ends with inline = "+inline);
				justWroteEOL=true;
				// convert byte[] buffer to string;
				oneLine = new String(lineBytes); 
				oneLine = oneLine.trim();// this had lots of trailing blanks
				//test to see if we have hit a particular string of junk which flags the end of meaningful data
				if(oneLine.indexOf(endFlag) >= 0) hitJunk=true; // this terminates parsing of the 
																// trailer file.
				else { // it's a good line; write it to the file.
					if(oneLine.length() > 12) { /* there are no valid lines of length <= 12 after counting the
								4 bytes that the html EOL occupy. I.e. 8-character lines are probably junk */
						// change to lower case to make it easier to read.
						lcstring = oneLine.toLowerCase();
						writeFileString(lcstring, trailfile);
						// pluckInterestingParams(lcstring); don't need this because parseTrailer() does it
						}
					inline =blankLine(lineBytes); // blank the line; prepare to accept next line.
					}
				} 
			++i;	
			} // when fall through, we have written the contents of the trailer.
			
		try { // write the html termination and close the file.
			for(i=0;i<htmlEnd.length;i++) // write the html ending
				writeFileString(htmlEnd[i], trailfile);
			// write some blanks to oerwrite any stuff left over from earlier version
			for(i=0;i<8;i++) // write the html ending
				writeFileString(("                                     "+htmlEOL), trailfile);
			trailfile.close();
			// send the URL for this file to the rgbFileInfoViewer in infoCCD
			Info.showRGBFileInfo("file:"+trailerFileName,irgb);	
			}
		catch(IOException e) {
			talk("got exception finishing the trail file ("+trailerFileName+") exception= "+e);
			return;
			}
		//talk("Wrote HTML bio-rad info file "+trailerFileName);
		// talk("The laserSharp info files are the same for all the channels, so we write only one of them.");
		//talk("   ");
		}

	private void parseTrailer() {
		boolean hitJunk=false;
		int i=0,k; // i counts through the array holding the trailer
		String oneLine, lcstring;
		byte [] lineBytes = new byte[256];
		byte [] oneByte = new byte[1];
		int inline=blankLine(lineBytes);
		String endFlag = "+,-./0123456789";
		boolean justWroteEOL=false;
		//talk("In parseTrailer()");
		while(i<trailerLength && !hitJunk) {
			oneByte[0] = theTrailer[i];
			if(isNormalCharacter(oneByte)) {
				//talk("got a <"+oneByte[0]+">, inline="+inline);
					lineBytes[inline++] = theTrailer[i];
					if(inline>249) inline=249;
					justWroteEOL=false;
					}
			else if(!justWroteEOL) { // this line just ended with some unknown junk
				justWroteEOL=true;
				if(inline<=0) {
					lineBytes[inline++] = (byte)' ';
					//talk("Came up with blank line??");
					}
				// convert byte[] buffer to string;
				oneLine = new String(lineBytes); 
				oneLine = oneLine.trim();// this had lots of trailing blanks
				//talk("Read in line of "+oneLine.length()+" characters");
				//test to see if we have hit a particular string of junk which flags the end of meaningful data
				if(oneLine.indexOf(endFlag) >= 0) {
					hitJunk=true; // this terminates parsing of the 
					//talk("hitJunk=true");
					}
																// trailer file.
				else { // it's a good line; write it to the file.
					//talk("got good line");
					if(oneLine.length() > 12) { /* there are no valid lines of length <= 12 after counting the
								4 bytes that the html EOL occupy. I.e. 8-character lines are probably junk */
						// change to lower case to make it easier to read.
						//talk("convert line to lower case");
						lcstring = oneLine.toLowerCase();
						//talk("pluck interesting parameters");
						pluckInterestingParams(lcstring);
						}
					inline =blankLine(lineBytes); // blank the line; prepare to accept next line.
					}
				} 
			++i;	
			} // when fall through, we're through.
		}
		
			
	private int blankLine(byte [] aline)	 {
		// blanks the line and returns 1;
		int len = aline.length;
		for(int i=0;i<len;i++)
			aline[i]=32;
		return 1;
		}
		
	private void writeFileString(String stg, RandomAccessFile outputFile) {
		byte [] buf = new byte[stg.length()+1];
		stg.getBytes(0,stg.length(),buf,0);
		buf[stg.length()] = 10;
			try {
			outputFile.write(buf);
			}
		catch (IOException ioex) {
			talk("write to file failed: "+ioex);
			}
		}

}
