package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.lang.*;
import java.util.*;
import java.awt.*;  
import javax.swing.*;


class BlackHoles extends JFrame implements java.awt.event.ActionListener,java.awt.event.WindowListener {
	ShowPicJPanel picJPanel;
	int nomwidth=170, nomheight=282;
	final static int PICK_NEAREST_HOLE = 73;
	final static int MOVE_SELECTED_HOLE = 74;
	final static int ADD_NEW_HOLE = 75;
	boolean actOnIncomingMouseLocs = false;
	double zCoordOfDrawnSection = 0.; 
	int [] zero = {0, 0, 0};
	SphericalHole rootHole = null;
	SphericalHole selectedHole = null;
	JTextField pixPeruMText=null;
	double pixPeruM=12.; // ratio of pixel width to depth of one section.
	JButton redBlot,greenBlot,blueBlot;
	boolean ifBLotRedChan=false,ifBLotGreenChan=false,ifBLotBlueChan=false;
	byte blotBlack = 0;

	public void blotOutHoles(BioradStackConfig BSC, byte [] ras, int ncols, int nrows, int nsect) {
		switch (BSC.ifRGB) {
			case 0: // red channel
				if(!ifBLotRedChan) return;
				break;
			
			case 1: // green channel
				if(!ifBLotGreenChan) return;
				break;
			
			default:
			case 2: // blue channel
				if(!ifBLotBlueChan) return;
				break;
			}
	    SphericalHole ahole = rootHole;
	    while(!ahole.isLast()) {
			ahole = ahole.next;
			ahole.blotOutData(BSC, ras,  ncols, nrows, nsect, pixPeruM, blotBlack);
			}
	    }
	
    public BlackHoles(ShowPicJPanel picJPanel) {   
	super("Black Hole Locator");
	this.picJPanel = picJPanel;
	WidgetMaker wmk = new WidgetMaker(this);
	// initiate the linked list of holes
	rootHole = new SphericalHole(this);
	selectedHole= rootHole;
	getContentPane().setLayout(new GridLayout(0, 1)); 
    setSize(nomwidth, nomheight);
	setBackground(Color.yellow);

	getContentPane().add(wmk.makeAJButton("add new hole at next click",255,255,0));	
	getContentPane().add(wmk.makeAJButton("shrink selected hole",245,245,8));	
	getContentPane().add(wmk.makeAJButton("enlarge selected hole",235,235,16));
	getContentPane().add(wmk.makeAJButton("GoTo selected hole",225,225,24));	
	getContentPane().add(wmk.makeAJButton("select next hole",215,215,32)); 
	getContentPane().add(wmk.makeAJButton("select previous hole",205,205,40)); 
	getContentPane().add(wmk.makeAJButton("Select first hole",180,195,60)); 
	getContentPane().add(wmk.makeAJButton("next click picks nearest hole",160,185,80)); 
	getContentPane().add(wmk.makeAJButton("next click moves selected hole",140,175,100)); 
	getContentPane().add(wmk.makeAJButton("show all holes",120,165,120)); 
	getContentPane().add(wmk.makeAJButton("hide all holes",90,160,140)); 
	getContentPane().add(wmk.makeAJButton("redraw",70,180,170)); 
	getContentPane().add(wmk.makeAJButton("DELETE SELECTED HOLE",50,190,180)); 

	Color bright = Color.yellow;
	JPanel pan = new JPanel();
	pan.setBackground(bright);
	pan.setLayout(new GridLayout(1,2));
	pan.add(wmk.makeAJButton("set pix/uM",bright));
	pixPeruMText = wmk.makeAJTextField((""+pixPeruM),6,bright);
	pan.add(pixPeruMText);
	getContentPane().add (pan);

	getContentPane().add(wmk.makeAJLabel("toggles tell if holes blot out data",bright));
	
	pan = new JPanel();
	pan.setBackground(Color.gray);
	pan.setLayout(new GridLayout(1,3));
	redBlot = wmk.makeAJButton("RedChan",Color.gray);
	greenBlot = wmk.makeAJButton("GreenChan",Color.gray);
	blueBlot = wmk.makeAJButton("BlueChan",Color.gray);
	pan.add(redBlot);
	pan.add(greenBlot);
	pan.add(blueBlot);
	getContentPane().add(pan);

	getContentPane().add( wmk.makeAJLabel("toggles are active when colored...",bright));
	
	addWindowListener(this);
	}

// catch WindowListener events  
	public void windowActivated(java.awt.event.WindowEvent event) {
		// talk("window activated");
		}
	public void windowDeactivated(java.awt.event.WindowEvent event) {
		//  talk("window Deactivated");
		}
	public void windowOpened(java.awt.event.WindowEvent event) {}
	public void windowIconified(java.awt.event.WindowEvent event) {}
	public void windowDeiconified(java.awt.event.WindowEvent event) {}
	public void windowClosing(java.awt.event.WindowEvent event) {
	    setVisible(false);
		}
	public void windowClosed(java.awt.event.WindowEvent event) {
		}
   
   public void writeHoleListToDisk(ArchiveMarks disker){
	String stg;
        SphericalHole cm;
	int radius;
	int [] xyz={0, 0, 0};
/* sample of what it's supposed to look like
@sphere 1 256 303 146  8    These taken from .rgb files (not 
@sphere 1 262 316 146 8   .tiff files, so the y-components
@sphere 1 492 215 136 8   are correct. No need to invert.
@sphere 1 495 217 144 8
@sphere 1 46 477 146 8
@sphere 1 52 484 148 8
@sphere 1 626 21 124 8
@sphere 1 633 32 124 8  */
	cm=rootHole;
	while(!cm.isLast()) { // there are some hole to write out...
	    cm = cm.next;
	    xyz = cm.tellAbsLocation();
	    xyz[1] = 511-xyz[1]; // invert the y-coordinate
	    xyz[2] *= 2;  // convert Z-coordinate to half-section double count
	    radius = cm.tellRadius();
	    stg = ("@sphere 1 "+xyz[0]+" "+xyz[1]+" "+xyz[2]+" "+radius);
	    disker.writeAString(stg);
	    }
	stg = ("@pixPeruM "+pixPeruM);
	disker.writeAString(stg);
    }

   public void archiveHoles(ArchiveMarks disker, boolean ifWrite) {
        SphericalHole cm;
	int ihole=0, radius;
	int [] xyz={0, 0, 0};
	double badval = 77777.7;
        if(ifWrite) {
	   cm=rootHole;
	   disker.writeNum(pixPeruM, " = pixels per micron in z-direction");
	   while(!cm.isLast()) { // there are some holes to write out...
		cm = cm.next;
		++ihole;
		xyz = cm.tellAbsLocation();
		radius = cm.tellRadius();
		disker.writeNum((double)radius, (" = radius of hole "+ihole));
		disker.writeNum((double)xyz[0], (" = x-coord of hole "+ihole));
		disker.writeNum((double)xyz[1], (" = y-coord of hole "+ihole));
		disker.writeNum((double)xyz[2], (" = z-coord of hole "+ihole));
		}
	    }
	else {
	    talk("eliminating current holes...");
	    deleteAllholes();
	    pixPeruM = disker.getNextFileDouble(badval);
	    talk("Recovered pixPeruM = "+pixPeruM);
 		pixPeruMText.setText(""+pixPeruM);	
	    ihole=0;
	    boolean ifgood = true;
	    cm = rootHole;
	    ifgood=true;
	    while(ifgood) {
		radius = (int)(disker.getNextFileDouble(badval));
		if(radius == (int)badval || radius==(-7777)) ifgood=false;
		talk("got radius == "+radius);
		if(ifgood) {
		    xyz[0] = (int)(disker.getNextFileDouble(badval));
		    if(xyz[0] == (int)badval) ifgood=false;
		    }
		if(ifgood) {
		    xyz[1] = (int)(disker.getNextFileDouble(badval));
		    if(xyz[1] == (int)badval) ifgood=false;
		    }
		if(ifgood) {
		    xyz[2] = (int)(disker.getNextFileDouble(badval));
		    if(xyz[2] == (int)badval) ifgood=false;
		    }
		if(ifgood) { // OK we read in a valid new hole
		    selectedHole = cm.insertNewSphericalHole(xyz, radius, this);
		    cm=selectedHole;
		    cm.appear();
		    cm.tellStats();		     
		    ++ihole;
		    }
		}
		talk("Recovered "+ihole+" holes");
		selectedHole=rootHole;
		showAllHoles();
	      }
	}
   
	public void drawHoles(Graphics g, int xorg, int yorg) {
	    SphericalHole cm = rootHole;
	    g.setColor(Color.white);
	    while(!cm.isLast()) {
		cm = cm.next;
		cm.draw(g,  xorg,  yorg);
		}
	    selectedHole.highlight(g,  xorg,  yorg);
	    }
	    
	public void deleteAllholes(){
	    SphericalHole cm = rootHole;
	    while(!cm.isLast()) cm=cm.next; // go to the last hole
	    while(!cm.isListRoot())
			cm = cm.excise();
		rootHole = new SphericalHole(this); // this discards old listRoot and installs a virgin root
	    }

 	public void hideAllHoles() {
	    SphericalHole cm = rootHole;
	    while(!cm.isLast()) {
		cm = cm.next;
		cm.disappear();
		}
	    }

 	public void showAllHoles() {
	    SphericalHole cm = rootHole;
	    while(!cm.isLast()) {
		cm = cm.next;
		cm.appear();
		}
	    }

   	private void primeMouseForHoles(int whatTask) {
		actOnIncomingMouseLocs=true;
   		if(picJPanel != null) {
			picJPanel.primeMouseForTask(whatTask,this);
			picJPanel.parent.show();
			picJPanel.requestFocus();
			}
		}
	
	public SphericalHole nearestHoleTo(int [] xyz) {
	    SphericalHole cm = rootHole;
	    double dst, dist = 777777.;
	    SphericalHole ClosestSphericalHole = cm;
	    while(!cm.isLast()) {
		cm = cm.next;
		dst = cm.distanceTo(xyz);
		if(dst<dist) {
		    dist=dst;
		    ClosestSphericalHole=cm;
		    }
		}
	    return ClosestSphericalHole;
	    }

	public void processMouseClick(double [] mouseLoc, int doWhat) {
		//  this method is called only by picJPanel after  primeMouseForTask was called
	    int [] xyz = {(int)(mouseLoc[0]+.499),(int)(mouseLoc[1]+.499),
			    (int)(mouseLoc[2]+.499)};
	    if(actOnIncomingMouseLocs) switch (doWhat) {
		case PICK_NEAREST_HOLE:	
		    selectedHole = nearestHoleTo(xyz);
		    selectedHole.appear();
		    reDraw();	
		    break;
		    
		case MOVE_SELECTED_HOLE:
		    selectedHole.moveTo(xyz);
		    reDraw();		
		    break;
		    
		case ADD_NEW_HOLE:
		    selectedHole = 
		      selectedHole.insertNewSphericalHole(xyz, 30, this);
		    selectedHole.appear();
			invalidateRasterCaches();
		    reDraw();		
		    break;
		}
	    else
		talk("No action taken by hole Locator because mouse news arrived too late");
		}
		
// catch button and menu events...
	public void actionPerformed(java.awt.event.ActionEvent event) {
		String arg = event.getActionCommand();
    		actOnIncomingMouseLocs=false; // defeat response to foreign mouse news
    		// if response did not come before some button on this panel was pushed...
			if(arg.equals("GoTo selected hole")) {
			    if(selectedHole.isListRoot()) {
				talk("there is no hole to go to yet");
				}
			selectedHole.appear();
			picJPanel.gotosection((selectedHole.tellAbsLocation())[2]); 
			}
			if(arg.equals("Select first hole")) {
				if(rootHole.isLast()) {
				    selectedHole=rootHole;
				    talk("there are no holes yet");
				    }
				else
				    selectedHole=rootHole.next;
				selectedHole.appear();
				selectedHole.tellStats();
				reDraw();
				}
				
			if(arg.equals("shrink selected hole")) {
			    selectedHole.shrink();
				invalidateRasterCaches();
			    reDraw();
			    }
			
			if(arg.equals("enlarge selected hole")) {
			    selectedHole.grow();
				invalidateRasterCaches();
			    reDraw();
			    }
			
			if(arg.equals("select next hole")) {
				if(selectedHole.isLast()) {
				    talk("you are at the last hole");
				    }
		 		else
				    selectedHole = selectedHole.next;
				selectedHole.appear();
				selectedHole.tellStats();
				reDraw();
		 		}
			if(arg.equals("select previous hole")) {
				if(selectedHole.isFirst()) {
				    talk("there is no previous hole");
				    }
				else
				    selectedHole = selectedHole.prior;
				selectedHole.appear();
				selectedHole.tellStats();
				reDraw();
				}
			if(arg.equals("DELETE SELECTED HOLE")) {
				selectedHole = selectedHole.excise();
				selectedHole.appear();
				invalidateRasterCaches();
				reDraw();
				}
			if(arg.equals("redraw")) {
				reDraw();
				}
			if(arg.equals("add new hole at next click")) { 
				talk("add new hole at next mouse click");
				primeMouseForHoles(ADD_NEW_HOLE);
				}

			if(arg.equals("next click picks nearest hole")) { 
				talk("next mouse picks nearest hole");
				primeMouseForHoles(PICK_NEAREST_HOLE);
				}
			if(arg.equals("next click moves selected hole")) { 
				talk("next mouse relocates selected hole");
				primeMouseForHoles(MOVE_SELECTED_HOLE);
				}				
			if(arg.equals("hide all holes")) { 
				hideAllHoles();
				reDraw();
				}
			if(arg.equals("show all holes")) { 
				showAllHoles();
				reDraw();
				}
			
			if(arg.equals("set pix/uM")) { 
 				pixPeruM = Double.valueOf(pixPeruMText.getText()).doubleValue();	
 				talk("ratio of pixel width to section depth reset to "+pixPeruM);
 				if(ifBLotRedChan) invalidateRasterCaches(0);
 				if(ifBLotGreenChan) invalidateRasterCaches(1);
 				if(ifBLotBlueChan) invalidateRasterCaches(2);
				}
			
			if(arg.equals("RedChan")) { 
				ifBLotRedChan=!ifBLotRedChan;
				if(ifBLotRedChan) 
					redBlot.setBackground(Color.red);
				else 
					redBlot.setBackground(Color.gray);
				invalidateRasterCaches(0);
				reDraw();
				}

			if(arg.equals("GreenChan")) { 
				ifBLotGreenChan=!ifBLotGreenChan;
				if(ifBLotGreenChan) 
					greenBlot.setBackground(Color.green);
				else 
					greenBlot.setBackground(Color.gray);
				invalidateRasterCaches(1);
				reDraw();
				}

			if(arg.equals("BlueChan")) { 
				ifBLotBlueChan=!ifBLotBlueChan;
				if(ifBLotBlueChan) 
					blueBlot.setBackground(Color.blue);
				else 
					blueBlot.setBackground(Color.gray);
				invalidateRasterCaches(2);
				reDraw();
				}


    }
   
   private void invalidateRasterCaches(int ifRGB) {
   		picJPanel.invalidateRasterCaches(ifRGB);
   		}
   
   private void invalidateRasterCaches() {
   		if(ifBLotRedChan)    invalidateRasterCaches(0);
   		if(ifBLotGreenChan)  invalidateRasterCaches(1);
   		if(ifBLotBlueChan)   invalidateRasterCaches(2);
   		}
   
   private void reDraw() {
       picJPanel.reDraw();
    }
   
	public void changeJPanelSize(double ratio) {
		int pwidth = getSize().width;
		int pheight = getSize().height;
		pwidth = (int)(ratio*pwidth +.4999);
		pheight = (int)(ratio*pheight +.4999);
		setSize(pwidth,pheight);
		show();
		}


	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}

class SphericalHole {
    int [] absloc = {0, 0, 0};
    int radius = 8;
    BlackHoles parent;
    boolean isVisible=false;
    SphericalHole prior=null;
    SphericalHole next=null;
    boolean ifLinkedListRoot=false; // the root BlackHoles does nothing
						// but cannot be deleted, does not draw;
    
    public SphericalHole(int [] xyz, int rad, BlackHoles parent) {
	this.parent = parent;
	radius=rad;
	absloc[0]=xyz[0];
	absloc[1]=xyz[1];
	absloc[2]=xyz[2];
	}
	
    public SphericalHole(BlackHoles parent) { // create the linked list root
	this.parent = parent;
	radius=2;
	absloc[0]=0;
	absloc[1]=0;
	absloc[2]=0;
	prior=null;
	next=null;
	ifLinkedListRoot=true;
	}
	
    public void moveTo(int [] xyz) {
	absloc[0]=xyz[0]; absloc[1]=xyz[1]; absloc[2]=xyz[2];
	}
	
	public boolean blotOutData(BioradStackConfig BSC, byte [] ras, int  ncols, int nrows, 
				int sectNum, double pixPeruM, byte blotColor) {
		double delz = (BSC.getZHeightFromSectNum(sectNum) - BSC.getZHeightFromSectNum(absloc[2]))*pixPeruM;
		double rhosq = (radius*radius - delz*delz);
		int offset,x,y,dx,dy;
		if(rhosq<=0) return false;
		int rsq = (int)(rhosq + .4999);
		int rho = (int)(Math.sqrt(rhosq) + .4999);
		int xlo = absloc[0]-rho; if(xlo<0) xlo=0; if(xlo >= ncols) xlo = ncols-1;
		int xhi = absloc[0]+rho; if(xhi<0) xhi=0; if(xhi >= ncols) xhi = ncols-1;
		int ylo = absloc[1]-rho; if(ylo<0) ylo=0; if(ylo >= nrows) ylo = nrows-1;
		int yhi = absloc[1]+rho; if(yhi<0) yhi=0; if(yhi >= nrows) yhi = nrows-1;
		for(y=ylo;y<=yhi;y++) {
			offset = y*ncols;
			for(x=xlo;x<=xhi;x++) {
				dx=x-absloc[0];	dy = y-absloc[1];
				if(dx*dx + dy*dy <rsq)
					ras[offset + x] = blotColor;
				}
			}
		return true;
		}
	
    public double distanceTo(int [] xyz) {
	double dd = xyz[0] - absloc[0];
	double dist = dd*dd;
	dd = xyz[1] - absloc[1];
	dist += (dd*dd);
	dd = xyz[2] - absloc[2];
	dist += (dd*dd);
	return(Math.sqrt(dist));
	}
		
    public void grow() {
	++radius;
	talk("radius grows to "+radius);
	}
	
     public void shrink() {
	if(--radius <1) radius=1;;
	talk("radius shrinks to "+radius);
	}
	
    public boolean isListRoot() {
	return ifLinkedListRoot;
	}

    public boolean isFirst() {
	if(ifLinkedListRoot || prior.isListRoot()) return true;
	else return false;
	}

    public boolean isLast() {
	if(next==null) return true;
	else return false;
	}

    public SphericalHole excise() {
	if(isListRoot()) return this;
/* before...
holeA.next-->  this.next----->
holeA<----\    this<------\   holeB
           \---this.prior  \--holeB.prior


after...
holeA.next-->  holeB.next----->
holeA<----\    holeB   
           \---holeB.prior 

 */
	if(!isLast()) {
	    prior.next = next;	   
	 	next.prior = prior;
	 	}
	else 
	    prior.next=null;
	next=null;
	if(!prior.isLast()) 
	    return prior.next;
	else
	    return prior; // there is always a prior (possibly the List root)
	}

    public SphericalHole insertNewSphericalHole(int [] xyz, int rad,  
		BlackHoles parent) {
	SphericalHole newhole = new SphericalHole(xyz, rad, parent);
/* before...
holeA.next-->  this.next----->
holeA<----\    this<------\   holeB
           \---this.prior  \--holeB.prior

after...
holeA.next-->  this.next-->  newhole.next----->
holeA<---\     this<-----\   newhole<------\    holeB        
          \----this.prior \--newhole.prior  \---holeB.prior

 */
	newhole.next = this.next;
	if(!isLast()) next.prior=newhole;
	newhole.prior=this;
	this.next=newhole;
	talk("New hole added...");
	newhole.tellStats();
	return newhole;
	}
   
    public SphericalHole previousCross() {
	if(prior.ifLinkedListRoot) return this;
	return prior;
	}
	
    public SphericalHole nextCross() {
	if(next==null) return this;
	return next;
	}
	
    public void setRadius(int newrad) {
	radius = newrad;
	}
	
    public int tellRadius() {
	return(radius);
	}

    public int [] tellAbsLocation() {
	int [] xyz = {absloc[0], absloc[1], absloc[2]};
	return xyz;
	}

    public  void tellStats() {
	if(isListRoot()) {
		talk("I'm not a hole; i'm the list root");
		return;
		}
	talk("hole of radius "+radius+" is at ("+absloc[0]+",  "+absloc[1]+",  "+absloc[2]+")");
	}
	
    public void draw(Graphics g, int xorg, int yorg) {
      if(!ifLinkedListRoot && isVisible) {
	int dia = (int)(2.*radius);
	if(dia<1) dia=1;
	int x = (absloc[0]-radius+xorg);
	int y = (absloc[1]-radius+yorg);
	g.drawOval(x, y, dia,  dia);
	}
      }
     
    public void appear() {
	isVisible=true;
	}
      
    public void disappear() {
	isVisible=false;
	}
      
    public void highlight(Graphics g, int xorg, int yorg) { // by drawing a tiny dot...
      if(!ifLinkedListRoot && isVisible) {
		int x = absloc[0]+xorg;
		int y = absloc[1]+yorg;
		g.drawLine(x-10, y, x+10, y);
		g.drawLine(x, y+10, x, y-10);
		}
     }

	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}
