package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.lang.*;
import java.util.*;
import java.awt.*;  
import javax.swing.*;


class Boundary {	
	int XMIN=0,XMAX=768,YMIN=0,YMAX=512; // the max raster size
			//  the bounds of the current histogram.
	Color mycolor = Color.white;
	ShowPicJPanel picJPanel=null;
	Stats statsJPanel=null;
	HistogramBox parent;
	int xcnt=384,ycnt=256;
	int boxShape=1;
	int [] polyxcoords=null;
	int [] polyycoords=null;
	int [] RGBdims = new int[3];
	int nverts=0;
	Polygon myPolygon=null;
	int draggingVertexNumber = -1;
	boolean ifOuterBoundary=false;
	double halfwidth=1;	
	double halfheight=1;
	double halfwidthheightsq;	
	boolean ifMonochrome=false;

	public void takeNewRGBdims(int [] rgbdims) {
		for(int i=0;i<3;i++) 
			RGBdims[i]=rgbdims[i];
		}
		
   	public void swapBoxesAndDims(Boundary mate) {
   		Boundary in,out;
   		int i;
     	if(ifOuterBoundary) { in=mate; out=this;}
     	else { out=mate; in=this;}
     	if(in.ifOuterBoundary == out.ifOuterBoundary) {
     		talk("OOPS inner/outer conflict cannot posibly occur.");
     		return;
			}  
		for(i=0;i<3;i++) in.RGBdims[i] = out.RGBdims[i]; 
		out.XMAX = in.XMAX;
		out.XMIN = in.XMIN;
		out.YMIN = in.YMIN;
     	}
     	
	public void sepuku() {
		polyxcoords=null;
		polyycoords=null;
		}
		
	public void readFromDisk(ArchiveMarks am) {
	  	double badval = -7777.777;
	  	int ibool,i;
			XMIN = (int)(am.getNextFileDouble(badval));
			XMAX = (int)(am.getNextFileDouble(badval));
			YMIN = (int)(am.getNextFileDouble(badval));
			YMAX = (int)(am.getNextFileDouble(badval));

			Color mycolor = Color.white;

		xcnt = (int)(am.getNextFileDouble(badval));
		ycnt = (int)(am.getNextFileDouble(badval));
		boxShape = (int)(am.getNextFileDouble(badval));
 		ibool = (int)(am.getNextFileDouble(badval)); ifOuterBoundary = ibool!=0? true : false;

	  	if(ifOuterBoundary) { // an outer boundary needs to get the appropriate XMIN,XMAX,YMIN
	  		RGBdims[0]=XMIN; // from its companion inner boundary
	  		RGBdims[1]=XMAX; // and the companion inner boundary needs to get its RGBdims
	  		RGBdims[2]=YMIN; // from its outer boundarty companion
	  		}

		halfwidth = am.getNextFileDouble(badval);
		halfheight = am.getNextFileDouble(badval);
		halfwidthheightsq = am.getNextFileDouble(badval);
		nverts = (int)(am.getNextFileDouble(badval));
		polyxcoords=new int[nverts];
		polyycoords=new int[nverts];
		//talk("boundary reading in old poly coord list, nverts="+nverts);
  		for(i=0;i<nverts;i++)
   			polyxcoords[i] = (int)(am.getNextFileDouble(badval));
  		for(i=0;i<nverts;i++)
   			polyycoords[i] = (int)(am.getNextFileDouble(badval));
		myPolygon = new Polygon(polyxcoords,polyycoords,nverts);	
		draggingVertexNumber = -1;
       	parent.bndChg("Boundary.readFromDisk ");
   	}
   	
 	public void writeToDisk(ArchiveMarks am, String stg) {
 		int i;
	  	if(ifOuterBoundary) {
   			am.writeNum(RGBdims[0],("    Rdim  <---"+stg));
   			am.writeNum(RGBdims[1],"    Gdim");
   			am.writeNum(RGBdims[2],"    Bdim"); 
   			}
   		else {
   			am.writeNum(XMIN,("    XMIN  <---"+stg));
   			am.writeNum(XMAX,("    XMAX"));
   			am.writeNum(YMIN,("    YMIN")); 
   			}
   		am.writeNum(YMAX,"    YMAX");
   		am.writeNum(xcnt,"    xcnt");
   		am.writeNum(ycnt,"    ycnt");
   		am.writeNum(boxShape,"    boxShape");
   		am.writeNum(ifOuterBoundary,"    ifOuterBoundary");
   		am.writeNum(halfwidth,"    halfwidth");
   		am.writeNum(halfheight,"    halfheight");
   		am.writeNum(halfwidthheightsq,"    halfwidthheightsq");
   		am.writeNum(nverts,"    nverts");
   		for(i=0;i<nverts;i++)
   			am.writeNum(polyxcoords[i],("   polyxcoords["+i+"]"));
  		for(i=0;i<nverts;i++)
   			am.writeNum(polyycoords[i],("   polyycoords["+i+"]"));
 		}
 		
 	public void writeCoreBoundaryToDisk(ArchiveMarks am, String stg) {
 		int i;
   		am.writeNum(xcnt,"    =xcenter for"+stg);
   		am.writeNum(ycnt,"    =ycenter for"+stg);
   		if(nverts==5) {
   			am.writeNum(boxShape,"    =boxShape (1 ==> rectangle)");
   			am.writeNum(2*halfwidth,"    =width");
   			am.writeNum(2*halfheight,"   =height");
   			}
   		else {
   			am.writeNum(boxShape,"    =boxShape = general polygon for "+stg);
   			am.writeNum(nverts,"    =nverts");
   			for(i=0;i<nverts;i++)
   				am.writeNum(polyxcoords[i],("   polyxcoords["+i+"]"));
  			for(i=0;i<nverts;i++)
   				am.writeNum(polyycoords[i],("   polyycoords["+i+"]"));
   			}
 		}
	
	public void transferDataTo(Boundary abnd) {
		int j;
		abnd.XMIN=XMIN; abnd.XMAX=XMAX; abnd.YMIN=YMIN; abnd.YMAX=YMAX;
		abnd.mycolor = mycolor;
		abnd.picJPanel=picJPanel;
		abnd.statsJPanel=statsJPanel;
		abnd.parent=parent;
		abnd.boxShape=boxShape;
		if(abnd.nverts != nverts) {
			abnd.nverts=nverts;
			abnd.polyxcoords = new int[nverts];
			abnd.polyycoords = new int[nverts];
			}
		for(j=0;j<nverts;j++) {
			abnd.polyxcoords[j] = polyxcoords[j];
			abnd.polyycoords[j] = polyycoords[j];
			}
			
		for(j=0;j<3;j++)
			abnd.RGBdims[j] = RGBdims[j];
			
		abnd.draggingVertexNumber = draggingVertexNumber;
		abnd.ifOuterBoundary=ifOuterBoundary;
		abnd.getBounds();
		}	
		
	public void blendDataFrom(double a1,Boundary b1,double a2,Boundary b2) {
		int j;
		if(b1.boxShape != b2.boxShape || b1.nverts != b2.nverts) { 
			b2.transferDataTo(this);
			return;	
			}
		XMIN=(int)(a1*b1.XMIN+a2*b2.XMIN +.49);
		XMAX=(int)(a1*b1.XMAX+a2*b2.XMAX +.49);
		YMIN=(int)(a1*b1.YMIN+a2*b2.YMIN +.49);
		YMAX=(int)(a1*b1.YMAX+a2*b2.YMAX +.49);
		mycolor = b1.mycolor;
		picJPanel=b1.picJPanel;
		statsJPanel=b1.statsJPanel;
		parent=b1.parent;
		if(nverts!=b1.nverts) {
			nverts=b1.nverts;
			polyxcoords = new int[nverts];
			polyycoords = new int[nverts];
			}
		for(j=0;j<nverts;j++) {
			polyxcoords[j] = (int)(a1*b1.polyxcoords[j] + a2*b2.polyxcoords[j]+.499);
			polyycoords[j] = (int)(a1*b1.polyycoords[j] + a2*b2.polyycoords[j]);
			}
			
		for(j=0;j<3;j++)
			RGBdims[j] = (int)(a1*b1.RGBdims[j] + a2*b2.RGBdims[j]);
			
		myPolygon = new Polygon(polyxcoords,polyycoords,nverts);	
		draggingVertexNumber = b1.draggingVertexNumber;
		ifOuterBoundary=b1.ifOuterBoundary;
		//talk("blendDataFrom moves xcenter from "+xcnt+" to "+a1+" * "+b1.xcnt+" + "+a2+" * "+b2.xcnt);
		getBounds();
		//talk("which is "+xcnt);
		return; // meaning that the boundary remains interpolatable...
		}


	public Boundary(ShowPicJPanel picJPanel, Stats statsJPanel, HistogramBox parent, 
					boolean ifOuterBoundary) {
		this.parent = parent;
		this.picJPanel=picJPanel;
		this.statsJPanel=statsJPanel;
		XMAX=picJPanel.fullWidth;
		YMAX=picJPanel.fullHeight;
		XMIN=picJPanel.XMIN;
		YMIN=picJPanel.YMIN;
		if(ifOuterBoundary) {
			makeRect(384-128,255-128,384+128,255+128, "Outer boundary constructor");
			this.ifOuterBoundary=true;
			}
		else {
			makeRect(384-50,255-50,384+50,255+50, "Inner boundary constructor");
			this.ifOuterBoundary=false;
			}					
		}
	
	public void reshapeBox(int shape) {
		boxShape=shape;
		if(boxShape==1 || boxShape==2)  // it's basically a rectangle...
			squareUp(-1);
		getBounds();
       	parent.bndChg("Boundary.reshapeBox ");
		statsJPanel.computeArea(); // to report histogram box area.
		}

		
	public void rescale(double enlargeRatio) { // boost or shrink size of the box...
		int [] cg = getCenterOfGravity();
		double ex,ey,norm,scalar;
		int i;
		for(i=0;i<nverts;i++) {
			ex = polyxcoords[i]-cg[0];
			ey = polyycoords[i]-cg[1]; 
			norm = Math.sqrt(.001+ex*ex+ey*ey);
			scalar = enlargeRatio*norm;
			ex = ex/norm;
			ey = ey/norm;// (ex,ey) is a unit vector from cg to vertex i
			polyxcoords[i] = (int)(.499+cg[0]+scalar*ex);
			polyycoords[i] = (int)(.499+cg[1]+scalar*ey);
			}
		getBounds();
		if(enlargeRatio != 1) {
       		parent.bndChg("Boundary.rescale "+enlargeRatio);
			statsJPanel.computeArea(); // to report histogram box area.
			}
		}
			
			
	private void makeRect(int xmin,int ymin, int xmax, int ymax, String stg) {
		nverts=5;
		polyxcoords=new int[nverts];
		polyycoords=new int[nverts];
		polyxcoords[0] = xmin;
		polyxcoords[1] = xmax;
		polyxcoords[2] = xmax;
		polyxcoords[3] = xmin;
		polyxcoords[4] = xmin;
		polyycoords[0] = ymin;
		polyycoords[1] = ymin;
		polyycoords[2] = ymax;
		polyycoords[3] = ymax;	
		polyycoords[4] = ymin;
		myPolygon = new Polygon(polyxcoords,polyycoords,nverts);	
       	parent.bndChg(("Boundary.makeRect from( "+stg));
		// for use in computing ellipsooid inside/outside function
		}

	private void squareUp(int vert) {
		if(vert<0) { // use bounding box...
			Rectangle bigRect = myPolygon.getBounds();
			makeRect(bigRect.x,bigRect.y,bigRect.x+bigRect.width,
				bigRect.y+bigRect.height," squareUP");
			}
		else  {
			switch (vert) {
				case 0 : 
					polyxcoords[4]=polyxcoords[vert];
					polyycoords[4]=polyycoords[vert];
					polyxcoords[3]=polyxcoords[vert];
					polyycoords[1]=polyycoords[vert];
					break;
				case 1:
					polyxcoords[2] = polyxcoords[vert];
					polyycoords[0] = polyycoords[vert];
					polyycoords[4] = polyycoords[vert];
					break;
				case 2:
					polyxcoords[1] = polyxcoords[vert];
					polyycoords[3] = polyycoords[vert];
					break;
				case 3:
					polyxcoords[4] = polyxcoords[vert];
					polyxcoords[0] = polyxcoords[vert];
					polyycoords[2] = polyycoords[vert];
					break;
				}					
			}
       	parent.bndChg("Boundary.squareUp ");
		}
		
	public int [] getCenterOfGravity() {
		int [] cg = {0,0};
		Rectangle bigRect = myPolygon.getBounds();
		cg[0] = bigRect.x + (1 + bigRect.width)/2;
		cg[1] = bigRect.y + (1 + bigRect.height)/2;
		return cg;
		}

	public void moveCenterTo(int x, int y) {
		int [] cg = getCenterOfGravity();
		int dx = x - cg[0];
		int dy = y - cg[1];
		int i;
		for(i=0;i<nverts;i++) {
			polyxcoords[i] += dx;
			polyycoords[i] += dy;
			}
		getBounds();
       	parent.bndChg("Boundary.moveCenterTo ");
		}
	
	public void draw(Graphics g, int xorg, int yorg) { // draw the x-y outline of the boundary
		int i, xfrom,yfrom,xto=polyxcoords[0], 
				yto=polyycoords[0];
		if(ifMonochrome) g.setColor(Color.black);
		else g.setColor(mycolor);
		switch (boxShape) {
			case 1:  // a rectangle
			case 3:  // a polygon
				for(i=1;i<nverts;i++) {
					xfrom=xto;		yfrom=yto;
					xto=polyxcoords[i]; 
					yto=polyycoords[i];
					g.drawLine(xorg+xfrom,yorg+yfrom,xorg+xto,yorg+yto);
					}
				break;
			case 2:  // ellipsoid
				g.drawOval(xorg+polyxcoords[0],
					yorg+polyycoords[0],
					(polyxcoords[2]-polyxcoords[0]),
					(polyycoords[2]-polyycoords[0]));
				break;
				}				
		}
		
	public void setupToDefineNewPolygon() {
		polyxcoords = new int[100];
		polyycoords = new int[100];
		nverts=0;
		myPolygon=null;
		}
	
	public int [] getBounds() {
		int [] xybounds=new int[4];
		myPolygon = new Polygon(polyxcoords,polyycoords,nverts);	
		Rectangle bx=myPolygon.getBounds();
		halfwidth = (bx.width+1)/2;	
		halfheight = (bx.height+1)/2;
		xcnt = (int)(bx.x+halfwidth);
		ycnt = (int)(bx.y+halfheight);
		halfwidthheightsq = halfwidth*halfwidth*halfheight*halfheight;	
		xybounds[0] = bx.x;
		xybounds[1] = bx.y;
		xybounds[2] = bx.x + bx.width;
		xybounds[3] = bx.y + bx.height;
		return xybounds;
		}
		
	public boolean isInBoundary(int x, int y) {
		double delx,dely;
		if(boxShape != 2) return (myPolygon.inside(x,y));	
			delx = halfheight*(x-xcnt);
			dely = halfwidth*(y-ycnt);
			if(delx*delx + dely*dely >= halfwidthheightsq)  
				return false;	
			else return true;
		}

	public boolean nextVertexOfPolygon(int x,int y) {
			polyxcoords[nverts]=x;
			polyycoords[nverts]=y;
			++nverts;
			if(nverts==1) { // it's the first vertex...
	    		//talk("first vertex defined");
				return true;
				}
			int dx=polyxcoords[0] - polyxcoords[nverts-1];
			int dy=polyycoords[0] - polyycoords[nverts-1];
			if(nverts>2 && (dx*dx + dy*dy)<=16 || (boxShape==1 && nverts==5) || nverts>=99) { 
						// close it
				polyxcoords[nverts-1]=polyxcoords[0];
				polyycoords[nverts-1]=polyycoords[0];
				//talk("OK the polygon is defined");
				getBounds();
       			parent.bndChg("Boundary.nextVertexOfPolygon ");
				return false;
				}
			talk("click on next vertex...");
			return true;
	    }
			
	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
		
	
	public void moveVertex(int vertnumber, int x, int y) {
			polyxcoords[vertnumber]=x;
			polyycoords[vertnumber]=y;
			if(vertnumber==nverts-1) {
				polyxcoords[0]=x;
				polyycoords[0]=y;
				}
			if(vertnumber==0) {
				polyxcoords[nverts-1]=x;
				polyycoords[nverts-1]=y;
				}			
			if(boxShape==1 || boxShape==2) squareUp(vertnumber);
       			parent.bndChg("Boundary.moveVertex3 ");
			getBounds();
			}
		
	public void primeToDragAVertex() {
		draggingVertexNumber=-1;
		}	
			
	public void moveVertex(int x, int y) { 
		if(draggingVertexNumber<0) 
			draggingVertexNumber=closestVertex(x,y);
		moveVertex(draggingVertexNumber,x,y);
			}
		
	private int distSquared(int x,int y,int xx, int yy) {
		int dx=x-xx;
		int dy=y-yy;
		return(dx*dx + dy*dy);
		}	
	
	private int closestVertex(int x, int y) {
		int dsq,ibest=0,i;
		int distsq=distSquared(polyxcoords[ibest],polyycoords[ibest],x,y);
		for(i=1;i<nverts;i++) {
			dsq=distSquared(polyxcoords[i],polyycoords[i],x,y);
			if(dsq<distsq) {
				ibest=i;
				distsq=dsq;	
				}
			}
		return ibest;
		}
}


	