package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.lang.*;
import java.util.*;
import java.awt.*;  
import javax.swing.*;

class ColorSeparator extends JFrame implements java.awt.event.ActionListener,java.awt.event.WindowListener {
	ShowPicJPanel picJPanel;
	PickColors colorSource;
	BioradCanvas drawCanvas;
	int nomwidth=420, nomheight=135;
	int [] redColors = {255,0,0};
	int [] greenColors = {0,255,0};
	int [] blueColors = {0,0,255};
	int [] redLUT;
	int [] greenLUT;
	int [] blueLUT;
	static int alphamask = (255 << 24);
	static int redmask = (255 << 16);
	static int greenmask = (255 << 8);
	static int bluemask = 255;
	int [] redpix;	 
	int [] greenpix;	
	int [] archivePicRas;
	int [] intPicRas;
	JTextField baseName,redTag,greenTag,blueTag,
		redGreenTag,redBlueTag,greenBlueTag,redGreenBlueTag;
	String pathFileName,pathName,shortFileName;

    public ColorSeparator(ShowPicJPanel picJPanel, PickColors colorSource, BioradCanvas drawCanvas) {  
        super("write colored separations to disk files");        
		this.picJPanel = picJPanel;
		this.colorSource = colorSource;
		this.drawCanvas = drawCanvas;
		redColors = colorSource.redColors;
		greenColors = colorSource.greenColors;
		blueColors = colorSource.blueColors;
		WidgetMaker wmk = new WidgetMaker(this);

		
		Color labelColor = new Color(160,80,40);
		Color leftColor = new Color(192,96,48);
    	setSize(nomwidth, nomheight);
		setBackground(Color.lightGray);
        getContentPane().setLayout(new GridLayout(1,5));    // NEED to specify a layout manager!!
        JPanel MainJButtons = new JPanel();
        MainJButtons.setLayout(new GridLayout(0,1));
        MainJButtons.setBackground(Color.lightGray);
        baseName = wmk.makeAJTextField("",24);
        baseName.setBackground(Color.white);
        MainJButtons.add(wmk.makeAJLabel("name's 1-st part is",leftColor));
        MainJButtons.add(baseName);
        MainJButtons.add(wmk.makeAJLabel("shared by all files",leftColor));
        MainJButtons.add(wmk.makeAJLabel("                 ",leftColor));
        MainJButtons.add(wmk.makeAJButton("Get tiff image", Color.yellow));
        MainJButtons.add(wmk.makeAJLabel("      from disk file",Color.yellow));
        MainJButtons.add(wmk.makeAJLabel("             ",leftColor));
        MainJButtons.add(wmk.makeAJLabel("             ",leftColor));
        MainJButtons.add(wmk.makeAJButton("Write ALL tiffs to disk", new Color(0,200,120)));
        getContentPane().add(MainJButtons);
        
        JPanel tagNames = new JPanel();
        tagNames.setLayout(new GridLayout(0,1));
        tagNames.setBackground(Color.lightGray);
        tagNames.add(wmk.makeAJLabel("name's 2-nd part",labelColor));
        tagNames.add(wmk.makeAJLabel("(blank-->inactive)",labelColor));
  		redTag = wmk.makeAJTextField("",20);    redTag.setBackground(Color.white);   tagNames.add(redTag);
  		greenTag = wmk.makeAJTextField("",20);   greenTag.setBackground(Color.white);   tagNames.add(greenTag);
		blueTag = wmk.makeAJTextField("",20);   blueTag.setBackground(Color.white);   tagNames.add(blueTag);
		redGreenTag = wmk.makeAJTextField("",20);    redGreenTag.setBackground(Color.white);  tagNames.add(redGreenTag);
		redBlueTag = wmk.makeAJTextField("",20);    redBlueTag.setBackground(Color.white);  tagNames.add(redBlueTag);
		greenBlueTag = wmk.makeAJTextField("",20);  greenBlueTag.setBackground(Color.white);  tagNames.add(greenBlueTag);
		redGreenBlueTag  = wmk.makeAJTextField("",20); redGreenBlueTag.setBackground(Color.white); tagNames.add(redGreenBlueTag); 
		getContentPane().add(tagNames);
		
        JPanel tagJLabels = new JPanel();
        tagJLabels.setLayout(new GridLayout(0,1));
        tagJLabels.setBackground(Color.lightGray);
        tagJLabels.add(wmk.makeAJLabel("   Combinations",labelColor));
        tagJLabels.add(wmk.makeAJLabel("       Available",labelColor));
		tagJLabels.add(wmk.makeAJLabel("red chanel")); 
		tagJLabels.add(wmk.makeAJLabel("green chanel")); 
		tagJLabels.add(wmk.makeAJLabel("blue channel")); 
		tagJLabels.add(wmk.makeAJLabel("red+green chans")); 
		tagJLabels.add(wmk.makeAJLabel("red+blue chans")); 
		tagJLabels.add(wmk.makeAJLabel("green+blue chans")); 
		tagJLabels.add(wmk.makeAJLabel("all 3 channels")); 
		getContentPane().add(tagJLabels);

        JPanel displayJButtons = new JPanel();
        displayJButtons.setLayout(new GridLayout(0,1));
        displayJButtons.setBackground(Color.lightGray);
        displayJButtons.add(wmk.makeAJLabel("push buttons to",labelColor));
        displayJButtons.add(wmk.makeAJLabel("display pictures",labelColor));
		displayJButtons.add(wmk.makeAJButton("red chan",Color.red));
		displayJButtons.add(wmk.makeAJButton("green chan",Color.green));
		displayJButtons.add(wmk.makeAJButton("blue chan",Color.blue));
		displayJButtons.add(wmk.makeAJButton("red+green",Color.yellow));
		displayJButtons.add(wmk.makeAJButton("red+blue",Color.magenta));
		displayJButtons.add(wmk.makeAJButton("green+blue",Color.cyan));
		displayJButtons.add(wmk.makeAJButton("r + g + b",Color.white));
		getContentPane().add(displayJButtons);

        JPanel writeTiffJButtons = new JPanel();
        writeTiffJButtons.setLayout(new GridLayout(0,1));
        writeTiffJButtons.setBackground(Color.lightGray);
        writeTiffJButtons.add(wmk.makeAJLabel("push buttons to",labelColor));
        writeTiffJButtons.add(wmk.makeAJLabel("write tiff files",labelColor));
		writeTiffJButtons.add(wmk.makeAJButton("Red chan",Color.red));
		writeTiffJButtons.add(wmk.makeAJButton("Green chan",Color.green));
		writeTiffJButtons.add(wmk.makeAJButton("Blue chan",Color.blue));
		writeTiffJButtons.add(wmk.makeAJButton("Red+Green",Color.yellow));
		writeTiffJButtons.add(wmk.makeAJButton("Red+Blue",Color.magenta));
		writeTiffJButtons.add(wmk.makeAJButton("Green+Blue",Color.cyan));
		writeTiffJButtons.add(wmk.makeAJButton("R + G + B",Color.white));
		getContentPane().add(writeTiffJButtons);
		addWindowListener(this);
        }

		
	private void writeOrDisplayTiffs(boolean ifWrite, boolean ifred, boolean ifgreen, boolean ifblue, boolean ifredgreen, boolean ifredblue,
		boolean ifgreenblue, boolean ifredgreenblue) {
		int npix = picJPanel.fullWidth*picJPanel.fullHeight;
		String canvasTitle;
		
		// get the current colors
		if(picJPanel.curWidth != picJPanel.fullWidth || picJPanel.curHeight != picJPanel.fullHeight) {
			talk("You can use this panel only on full-sizeed biorad images");
			return;
			}
		
		// see which secondPartNames are non-blank... 
		String part1name = baseName.getText();
		if(part1name.length() ==0) {
			talk("You must enter a non-blank first part of the file(s) name in order to");
			talk("write any tiff images to disk.");
			return;
			}
			
		String redTagName = redTag.getText(); 
		String greenTagName = greenTag.getText();
		String blueTagName = blueTag.getText();
		String redgreenTagName = redGreenTag.getText();
		String redblueTagName = redBlueTag.getText();
		String greenblueTagName = greenBlueTag.getText();
		String redgreenblueTagName = redGreenBlueTag.getText();
		ifred = ifred && (redTagName.length() > 0);
		ifgreen = ifgreen && (greenTagName.length() > 0);
		ifblue = ifblue && (blueTagName.length() > 0);
		ifredgreen = ifredgreen && (redgreenTagName.length() > 0);
		ifredblue = ifredblue && (redblueTagName.length() > 0);
		ifgreenblue = ifgreenblue && (greenblueTagName.length() > 0);
		ifredgreenblue = ifredgreenblue && (redgreenblueTagName.length() > 0);
		if(!ifWrite) { // it's a request to display only one combo.
			canvasTitle = "blend of ";
			boolean ifUseRed = ifred || ifredgreen || ifredblue || ifredgreenblue;
			boolean ifUseGreen = ifgreen || ifredgreen || ifgreenblue || ifredgreenblue;
			boolean ifUseBlue = ifblue || ifredblue || ifgreenblue || ifredgreenblue;
			blendChanels(ifUseRed,ifUseGreen,ifUseBlue);
			if(ifUseRed) canvasTitle += "R ";
			if(ifUseGreen) canvasTitle += "G ";
			if(ifUseBlue) canvasTitle += "B ";
			canvasTitle += "chans of "+pathFileName;
			drawCanvas.showThisRaster(intPicRas, canvasTitle);
			return;
			}
			
		else { // write one or more files... 
			TiffFuncs fileWriter = new TiffFuncs();

			if(ifred) {
				blendChanels(true,false,false);
				canvasTitle = (pathName+part1name+redTagName+".tiff");
				intPicRas=drawCanvas.showThisRaster(intPicRas, canvasTitle);
				fileWriter.writeColor768Tiff(canvasTitle, intPicRas);
				}				
			if(ifgreen) {
				blendChanels(false,true,false);
				canvasTitle = (pathName+part1name+greenTagName+".tiff");
				intPicRas=drawCanvas.showThisRaster(intPicRas, canvasTitle);
				fileWriter.writeColor768Tiff(canvasTitle, intPicRas);
				}
			if(ifblue) {
				blendChanels(false,false,true);
				canvasTitle = (pathName+part1name+blueTagName+".tiff");
				intPicRas=drawCanvas.showThisRaster(intPicRas, canvasTitle);
				fileWriter.writeColor768Tiff(canvasTitle, intPicRas);
				}

			if(ifredgreen) {
				canvasTitle = (pathName+part1name+redgreenTagName+".tiff");
				intPicRas=drawCanvas.showThisRaster(intPicRas, canvasTitle);
				blendChanels(true,true,false);
				fileWriter.writeColor768Tiff(canvasTitle, intPicRas);
				}
			if(ifredblue) {
				blendChanels(true,false,true);
				canvasTitle = (pathName+part1name+redblueTagName+".tiff");
				intPicRas=drawCanvas.showThisRaster(intPicRas, canvasTitle);
				fileWriter.writeColor768Tiff(canvasTitle, intPicRas);
				}
			if(ifgreenblue) {
				blendChanels(false,true,true);
				canvasTitle = (pathName+part1name+greenblueTagName+".tiff");
				intPicRas=drawCanvas.showThisRaster(intPicRas, canvasTitle);
				fileWriter.writeColor768Tiff(canvasTitle, intPicRas);
				}
			if(ifredgreenblue) {
				blendChanels(true,true,true);
				canvasTitle = (pathName+part1name+redgreenblueTagName+".tiff");
				intPicRas=drawCanvas.showThisRaster(intPicRas, canvasTitle);
				fileWriter.writeColor768Tiff(canvasTitle, intPicRas);
				}
			}
		}

	private void writeTiffFile(String stg) {
		talk("Would write "+stg);
		}			

// catch button events...
	public void actionPerformed(java.awt.event.ActionEvent event) {
		String arg = event.getActionCommand();
		int npix = picJPanel.fullWidth*picJPanel.fullHeight;

		talk(arg);
		if(arg.equals("Write ALL tiffs to disk")) {
			writeOrDisplayTiffs(true,true,true,true,true,true,true,true);
			return;
			}
					
		if(arg.equals("Get tiff image")) { // get the file
			pathFileName=null;
			pathName=null;
			GMORasFuncs fileGetter = new GMORasFuncs();
			try {intPicRas=fileGetter.readGmoRgbRas();}
			catch(IOException e) {
				talk("failed to get a valid file.");
				intPicRas=null;
				return;
				}
			if(intPicRas != null) {
				talk("image successfully read in... ");
				pathFileName = fileGetter.mostRecentInputFileName;
				pathName = fileGetter.mostRecentInputPathName;
				shortFileName = purgeTrailingTiff(fileGetter.mostRecentFileNameShort);
				baseName.setText(shortFileName);
				if(archivePicRas==null) archivePicRas = new int[npix];
				for(int i=0;i<npix;i++) // copy the just-read-raster to the archival archivePicRas[]
					archivePicRas[i] = intPicRas[i];
				drawCanvas.showThisRaster(intPicRas, pathFileName);
				}
			fileGetter=null;
			return;
			}
		if(arg.equals("red chan")) { // display
			writeOrDisplayTiffs(false, true, false, false, false, false, false, false);
			return;
			}
		if(arg.equals("green chan")) {// display
			writeOrDisplayTiffs(false, false, true, false, false, false, false, false);
			return;
			}
		if(arg.equals("blue chan")) {// display
			writeOrDisplayTiffs(false, false, false, true, false, false, false, false);
			return;
			}
		if(arg.equals("red+green")) {// display
			writeOrDisplayTiffs(false, false, false, false, true, false, false, false);
			return;
			}
		if(arg.equals("red+blue")) {// display
			writeOrDisplayTiffs(false, false, false, false, false, true, false, false);
			return;
			}
		if(arg.equals("green+blue")) {// display
			writeOrDisplayTiffs(false, false, false, false, false, false, true, false);
			return;
			}
		if(arg.equals("r + g + b")) {// display
			writeOrDisplayTiffs(false, false, false, false, false, false, false, true);
			return;
			}
			
			
		if(arg.equals("Red chan")) { // write
			writeOrDisplayTiffs(true, true, false, false, false, false, false, false);
			return;
			}
		if(arg.equals("Green chan")) {// write
			writeOrDisplayTiffs(true, false, true, false, false, false, false, false);
			return;
			}
		if(arg.equals("Blue chan")) {// write
			writeOrDisplayTiffs(true, false, false, true, false, false, false, false);
			return;
			}
		if(arg.equals("Red+Green")) {// write
			writeOrDisplayTiffs(true, false, false, false, true, false, false, false);
			return;
			}
		if(arg.equals("Red+Blue")) {// write
			writeOrDisplayTiffs(true, false, false, false, false, true, false, false);
			return;
			}
		if(arg.equals("Green+Blue")) {// write
			writeOrDisplayTiffs(true, false, false, false, false, false, true, false);
			return;
			}
		if(arg.equals("R + G + B")) {// write
			writeOrDisplayTiffs(true, false, false, false, false, false, false, true);
			return;
			}
		}
	
	private String purgeTrailingTiff(String stg) {
		int i;
		if((i=stg.indexOf(".tiff")) >0 ) {
			return stg.substring(0,i);
			}			
		if(stg.indexOf(".TIFF")>0) {
			return stg.substring(0,i);
			}		
		return stg;
		}
		
		
	private void blendChanels(boolean ifRed, boolean ifGreen, boolean ifBlue) {
		int i;
		int npix = picJPanel.fullWidth*picJPanel.fullHeight;
		int mask = alphamask;
		if(ifRed) mask = mask | redmask;
		if(ifGreen) mask = mask | greenmask;
		if(ifBlue) mask = mask | bluemask;
		for(i=0;i<npix;i++) 			
			intPicRas[i] = archivePicRas[i] & mask;				
		}
		
// catch WindowListener events  
	public void windowActivated(java.awt.event.WindowEvent event) {
		// talk("window activated");
		}
	public void windowDeactivated(java.awt.event.WindowEvent event) {
		//  talk("window Deactivated");
		}
	public void windowOpened(java.awt.event.WindowEvent event) {}
	public void windowIconified(java.awt.event.WindowEvent event) {}
	public void windowDeiconified(java.awt.event.WindowEvent event) {}
	public void windowClosing(java.awt.event.WindowEvent event) {
		talk("Color separations panel discards all buffers to save memory");
		intPicRas=null;
		archivePicRas=null;
		baseName.setText("");
		pathFileName="";
		pathName="";
		shortFileName="";
	    setVisible(false);
		}
		
	public void windowClosed(java.awt.event.WindowEvent event) {
		}


static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}	 
