package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  
import javax.swing.*;

class DimEdgesFltr {
	SegmenterNode nd=null;
	short [] work=null;
	int workrows=-1, workcols=-1;
	
	public DimEdgesFltr(SegmenterNode nd) {
		this.nd=nd;
		}
	
	private void makeWork(int incols, int inrows) {
		if(work==null || incols != workrows || inrows != workrows) {
			workrows=inrows;
			workcols=incols;
			work = new short[inrows*incols];
			}
		}
	
	public void tellWhatYouDo() {
		nd.talk("");
		nd.talk("The DimEdgesFltr has a single raster as input and two parameter, 'Thresh' & iter8'.");
		nd.talk("For each pixel, the output is zero if the input pixel is < Thresh.");
		nd.talk("When the input pixel >= Thresh, then the output brightness is proportional to the number");
		nd.talk("neighbors that pixel has with values exceeding Thresh, accoding to the following recipe:");
		nd.talk("#aboveThreshNeighbors->  0    1    2    3    4    5    6    7    8");
		nd.talk("output brightness->      0    32   64   96  128  160  192  224  255");
		nd.talk("");
		nd.talk("The DimEdgesFltr performs iter8 successive iterations of this process.");
		}
		
/* replace every pixel that has brightness>=thresh with a brightness proportional to how many neighbors it has */

public boolean doit(short [] inras, int incols, int inrows, int threshint, int iter8, short [] outras) {
		int i,row, col, npix = incols*inrows,rcoff,j,neigh;
		int [] offsets = {-1-incols, -incols, -incols+1,
		                     -1,                 1,
		                    incols-1,  incols,  incols+1};
		short [] bright = {0,32,64,96,128,160,192,224,255};
		short on=(short)255, off = (short)0, thresh = (short)(threshint);
		if(nd.diags) nd.talk("DimEdgesFltr with threshint="+threshint+" and iter8="+iter8);
		makeWork(incols,inrows);
		for(i=0;i<npix;i++)  // store input array in the work array
			work[i] = inras[i];	
		while(iter8>0) {	
		  for(row=1;row<inrows-1;row++)
    		for(col=1;col<incols-1;col++) {
				rcoff = row*incols + col;
				if(work[rcoff]>0) {
					j=0; neigh=0;
	    			while(j<8) {
						if(work[rcoff+offsets[j]] >= thresh) 
							++neigh;
		    			++j;
		    			}
		    		if(work[rcoff] >= thresh) outras[rcoff] = bright[neigh];
		    		else outras[rcoff] = 0; // this does the first thresholding
		    		}
				else
	    			outras[rcoff]=0;
	    		}
		    /* copy border pixels */
    		for(i=0; i<incols; i++) 
    			outras[i]= inras[i]>thresh? inras[i] : 0;          // top row
    		for(i=npix-incols; i<npix; i++) 
    			outras[i] =inras[i]>thresh? inras[i] : 0;         // bottom row
    		for(row=1; row<inrows; row++) {
				rcoff = row*incols;
				outras[rcoff]  =  inras[rcoff]>thresh? inras[rcoff] : 0;
				outras[rcoff-1] = inras[rcoff-1]>thresh? inras[rcoff-1] : 0;
				}
			--iter8;
			if(iter8>0) 
				for(i=0;i<npix;i++)
					work[i]=outras[i];
			}
			
	return true;
	}

}