package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  
import javax.swing.*;

class ErodeFltr {
	SegmenterNode nd=null;
	short [] work=null;
	int workrows=-1, workcols=-1;
	
	public ErodeFltr(SegmenterNode nd) {
		this.nd=nd;
		}
	
	private void makeWork(int incols, int inrows) {
		if(work==null || incols != workrows || inrows != workrows) {
			workrows=inrows;
			workcols=incols;
			work = new short[inrows*incols];
			}
		}
		
	public void tellWhatYouDo() {
		nd.talk("");
		nd.talk("The ErodeFltr has a single input raster and two parameters, 'cutoff' and 'iter8'.");
		nd.talk("It performs iter8 successive iterations of the following process which eats away");
		nd.talk("pixels having at least one empty neighbor. First, and only on the first iteration,");
		nd.talk("every input pixel < cutoff is set to zero while every input pixel >= cutoff is set");
		nd.talk("to 255. Then, in each iteration, any 'on' pixel having an 'off' neighbor is turned");
		nd.talk("off.");
		nd.talk("This means, for example, any isolated 'round' spot with a radius < iter8 will be erased.");
		nd.talk("Any meandering stripe with a width < 2*iter8 will be erased. ");
		}
		
/* every pixel >threshold gets set to 255, and every pixel < threshold gets set to 0. 
Then iter8 erosion passes are done, in each of which any pixel with
a 0 neighbors gets set to 0. */

public boolean doit(short [] inras, int incols, int inrows, int threshint, int iter8,  short [] outras) {
		int i,row, col, npix = incols*inrows,rcoff,j;
		boolean ifTurnedOff=false;
		int nturnedoff=0;
		int [] offsets = {-1-incols, -incols, -incols+1,
		                     -1,                 1,
		                    incols-1,  incols,  incols+1};
		short on=(short)255, off = (short)0, thresh = (short)(threshint);
		if(nd.diags) nd.talk("ErodeFltr with threshint="+threshint+" and iter8="+iter8);
		makeWork(incols,inrows);
		for(i=0;i<npix;i++) { // blackwhite the input array
			if(inras[i] >= thresh)
				work[i] = on;	
			else
				work[i] = off;
			}	
			
		if(iter8<0) 
			iter8=(-iter8);
		int npass=iter8;
		while(iter8>0) {
		  for(row=1;row<inrows-1;row++)
    		for(col=1;col<incols-1;col++) {
				rcoff = row*incols + col;
				if(work[rcoff]>0) {
					ifTurnedOff=false;
					j=0;
	    			while(j<8 && !ifTurnedOff) {
						if(work[rcoff+offsets[j]] == 0) {
		    				outras[rcoff]=0;
		    				ifTurnedOff=true;
							++nturnedoff;
		    				}
		    			++j;
		    			}
		    		if(j>7) outras[rcoff]=work[rcoff];
		    		}
				else
	    			outras[rcoff]=0;
	    		}
		    /* copy border pixels */
    		for(i=0; i<incols; i++) 
    			outras[i]= work[i];          // top row
    		for(i=npix-incols; i<npix; i++) 
    			outras[i] = work[i];         // bottom row
    		for(row=1; row<inrows; row++) {
				rcoff = row*incols;
				outras[rcoff]  =  work[rcoff];
				outras[rcoff-1] = work[rcoff-1];
				}
			--iter8;
			if(iter8>0) // copy outras to work
				for(i=0; i<npix; i++) 
					work[i] = outras[i];
			}
	return true;
	}

}