package staxx;  

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.lang.*;
import java.util.*;
import java.awt.*;  
import javax.swing.*;

class HistogramBox { // implements the inner and outer boundaries of the histogram region
	// of interest. It uses no z--coordinate. 
	ShowPicJPanel picJPanel=null;
	Stats statsJPanel=null;
	int numLowPassIterations=0;
	boolean [] ifCoreBoxPinned=null;
	boolean ifCoreBoxesWander=false;
	Boundary innerBoundary,outerBoundary;
	Boundary [] innerCoreBoundaryStack=null;
	Boundary [] outerCoreBoundaryStack=null;
	TransectLine myTransect;
	int numCoreSects=1;
	int atSect=0,Rthresh=0,Gthresh=0,Bthresh=0;
	boolean makingHistograms=true; // when false we bypass the voxPointers computation.
	
	int XMIN=0,XMAX=10,YMIN=0,YMAX=10;
	int xmin=0,xmax=1,ymin=0,ymax=1;  // boundaries that contain both inner and
		 // outer boundaries... 
	
	int xnext=-1, ynext=-1;
		// used to return next voxel by nextVox() method
	boolean ifPastLast = true;
	boolean ifBoxIsVisible = true;
	boolean ifInnerBoxValid = true;
	boolean ifTransectLineVisible=true;
	int nowDefiningWhichPolygon = 0;    // 1 means now defining inner polygon
										// 2 means now defining outer polygon
										// 0 means not defining anything;
	Color mycolor = Color.white;
	int [] xcoords=null; // will hold the array of x coords inside the histogram box
	int [] ycoords=null; // will hold the array of ycoords inside the histogram box
	int NPixInHistobox = 0; // will hold the number of points inside the histogram box
	int nextVox=0;
	boolean ifBoundaryChanged=true;
	int outerBoxShape=1;
	int innerBoxShape=1;
	int dragVertexOfWhat =  -1; // == -1 means dragging no vertex
								// == 2 means dragging vertex of outer box
								// == 1 means dragging vertex of inner box

	final static int MOUSE_REPORTS_LOCATION_COLORS = 12;
	final static int MOUSE_MOVES_OUTERBOX_CENTER = 23;
	final static int MOUSE_MOVES_INNERBOX_CENTER = 26;
	final static int MOUSE_CLICKS_DEFINE_POLYGON = 24;	
	final static int MOUSE_DRAGS_POLYGON_VERTEX = 25;	
	final static int MOUSE_CLICKS_DEFINE_TRANSECT = 28;	
	final static int MOUSE_DRAGS_TRANSECT_VERTEX = 29;	
	final static int MOUSE_MOVES_TRANSECT_CENTER = 27;
	boolean ifMonochrome=false;
/*  table of which pre-scan sections correspond to which finely-spaced section.
 pre-scan-sect  .2uM spaced sect
      0.000     28.000
      1.000     33.000
      2.000     38.000
      3.000     43.000
      4.000     48.000
      5.000     53.000
      6.000     58.000
      7.000     63.000
      8.000     68.000
      9.000     73.000
     10.000     78.000
     11.000     83.000
     12.000     88.000
     13.000     93.000
     14.000     98.000
     15.000    103.000
     16.000    108.000
     17.000    113.000
     18.000    118.000
     19.000    123.000
     20.000    128.000
     21.000    133.000
     22.000    138.000
     23.000    143.000
     24.000    148.000
     25.000    153.000
     26.000    158.000
     27.000    163.000   */

	public void sepuku() {
		talk("HistogramBox.sepuku()");
		if(innerBoundary != null)  innerBoundary.sepuku();
		if(outerBoundary != null)  outerBoundary.sepuku();
		innerBoundary=null;
		outerBoundary=null;
		destroyCoreBoundaryStacks();
		xcoords=null;
		ycoords=null;
		picJPanel=null;
		statsJPanel=null;
		if(myTransect != null) myTransect.sepuku();
		myTransect=null;
		}
	
	public void upDateRGBdims() {
		if(!ifCoreBoxesWander) return;
		Rthresh=outerBoundary.RGBdims[0];
		Gthresh=outerBoundary.RGBdims[1];
		Bthresh=outerBoundary.RGBdims[2];
		statsJPanel.setRGBdimText(outerBoundary.RGBdims);
		}
			
	public void takeNewRGBdims(int [] rgbdims) {
		if(innerBoundary != null) innerBoundary.takeNewRGBdims(rgbdims);
		if(outerBoundary != null) outerBoundary.takeNewRGBdims(rgbdims);
		}
	
	public void destroyCoreBoundaryStacks() {
	ifCoreBoxPinned=null;
	ifCoreBoxesWander=false;
	numCoreSects=0;
	if(outerCoreBoundaryStack!=null && innerCoreBoundaryStack!=null)
	  for(int i=0;i<numCoreSects;i++) { // initially every section's boundaries are  a copies of innerBoundary and outerBoundary
		if(outerCoreBoundaryStack[i] != null) outerCoreBoundaryStack[i].sepuku();
		outerCoreBoundaryStack[i]=null;
		if(innerCoreBoundaryStack[i] != null) innerCoreBoundaryStack[i].sepuku();
		innerCoreBoundaryStack[i]=null;
		}
	innerCoreBoundaryStack=null;
	outerCoreBoundaryStack=null;
	}
		
	public void finalize() {
		talk("                                                                           HistogramBox snuffs");
		}
		
	public void makeNewCoreBoundaryStacks(int nsects) {
	numCoreSects=nsects;
	ifCoreBoxPinned=new boolean[numCoreSects];
	ifCoreBoxesWander=true;
	innerCoreBoundaryStack=new Boundary[numCoreSects];
	outerCoreBoundaryStack=new Boundary[numCoreSects];
	for(int i=0;i<numCoreSects;i++) { // initially every section's boundaries are copies of innerBoundary and outerBoundary
		outerCoreBoundaryStack[i]=new Boundary(picJPanel, statsJPanel, this, true);
		outerBoundary.transferDataTo(outerCoreBoundaryStack[i]);
		innerCoreBoundaryStack[i]=new Boundary(picJPanel, statsJPanel, this, false);
		innerBoundary.transferDataTo(innerCoreBoundaryStack[i]);
		ifCoreBoxPinned[i]=false;
		}
	ifCoreBoxPinned[0]=true;
	ifCoreBoxPinned[nsects-1]=true;
	}

	public int identifyPinnedSectionNumber(int iwh) {
		int i,j;
		if(!ifCoreBoxesWander || numCoreSects<=1)
			return(-1);
		if(iwh <0) {// return number of nearest previous pinned section
			if(atSect==0) return 0;
			i=atSect-1;
			while(i>0 && !ifCoreBoxPinned[i]) --i;
			return i;
			}
		if(iwh > 0) {// return number of nearest following pinned section
			if(atSect==numCoreSects-1) return atSect;
			i=atSect+1;
			while(i < (numCoreSects-1) && !ifCoreBoxPinned[i]) ++i;
			return i;
			}
		// return the nearest pinned section unless this one is pinned.
		if(ifCoreBoxPinned[atSect]) return atSect;
		// find number of nearest previous pinned section
			i=atSect-1;
			while(i>0 && !ifCoreBoxPinned[i]) --i;
		// find number of nearest following pinned section
			j=atSect+1;
			while(j < (numCoreSects-1) && !ifCoreBoxPinned[j]) ++j;
			if(j-atSect < atSect-i)
				return j;
			else return i;
		}
	
	public int takeWanderingCoreBoundaryFromSection(int isect) {
		if(picJPanel.redfile != null) // use the first non-null BioradFile we find in picJPanel
			return picJPanel.redfile.myConfig.takeWanderingCoreBoundaryFromSection(isect);
		if(picJPanel.greenfile != null) // use the first non-null BioradFile we find in picJPanel
			return picJPanel.greenfile.myConfig.takeWanderingCoreBoundaryFromSection(isect);
		if(picJPanel.bluefile != null) // use the first non-null BioradFile we find in picJPanel
			return picJPanel.bluefile.myConfig.takeWanderingCoreBoundaryFromSection(isect);
		return isect;
		}
				
	public void setSection(int isect) {
		if(!ifCoreBoxesWander) {
			atSect=isect;	
			return;
			}
		int boundsect = takeWanderingCoreBoundaryFromSection(isect);
		if(boundsect != isect) 
			talk("Section "+isect+" takes its wandering core boundary from section "+boundsect);
		if(atSect<numCoreSects && isect<numCoreSects) {
			int [] rgbdims = statsJPanel.getRGBdimValues();
			outerBoundary.takeNewRGBdims(rgbdims);
			innerBoundary.takeNewRGBdims(rgbdims);
			outerBoundary.transferDataTo(outerCoreBoundaryStack[atSect]);
			innerBoundary.transferDataTo(innerCoreBoundaryStack[atSect]);
			atSect=isect;
			ifCoreBoxPinned[0]=true; // always keep first and last section pinned
			ifCoreBoxPinned[numCoreSects-1]=true; // always keep first and last section pinned
			if(ifCoreBoxPinned[boundsect]) {
				outerCoreBoundaryStack[boundsect].transferDataTo(outerBoundary);
				innerCoreBoundaryStack[boundsect].transferDataTo(innerBoundary);
				}
			else { // interpolate...
				int pinlo=boundsect-1,pinhi=boundsect+1;
				while(!ifCoreBoxPinned[pinlo]) --pinlo;
				while(!ifCoreBoxPinned[pinhi]) ++pinhi;
				talk("interpolated core boundaries of section "+boundsect+" lies between pinned boundary sections "+pinlo+
					" and "+pinhi); 
				outerBoundary.blendDataFrom((double)(boundsect - pinlo)/(double)(pinhi-pinlo),outerCoreBoundaryStack[pinhi],
					  (double)(pinhi-boundsect)/(double)(pinhi-pinlo),outerCoreBoundaryStack[pinlo]);
					  
				innerBoundary.blendDataFrom((double)(boundsect - pinlo)/(double)(pinhi-pinlo),innerCoreBoundaryStack[pinhi],
					   (double)(pinhi-boundsect)/(double)(pinhi-pinlo),innerCoreBoundaryStack[pinlo]);
					   
				outerBoundary.transferDataTo(outerCoreBoundaryStack[atSect]);
				innerBoundary.transferDataTo(innerCoreBoundaryStack[atSect]);
				}
			statsJPanel.setRGBdimText(outerBoundary.RGBdims);
			bndChg("setSection() did it");
			}
		}
	public void setColors(boolean ifMonochrome) {
		this.ifMonochrome=ifMonochrome;
		myTransect.ifMonochrome=false;
		innerBoundary.ifMonochrome=false;
		outerBoundary.ifMonochrome=false;
		}

	public void bndChg(String stg) {
		ifBoundaryChanged = true;
		//talk(stg+"  changed the boundary");
		}
		
  	public void readFromDisk(ArchiveMarks am) {
	  	double badval = -7777.777;
	  	int ibool,i,npinned,pin,isect,XMIN,XMAX,YMIN,YMAX;
		XMIN = (int)(am.getNextFileDouble(badval));
		XMAX = (int)(am.getNextFileDouble(badval));
		YMIN = (int)(am.getNextFileDouble(badval));
		YMAX = (int)(am.getNextFileDouble(badval));
		xmin = (int)(am.getNextFileDouble(badval));
		xmax = (int)(am.getNextFileDouble(badval));
		ymin = (int)(am.getNextFileDouble(badval));
		ymax = (int)(am.getNextFileDouble(badval));
 		ibool = (int)(am.getNextFileDouble(badval)); 
 		ifBoxIsVisible = ibool!=0? true : false;
 		ibool = (int)(am.getNextFileDouble(badval)); 
 		ifInnerBoxValid = ibool!=0? true : false;
 		ibool = (int)(am.getNextFileDouble(badval)); 
 		ifTransectLineVisible = ibool!=0? true : false;

		ifPastLast = true;
		nowDefiningWhichPolygon = 0;
		mycolor = Color.white;
		xcoords=null; 
		ycoords=null; 
		NPixInHistobox = 0;
		nextVox=0;
		dragVertexOfWhat =  -1;
		
		outerBoxShape = (int)(am.getNextFileDouble(badval));
		innerBoxShape = (int)(am.getNextFileDouble(badval));
  	    innerBoundary.readFromDisk(am);
     	outerBoundary.readFromDisk(am);
     	outerBoundary.swapBoxesAndDims(innerBoundary);
     	if((i=myTransect.readFromDisk(am)) >= 0 ) { // must have failed because there is a wandering core stack
     		// Now read in the pinned core stack boundaries if they are there....
			numCoreSects=i;
  			ifCoreBoxesWander=true;
 			ifCoreBoxPinned=new boolean[numCoreSects];
 			innerCoreBoundaryStack = new Boundary[numCoreSects];
 			outerCoreBoundaryStack = new Boundary[numCoreSects];
 			for(i=0;i<numCoreSects;i++)  { 
 				ifCoreBoxPinned[i]=false;
  				innerCoreBoundaryStack[i] =new Boundary(picJPanel, statsJPanel, this, false); 
  				outerCoreBoundaryStack[i] =new Boundary(picJPanel, statsJPanel, this, true); 		
  				}		
			npinned = (int)(am.getNextFileDouble(badval));
 			for(pin=0;pin<npinned;pin++) {
 				isect = (int)(am.getNextFileDouble(badval,true));
 				talk("recover ..."+ am.mostRecentLineRead);
  				ifCoreBoxPinned[isect]=true;
				innerCoreBoundaryStack[isect].readFromDisk(am);
				}
 			for(pin=0;pin<npinned;pin++) {
 				isect = (int)(am.getNextFileDouble(badval,true));
 				talk("recover ..."+ am.mostRecentLineRead);
  				ifCoreBoxPinned[isect]=true;
				outerCoreBoundaryStack[isect].readFromDisk(am);
				}
			
			for(isect=0;isect<numCoreSects;isect++)
				if(ifCoreBoxPinned[isect])
					outerCoreBoundaryStack[isect].swapBoxesAndDims(innerCoreBoundaryStack[isect]);

     		myTransect.readFromDisk(am);
   			}
   		else {
   			if(ifCoreBoxesWander) {
   				for(i=0;i<numCoreSects;i++)  { 
  					innerCoreBoundaryStack[i].sepuku();
  					innerCoreBoundaryStack[i]=null; 
  					outerCoreBoundaryStack[i].sepuku(); 
  					outerCoreBoundaryStack[i]=null;		
  					}}
  			ifCoreBoxPinned=null;
   			ifCoreBoxesWander=false;
			}
     	bndChg("readFromDisk ");
     	}

	public void writeToDisk(ArchiveMarks am) {
		int i,npinned;
		XMIN=picJPanel.XMIN;
		XMAX=picJPanel.fullWidth;
		YMIN=picJPanel.YMIN;
		YMAX=picJPanel.fullHeight;
  		am.writeNum(XMIN,("    XMIN  <---HistogramBox"));
   		am.writeNum(XMAX,("    XMAX"));
   		am.writeNum(YMIN,("    YMIN"));
   		am.writeNum(YMAX,("    YMAX"));
  		am.writeNum(xmin,("    xmin"));
   		am.writeNum(xmax,("    xmax"));
   		am.writeNum(ymin,("    ymin"));
   		am.writeNum(ymax,("    ymax"));
   		am.writeNum(ifBoxIsVisible,"    ifBoxIsVisible");
   		am.writeNum(ifInnerBoxValid,"    ifInnerBoxValid");
   		am.writeNum(ifTransectLineVisible,"    ifTransectLineVisible");
     	am.writeNum(outerBoxShape,"    outerBoxShape");
   		am.writeNum(innerBoxShape,"    innerBoxShape");
      	innerBoundary.writeToDisk(am,"inner current boundary");
     	outerBoundary.writeToDisk(am,"outer current boundary");
		if(ifCoreBoxesWander && innerCoreBoundaryStack!= null && outerCoreBoundaryStack!=null &&
				ifCoreBoxPinned != null && numCoreSects>1) {
			am.writeNum(numCoreSects," = numCoreSects in a wandering core stack");
			npinned=0;
			for(i=0;i<numCoreSects;i++)
				if(ifCoreBoxPinned[i]) ++npinned;
			am.writeNum(npinned," = number of pinned sections in the stack");			
			String stg = ("core sect # of a pinned inner boundary in a stack of "+ numCoreSects);			 
			for(i=0;i<numCoreSects;i++) // write to disk each of the pinned inner boundaries.
				if(ifCoreBoxPinned[i]) { // it's a pinned section boundary
					am.writeNum(i,stg);
					innerCoreBoundaryStack[i].writeToDisk(am,("inner pinned core sect # "+i));
					}
			stg = ("core sect # of a pinned outer boundary in a stack of "+ numCoreSects);			 
			for(i=0;i<numCoreSects;i++) // write to disk each of the pinned outer boundaries.
				if(ifCoreBoxPinned[i]) { // it's a pinned section boundary
					am.writeNum(i,stg);
					outerCoreBoundaryStack[i].writeToDisk(am,("outerPinnedCoreSect # "+i));
					}
			}
     	myTransect.writeToDisk(am, "transectLine");
		}
	
	public void writeCoreBoxToDisk(ArchiveMarks am) {
   		am.writeAString(":::::::::::::::::::");
     	am.writeNum(outerBoxShape,"    =outerBoxShape code (1 ==> rectangle)");
  		am.writeNum(xmin,("    =xmin for outer box boundary"));
   		am.writeNum(xmax,("    =xmax"));
   		am.writeNum(ymin,("    =ymin"));
   		am.writeNum(ymax,("    =ymax"));
   		outerBoundary.writeCoreBoundaryToDisk(am," Outer box boundary");
    	am.writeAString(":::::::::::::::::::");
  		if(ifInnerBoxValid) {
   			am.writeAString("*******************");
   			am.writeAString("  The inner boundary is active; annulus between outer and inner boundary scored ");
   			innerBoundary.writeCoreBoundaryToDisk(am," inner box boundary");
   			am.writeAString("*******************");
   			}
		}
	
	public void transferDataTo(HistogramBox abox) { 
		int j;
		// copies all current data into aBox
		abox.xmin=xmin; abox.xmax=xmax; abox.ymin=ymin; abox.ymax=ymax;  
		abox.xnext=xnext;  abox.ynext=ynext;
		abox.ifPastLast = ifPastLast;
		abox.ifBoxIsVisible = ifBoxIsVisible;
		abox.ifInnerBoxValid = ifInnerBoxValid;
		abox.ifTransectLineVisible=ifTransectLineVisible;
		abox.nowDefiningWhichPolygon = nowDefiningWhichPolygon;    
		abox.mycolor = Color.white;
		abox.picJPanel=picJPanel;
		abox.statsJPanel=statsJPanel;
		abox.NPixInHistobox = NPixInHistobox; 
		abox.xcoords=new int[NPixInHistobox];
		abox.ycoords=new int[NPixInHistobox]; 
		for(j=0;j<NPixInHistobox;j++) {
			abox.xcoords[j] = xcoords[j];
			abox.ycoords[j] = ycoords[j];
			}
		abox.nextVox=nextVox;
		abox.innerBoundary=new Boundary(picJPanel, statsJPanel, this, false); 		
		abox.outerBoundary=new Boundary(picJPanel, statsJPanel, this, true);
		abox.myTransect=new TransectLine(picJPanel, statsJPanel, this);
		abox.outerBoxShape=outerBoxShape;
		abox.innerBoxShape=innerBoxShape;
		abox.dragVertexOfWhat = dragVertexOfWhat;
		// now transfer data to the boundaries and transect line
		innerBoundary.transferDataTo(abox.innerBoundary);
		outerBoundary.transferDataTo(abox.outerBoundary);
		
		if(abox.ifCoreBoxesWander && abox.innerCoreBoundaryStack!= null && abox.outerCoreBoundaryStack!=null){
			for(int i=0;i<abox.numCoreSects;i++) { 
				abox.outerCoreBoundaryStack[i].sepuku();
				outerCoreBoundaryStack[i]=null;
				abox.innerCoreBoundaryStack[i].sepuku();
				innerCoreBoundaryStack[i] = null;
				}
			ifCoreBoxPinned=null;
			}
		
		if(ifCoreBoxesWander && innerCoreBoundaryStack!= null && outerCoreBoundaryStack!=null &&
				ifCoreBoxPinned != null) { // make and transfer data to the 
			abox.numCoreSects=numCoreSects;
			abox.atSect=atSect;
			abox.ifCoreBoxPinned=new boolean[numCoreSects];
			abox.ifCoreBoxesWander=true;
			abox.innerCoreBoundaryStack=new Boundary[numCoreSects];
			abox.outerCoreBoundaryStack=new Boundary[numCoreSects];
			for(int i=0;i<numCoreSects;i++) { // initially every section's boundaries are copies of innerBoundary and outerBoundary
				abox.outerCoreBoundaryStack[i]=new Boundary(picJPanel, statsJPanel, this, true);
				outerCoreBoundaryStack[i].transferDataTo(abox.outerCoreBoundaryStack[i]);
				abox.innerCoreBoundaryStack[i]=new Boundary(picJPanel, statsJPanel, this, false);
				innerCoreBoundaryStack[i].transferDataTo(abox.innerCoreBoundaryStack[i]);
				abox.ifCoreBoxPinned[i]=ifCoreBoxPinned[i];
				}
			abox.ifCoreBoxPinned[0]=true;
			abox.ifCoreBoxPinned[numCoreSects-1]=true;
			}
		myTransect.transferDataTo(abox.myTransect);
	}
	
	public HistogramBox(ShowPicJPanel picJPanel,Stats statsJPanel) {
		this.picJPanel=picJPanel;
		this.statsJPanel = statsJPanel;
		setRasterLimits(0,2048,0,2048);
		innerBoundary = new Boundary(picJPanel, statsJPanel, this, false);
		outerBoundary = new Boundary(picJPanel, statsJPanel, this, true);
		myTransect= new TransectLine(picJPanel, statsJPanel, this);
     	ifBoundaryChanged=true;
		}
	
	public void rescaleBoundary(boolean ifOuter, double enlargeRatio) {
		if(ifOuter) {
			outerBoundary.rescale(enlargeRatio);
			}
		else if(ifInnerBoxValid) {
			innerBoundary.rescale(enlargeRatio);
			}
     	bndChg("HistogramBox.rescaleBoundary ");
		}

	public void newOuterPolygon() {
		nowDefiningWhichPolygon=2;
		outerBoundary.setupToDefineNewPolygon();
		}
		
	public void newTransectLine() {
		myTransect.setupToDefineNewTransect();
		}
			
	public void newInnerPolygon() {
		nowDefiningWhichPolygon=1;
		innerBoundary.setupToDefineNewPolygon();
		}
		
	public void toggleInnerBoxValidity() {
		ifInnerBoxValid = !ifInnerBoxValid;
		if(ifInnerBoxValid) 
			talk("inner box is on");
		else talk("inner box off");
		bndChg("HistogramBox.toggleInnerBoxValidity ");
		}
		
	public void primeToDragVertex(int mouseTask) {
		switch (mouseTask) {
			case 3: // drag outer polygon vertex
				outerBoundary.primeToDragAVertex();
				dragVertexOfWhat=2;
				break;
			case 2:  // drag inner polygon vertex
				innerBoundary.primeToDragAVertex();
				dragVertexOfWhat=1;
				break;
			case 7: // drag transect line vertex
				myTransect.primeToDragAVertex();
				dragVertexOfWhat=3;
				break;
			}
		}

	public void moveVertex(int x, int y, int doWhat) {
		if(doWhat==MOUSE_DRAGS_POLYGON_VERTEX) {
			if(dragVertexOfWhat==2) 
				outerBoundary.moveVertex(x, y);
			else if(ifInnerBoxValid && dragVertexOfWhat==1) 
				innerBoundary.moveVertex(x, y);
			}
		if(doWhat==MOUSE_DRAGS_TRANSECT_VERTEX)
			myTransect.moveVertex(x,y);
		}

	public void reshapeBox(boolean ifOuter, int boxShape) {
		if(ifOuter) outerBoundary.reshapeBox(boxShape);
		else innerBoundary.reshapeBox(boxShape);
		bndChg("HistogramBox.reshapeBox ");
		}
						
	public void setParams(int width, int blur, double pixPerMic) {
		if(pixPerMic<4) {
			talk("*********histogramBox.setParams  warning!!! pixelsPerMicrometer= "+pixPerMic);
			pixPerMic=4;
			}
		myTransect.setParams(width,blur,pixPerMic);
		}
						
	public boolean nextVertexOfPolygon(double [] mouseLoc, int doWhat) {
		if(doWhat == MOUSE_CLICKS_DEFINE_POLYGON) {
			int x=(int)(mouseLoc[0]+.499);
			int y=(int)(mouseLoc[1]+.499);
			if(nowDefiningWhichPolygon==1) 
				return(innerBoundary.nextVertexOfPolygon(x,y));
			if(nowDefiningWhichPolygon==2) 
				return(outerBoundary.nextVertexOfPolygon(x,y));
			}
		if(doWhat == MOUSE_CLICKS_DEFINE_TRANSECT) {
			int x=(int)(mouseLoc[0]+.499);
			int y=(int)(mouseLoc[1]+.499);
			return(myTransect.nextVertexOfTransect(x,y));
			}
		return false;
		}

	public long getNvox() {
		return NPixInHistobox;
		}
	
	public void setRasterLimits(int minx,int maxx,int miny,int maxy) {
		XMIN = minx<0? 0 : minx;
		XMAX = maxx > picJPanel.fullWidth? picJPanel.fullWidth : maxx;
		YMIN = miny<0? 0 : miny;
		YMAX = maxy>picJPanel.fullHeight? picJPanel.fullHeight : maxy;
		}
									
	public void resetVoxCycle() {
		nextVox=0;
		}
		
	public boolean voxCycle(int [] xy) {
		if(nextVox >= NPixInHistobox)
			return false;
		xy[0]=xcoords[nextVox];
		xy[1]=ycoords[nextVox];
		++nextVox;
		return true;
		}

	public int resetVoxPointers() { /* initializes nextVox() method to sweep through
				multi-section cube */
		if(!ifBoundaryChanged) {
			talk("bypassing resetVoxPointers()");
			return NPixInHistobox;
			}
		int [] xy = {0,0};
		ifPastLast=false;
		talk("resetVoxPointers()");
		int [] xybounds = outerBoundary.getBounds();
		if(ifInnerBoxValid) {
			int [] xyinbounds = innerBoundary.getBounds();
			if(xybounds[0] > xyinbounds[0]) xybounds[0]=xyinbounds[0] ;
			if(xybounds[1] > xyinbounds[1]) xybounds[1]=xyinbounds[1] ;
			if(xybounds[2] < xyinbounds[2]) xybounds[2]=xyinbounds[2] ;
			if(xybounds[3] < xyinbounds[3]) xybounds[3]=xyinbounds[3] ;
			}
		if((xmin=xybounds[0]) < XMIN) xmin=XMIN; 
		if((xmax=xybounds[2]) >= XMAX) xmax=XMAX-1;
		if((ymin=xybounds[1]) <YMIN) ymin=YMIN;
		if((ymax=xybounds[3]) >= YMAX) ymax=YMAX-1;
		if(makingHistograms) {
			xnext = xmin-1; // so that the first increment to xnext produces 0 value
			ynext = ymin;
			NPixInHistobox = (1+xmax-xmin)*(1+ymax-ymin);
			xcoords = new int[NPixInHistobox];
			ycoords = new int[NPixInHistobox];
			NPixInHistobox = 0;
			while(nextVoxPointer(xy)) {
				xcoords[NPixInHistobox] = xy[0];
				ycoords[NPixInHistobox] = xy[1];
				++NPixInHistobox;
				}
			}
		resetVoxCycle();
		statsJPanel.areaPixels.setText(""+NPixInHistobox);
		myTransect.rescale(1.0);
		statsJPanel.transectPixels.setText(""+myTransect.transectLength);
		ifBoundaryChanged=false;
		return NPixInHistobox;
		}
	
	public boolean nextVoxPointer(int [] nextCoords) { /* returns in nextCoords 
								 the coordinates of the next voxel 
								in the histogram volume and returns true if
								there is a next valid voxel.
								returns  false to indicatethat there 
							    is no next voxel.
							     Usee resetNextVox() to reset count.    */
	long delx,dely;
	if(ifPastLast) return false;
	boolean isVoxelValid = false;
	while(!isVoxelValid) {
	  // cycle through x first, then y
	  if(++xnext >= xmax) {
		xnext=xmin; // reset x-counter to beginning of next row and increment y-counter
		if(++ynext >= ymax) { // we just swept through the whole raster...
			ifPastLast=true;
			return false;
			}
		}
		isVoxelValid= (outerBoundary.isInBoundary(xnext,ynext) && (!ifInnerBoxValid ||
				!innerBoundary.isInBoundary(xnext,ynext)));
	  }
	// when fall through, we have a valid voxel whose coordinates we return...
	nextCoords[0] = xnext;
	nextCoords[1] = ynext;
	return true;
	}

	public void draw(Graphics g, int xorg, int yorg) { // draw the x-y outline of the volume
		if(ifTransectLineVisible) myTransect.draw(g,  xorg,  yorg);
		if(!ifBoxIsVisible) return;
		outerBoundary.draw(g,  xorg,  yorg);
		if(ifInnerBoxValid)  {
			innerBoundary.draw(g,  xorg,  yorg);
			}
		}
			
	private void solidFill(Graphics g) { // this is for testing...
		g.setColor(Color.blue);
		int [] xy={0,0,0};
		int xto,yto,skip=10;
		talk("solidFill");
		resetVoxCycle();
		if(!voxCycle(xy)) return;
		while(voxCycle(xy)) {
			if(--skip <0) {
				xto=xy[0];
				yto=xy[1];
				g.drawLine(xto,yto,xto,yto);
				skip=10;
				}
			}
		}
		
	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}