package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;

class  HolePatcher {
	SegmenterNode nd=null;
	static final int [] pows2 = {128,64,32,16,8,4,2,1};
	int rasColDim=-1;
	int [] neighOffs = new int[8];
	short initOff=7,erasedOff=1,filledOn=255,initOn=220,deepOff=5;
	short [] workras = null;
	int workpix=-1; // number of pixels in workras array

//       0 or 2 => no change to center pixel, 1 => erase center pixel,
//            3 => fill central pixel
static final int [] changes = {
2, 0, 0, 0, 0, 0, 0, 1, 0, 0, 1, 0, 3, 0, 0, 0, 0, 3, 1, 0, 0, 0, 0, 0, 3, 3, 2, 2, 3, 0, 2, 2, 
0, 0, 3, 0, 3, 0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 2, 3, 0, 0, 0, 0, 0, 0, 0, 3, 0, 2, 0, 3, 0, 0, 0, 
0, 3, 3, 3, 3, 0, 3, 0, 1, 0, 2, 2, 0, 0, 2, 0, 1, 0, 2, 2, 0, 0, 2, 0, 2, 2, 2, 2, 2, 0, 2, 2, 
0, 0, 3, 0, 0, 0, 3, 0, 0, 0, 2, 2, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 2, 0, 2, 2, 0, 0, 0, 2, 
0, 3, 3, 0, 0, 0, 0, 0, 3, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 2, 3, 3, 2, 0, 0, 0, 0, 0, 
0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
0, 0, 3, 3, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 2, 0, 2, 0, 2, 0, 0, 0, 2, 2, 
1, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 2, 0, 2, 2, 0, 0, 2, 2};
	
	
	public HolePatcher(SegmenterNode nd) {
		this.nd=nd;
		} 
		
	public void tellWhatYouDo() {
		nd.talk("");
		nd.talk("The gapFill filter has one parameters, iter8, which tells how many");
		nd.talk("successive iterations to perform.");
		nd.talk("It fills in pixels which appear to be holes in lines according to");
		nd.talk("the following recipe: index the 8 neighbors of a typical pixel like this");
		nd.talk("6	7	8      If the central pixel is off, then we turn it on (to 255)");
		nd.talk("5		1      if it has exactly any of these two neighbors 'on':");
		nd.talk("4	3	2      1&5, 2&6, 3&7, 4&8, 1&4, 2&5, 3&6, 4&7, 5&8, 1&6, 2&7, 3&8");
		nd.talk("We also turn on the central pixel if has exactly three neighbors on,");
		nd.talk("EXCEPT the following three neighbors patterns:");
		nd.talk("1&2&8, 1&2&3, 1&7&8, 2&3&4, 3&4&5, 4&5&6, 5&6&7, 6&7&8.");
		nd.talk("");
		nd.talk("A pixel with 0, 1, or more than three neighbors on is left alone.");
		}
		



/* scheme to patch particular holes in the middle of a 3x3 grid. Number the neighbors like this:

			6	7	8
			5		1
			4	3	2
Fill the center previously empty hole when these sets of just two neighbors are on:
1&5, 2&6, 3&7, 4&8 (the diagonals, whose recipe is i&(i+4) with i the lowest index
1&4, 2&5, 3&6, 4&7, 5&8, 1&6, 2&7, 3&8, (the near diagonals, with recipes i&(i+3)
										or i&(i+5), with i the lowest index.
When just three neighbors are on, fill the central previously empty hole EXCEPT in 
these cases:
6&7&8, 8&1&2, 2&3&4, 4&5&6, {the box edges, with recipes i&(i+1)&(i+2) mod 8}
1&2&3, 3&4&5, 5&6&7, 7&8&1, {the box corners with the same recipe as above}
*/

	public int processOnce(int [] whichKind, short [] inras,int NCOLS,int NROWS, short[] outras) {
		int i,npix=	NCOLS*NROWS,pixelsChanged;
		for(i=0;i<npix;i++) // put current ras[] into rasOld[]
			if(inras[i]>127)
				inras[i] = initOn;
			else
				inras[i] = initOff;
		pixelsChanged=applyEightPointPatternFilter(inras,outras,NCOLS,NROWS,whichKind);
		return pixelsChanged;
		}

		
	public int applyEightPointPatternFilter(short [] fromras, short [] toras,
			int ncols, int nrows, int [] doWhat) {
		int row,col,patrn,j,i,pixoff,act,numChanged=0;
		if(rasColDim != ncols) { // rebuild offsets
			talk("new GapFill raster size with "+ncols+" columns and "+nrows+" rows.");
			rasColDim=ncols;
			neighOffs[0]=-ncols-1; neighOffs[1]=-ncols; neighOffs[2]=-ncols+1;
			neighOffs[3]=-1;                         neighOffs[4]=1;
			neighOffs[5]=ncols-1;  neighOffs[6]=ncols;  neighOffs[7]=ncols+1;
			}

		for(row=1;row<nrows-1;row++) // omit boundaries
			for(col=1;col<ncols-1;col++) {
				pixoff = row*ncols+col;
				patrn = 0;
				for(j=0;j<8;j++)
					if(fromras[pixoff+neighOffs[j]]>127)
						patrn = patrn | pows2[j];
                if((act = doWhat[patrn]) == 2 || act==0) 
                     //                    noChange     undecided
					toras[pixoff] = fromras[pixoff];
				else if(act==1) {// erase it
					if(fromras[pixoff] > 127) {
						toras[pixoff] = erasedOff;
						++numChanged;
						}
					else
						toras[pixoff] = fromras[pixoff];
					}
				else if (act==3) { // FILL IT
					if(fromras[pixoff] > 127) 
						toras[pixoff] = fromras[pixoff];	
					else {
						toras[pixoff] = filledOn;	
						++numChanged;
						}
					}
				}
		return numChanged;				
		}

public boolean doit(short [] inras,int NCOLS,int NROWS, int iter8, short[] outras) {
	int i, npix=NCOLS*NROWS,offset;
	boolean [] ngh = new boolean[14];
	boolean iffill;
	if(workpix != npix) { // make new work array
		workras = new short[npix];
		workpix=npix;
		}
	// don't change inras...
	for(i=0;i<npix;i++)
		workras[i]=inras[i];
	
	if (iter8<0)	// means keep applying filter until there are no further changes
		iter8=256;  // or until we do it 256 times
		
	while(iter8>0) {
		processOnce(changes, workras, NCOLS, NROWS, outras);
		--iter8;
		if(iter8<=0) return true;
		// do it again, but put answer back into workras
		processOnce(changes, outras, NCOLS, NROWS, workras);
		--iter8;		
		}
	for(i=0;i<npix;i++) 
		outras[i]=workras[i]; // copy workras to the output array	
	return true;
	}
	
	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
			
}