package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * Jpegger.java
 *
 * @author Created by Omnicore CodeGuide
 */

/*
	Jpegger ...  a class that writes jpegs
*/

import java.io.*;
import java.awt.*;
import java.awt.image.*;

import javax.imageio.*;
import javax.imageio.stream.*;
import java.util.Iterator;

public class Jpegger extends Canvas {


    public static BufferedImage BufferedImageFromIntRaster(int [] curIm, int width, int height) {
    	int npix = width*height;
		BufferedImage src = new BufferedImage(width, height,1);
		src.setRGB(0,0,width,height,curIm,0, width);
        src.coerceData(false);
        return src;
        }
    
	public static void jpegFromIntRaster(int w, int h, int [] curIm, String fileName) {
		// convert the curIm array into bufImage
		BufferedImage bufImage=BufferedImageFromIntRaster(curIm,w,h);
		jpegFromBufferedImage(3, bufImage, fileName);
		}
			
	public static void jpegFromRGBByteRaster(int w, int h, byte [] rgbbytes, String fileName) {
		// convert the curIm array into bufImage
		int i,joff=0,npix=w*h;
		int [] curIm = new int[npix];
		for (i=0;i<npix;i++)
			curIm[i] = rgbbytes[joff++]<<16 | rgbbytes[joff++]<<8 | rgbbytes[joff++];
		BufferedImage bufImage=BufferedImageFromIntRaster(curIm,w,h);
		jpegFromBufferedImage(3, bufImage, fileName);
		}
			
	public static void jpegFromBufferedImage(int whichHalf, BufferedImage bImg, String fileName) {
		// whichHalf==1 means left half, whichHalf==2 means right half, whichHalf==3 means both halves of original image
		
	    try {
		  // create output stream
	   	  File outFile = new File(fileName);
		  FileImageOutputStream fios = new FileImageOutputStream (outFile);
		  int origWidth = bImg.getWidth(), origHeight=bImg.getHeight();
		  int xorg = whichHalf ==2? origWidth/2 : 0;
		  int xOutWidth = whichHalf==3? origWidth : origWidth/2;
		  // conversion for colors etc ?
		  BufferedImage altBI = new BufferedImage(xOutWidth, origHeight, BufferedImage.TYPE_INT_RGB );
		  Graphics g = altBI.getGraphics();
		  g.drawImage( bImg, 0,0, xOutWidth, origHeight,   xorg, 0, xorg+xOutWidth, origHeight,  null );
		  //                   rectangle to draw to               rectangle to draw from           imageObserver
	      // Get Writer and set compression
	      Iterator iter = ImageIO.getImageWritersByFormatName( "JPG" );
	      if( iter.hasNext() ) {
	        ImageWriter writer = (ImageWriter)iter.next();
	        ImageWriteParam iwp = writer.getDefaultWriteParam();
	        iwp.setCompressionMode( ImageWriteParam.MODE_EXPLICIT );
	        iwp.setCompressionQuality( 1.0f );
	        writer.setOutput( fios );
	        IIOImage image = new IIOImage(altBI, null, null);
	        writer.write(null, image, iwp);
	        fios.close();
	        fios=null;
	        writer.dispose();
	        writer=null;
	        g.dispose();
	        g=null;
	        iwp=null;
	        altBI=null;
	      }
	    }
	    catch( Exception e )
	    {
	      e.printStackTrace();
	      System.out.println("Jpegger.jpegFromBufferedImage() failed for filename= "+fileName);
	    }
	 }
}

