package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  
import javax.swing.*;

class NiceTubes { 
// replaces ragged connected pieces by smoothed ones.
	SegmenterNode nd=null;
	
	
	public NiceTubes(SegmenterNode nd) {
		this.nd=nd;
		} 

	public void tellWhatYouDo() {
		nd.talk("The niceTubes filter filter does this...");
		}
		
public boolean doit(short [] red,int NCOLS,int NROWS,int spotthresh,double linetol, 
	double maxtubewidth,short [] outras) {
int row, col, i, npix, pararclen,
    nswarms, nspots, nparabs, parabpoints, spotpoints, swarmpoints; 
Parabola3d parabola;
double radius;
long nvoxon, nvoxoffed;
SwarmOfPoints firstSwarm=null,lastSwarm=null,aswarm,bswarm;
boolean ifinit=false;

	nswarms=0;
	nspots=0;
	nparabs=0;
	nvoxon=0;
	nvoxoffed=0;
	npix=NCOLS*NROWS;  
	short [] ras = new short[npix];
	for(i=0;i<npix;i++) {
		ras[i]=0;
		outras[i]=0;
		}

    row=0;
    col=0; 
	while(row < NROWS) {
		while(col <NCOLS) {
	    if(israspixelon(row, col, red, NCOLS, NROWS) ) { 
			    /* find a voxel that is on */ 
			aswarm = new SwarmOfPoints(red,NCOLS,NROWS,col,row,ras);
			aswarm.collectConnectedSwarm(); // transfers points in aswarm from red[] to ras[]
			if(firstSwarm==null) { // this is the first swarm found
				firstSwarm=aswarm;
				lastSwarm=aswarm;
				}
			else {
				lastSwarm.nextSwarm=aswarm;
				lastSwarm=aswarm;
				}
			}
			++col;
			}
		col=0;
		++row;
		} // when fall through, we have found all connected pieces.
		// ras[] has all the points in it, red[] has none. Restore red[] 
	for(i=0;i<npix;i++)
		red[i]=ras[i];
		
    parabola = new Parabola3d(); 
          	
	aswarm = firstSwarm; // now process the connected swarms.
	// if a swsarm can be replaced by a mathematical parabolic track, do so
	// if a swarm can be replaced by a round dot, do so
	// else just transfer it to the outras array.
	while(aswarm != null) {
		if(aswarm.getLocations() <= spotthresh) { 
          				// convert this swarm to a round dot centered 
		    ++nspots;   //  at the swarm's center of gravity
			// how big should the spot radius be ? Pi * radius^2 = numPoints  
		    radius = Math.sqrt( (double)(aswarm.numPoints)/3.14159);
		    if(radius > .66666666*(double )(aswarm.diameter)) 
		    	radius =  .66666666*(double )(aswarm.diameter);
		    spotpoints=paintspotonoutmask(aswarm.locations[0][2], 
			      aswarm.locations[1][2], radius, outras, NCOLS, NROWS); 
		    nvoxon += spotpoints;
		    nvoxoffed += aswarm.numPoints-spotpoints;
			aswarm = aswarm.dissolve();
			}
		else { // the swarm is too big to be replaced by a round spot
			if(parabola.parabolaQ(aswarm, spotthresh, linetol)) {
				// parabola is a good fit and it now stores the best 2-D coeffs 
		    	if(parabola.width > maxtubewidth) 
			    	parabola.width = maxtubewidth;
			    spotpoints = parabola.paintOnOutRas(outras, NCOLS, NROWS, ras);
				++nparabs;
		    	nvoxon += spotpoints;
		    	nvoxoffed += aswarm.numPoints-spotpoints;
				aswarm = aswarm.dissolve();
				}
			else { // just copy the swarm to outras
		    	nvoxon += aswarm.numPoints;
		    	++nswarms;
				aswarm = aswarm.migrateAndDissolve(ras, outras);
				}
			}
		}
	tellResults(nswarms, nspots, nparabs, nvoxon, nvoxoffed);
	return true;
	}

	
	private boolean israspixelon(int y, int x, short [] ras, int NCOLS, int NROWS) {
    	if(x<0 || x>=NCOLS) return false;
    	if(y<0 || y>=NROWS) return false;
    	if(ras[y*NCOLS+x]>127) return true;
    	else return false;    
		}
	
	private int paintspotonoutmask(int xctr, int yctr, double radius, short [] outras, 
			int NCOLS, int NROWS) {
		int i,j,delx,dely,radsq,rad = (int )(radius +.499);
		int npainted=0;
		if(rad < 1) rad=1;
		radsq=rad*rad;
		int xlo = xctr-rad, xhi=xctr+rad, ylo=yctr-rad, yhi=yctr+rad;
		if(xlo<0) xlo=0;
		if(xhi >= NCOLS) xhi = NCOLS-1; 
		if(ylo<0) ylo=0;
		if(yhi >= NROWS) yhi = NROWS-1;
		for(i=ylo; i <= yhi; i++)
			for(j=xlo; j <= xhi; j++) {
				 delx=j-xctr;
				 dely=i-yctr;
				 if(delx*delx + dely*dely <= radsq) {
				 	++npainted;
				 	outras[i*NCOLS + j]=255;
				 	}
				 }
		return npainted;
		}

	private void tellResults(int nswarms, int nspots, int nparabs,
				long nvoxon, long nvoxoffed) {
		nd.talk("niceTubes(): "+nswarms+" large swarms found that could not be converted....");	
		nd.talk("niceTubes(): "+nspots+" small swarms turned to round spots.");
		nd.talk("niceTubes(): "+nparabs+" swarms turned to parabolic segments.");
		nd.talk("niceTubes(): "+nvoxon+" voxels on in output mask");
		if(nvoxoffed > 0)
			nd.talk("niceTubes(): "+nvoxoffed+" voxels in input mask were eliminated");
		else 
			nd.talk("niceTubes(): "+(-nvoxoffed)+" voxels added to those on in input mask");
		}
		
}

		 		