package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;

class OneProfile extends JFrame{ //This class computes and holds single profiles
	int width=20,blur=10;
	int thresh  = 0; // all intensities less than this are ignored.
	int [] prof = null;
	int [] normalProf = null;
	int profmax=1;
	int captionX = 10;
	boolean hasProfileChanged=true;
	Color mycolor = Color.red;
	HistogramBox myBox=null;
	TransectLine myTransect;
	int npts=256;
	
	public OneProfile(HistogramBox myBox,int thresh, Color crayon, int captionX) {
		this.thresh = thresh;
		this.captionX = captionX;
		this.myBox=myBox;
		mycolor = crayon;
		myTransect = myBox.myTransect;
		npts = (int)(myTransect.transectLength); // the number of points in the 
								// horizontal axis of the profile
		talk("making OneProfile for transect length =="+npts); 
		width = myTransect.width;
		blur = myTransect.blur;
		prof = new int[npts+1];
		normalProf = new int[npts+1];
		for(int i=0;i<npts;i++) prof[npts]=1;
		hasProfileChanged=true;
		}

	public int [] getRawProf() {
		return prof;
		}
	
	public void sepuku() {
		prof=null;
		normalProf=null;
		myBox=null;
		myTransect=null;
		}
	
	public void finalize() { 
		talk("                                                                           OneProfile snuffs"); 
		}
		
	public void makeNewProfile(byte [] ras) {
		int i,pix,npix,w,b;
		int [] xy={0,0};
		long collect;
		int fullWidth=myBox.picJPanel.fullWidth;
		profmax=1;
	    for(i=0;i<npts;i++) { // i counts points along the transect.
	    	collect=0;
			npix=0;
			for(w=0;w<width;w++)
				for(b=0;b<blur;b++) 
					if(myTransect.blurCoords(i,b,w,xy)) {
						if((pix = ras[xy[1]*fullWidth+xy[0]]) <0) 
							pix += 256;
						if(pix>=thresh) {
							collect += pix;
							++npix;
							}
						}
			if(npix<=0) npix=1;
			prof[i] = (int)(collect/npix);
			
			myTransect.blurCoords(i,blur/2,width/2,xy);
/*			talk("i="+i+", npix="+npix+", collect="+collect+", profi="+prof[i]+
				", midpoiont=("+xy[0]+", "+xy[1]+")"); */
				
			if(profmax<prof[i]) profmax=prof[i]; // find the max as we go along;
			}
		hasProfileChanged=true;
	    }
	    
	public void draw(Graphics g, int howTall, double scale, String fileName, String hstName) { // draw a single profile
		int i,j;
		int x,y,xn=0, yn, capdy=15,capy;
		double tic;

		scale = 255.0; // new july 1998 to draw a single profile on an absolute scale
		if(myBox.ifMonochrome)
			g.setColor(Color.black);
		else
			g.setColor(Color.yellow);
		tic = -5*myTransect.pixelsPerMicrometer;
		for(i=0;i<myTransect.transectLength;i++) // draw the tic marks on the horizontal axis
			if(i-tic >= 5*myTransect.pixelsPerMicrometer) {
				g.drawLine(i,howTall,i,howTall+6);
				tic += 5*myTransect.pixelsPerMicrometer;
				}
		xn=(int)myTransect.transectLength;
							
		if(hasProfileChanged) 
			getNormalizedProfile();
		if(myBox.ifMonochrome)
			g.setColor(Color.black);
		else
			g.setColor(mycolor);
			
		yn=255-prof[0];
		xn=0;
		for(i=1;i<npts;i++) {
			x=xn;	y=yn; 
			xn=i;	yn=255-prof[i];
			g.drawLine(x,y,xn,yn);
			}  
		capy=howTall+capdy;
		g.drawString(("max bright="+profmax), captionX,capy);		
		capy += capdy;
		g.drawString("Transect length = ",captionX,capy);
		capy += capdy;
		g.drawString((int)(myTransect.transectLength)+" pixels = "+
			(int)(.499+myTransect.transectLength/myTransect.pixelsPerMicrometer)+" um",captionX,capy);
		capy += capdy;
		g.drawString("Avgeraging box is ", captionX,capy);
		capy += capdy;
		g.drawString(+myTransect.width+" pix wide by"+myTransect.blur, captionX,capy);
		capy += capdy;		
		g.drawString("pix along transect", captionX,capy);
		capy += capdy;
		g.drawString(("file="+fileName),captionX,capy);
		if(hstName != null) {
				capy += capdy;
				g.drawString(("hst = "+hstName),2,capy);
				}
			}

	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
		
	public int [] getNormalizedProfile() {
		int i;
		if(!hasProfileChanged) return normalProf;
		for(i=0;i<npts;i++) 
			normalProf[i] = (255*prof[i])/profmax;
		hasProfileChanged=false;
		return normalProf;	
		}	
}  
