package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;

class PainterColorPicker extends JFrame  {
	int red=128,green=128,blue=128;
	double HUE=0, CHROMA=0., WHITENESS=.5;
	/* august 2001 attempt to copy  the style of color chooser 
	used by the "painter" program. We need a circular rim  to pick the hue, h, which ranges from
	0.0 to 1.0
	h==0. is pure blue 
	360. and should not include white or black. 
	When h=hue is between  0  and 1/3, R = 255*3*h; G = 0; B = 255*(1 -3*h)
	When h=hue is between 1/3 and 2/3, R = 255*(2 -3*h)); G = 255*(3*h-1); B = 0
	When h=hue is between 2/3 and 3/3, R = 0; G = 255*(3 - 3*h); B = 255*(3*h-2)
	
	Inside the circulaar rim we need a triangle.
	*/
	static final double oneThird=1.0/3.0;
	static final double twoThirds=2.0/3.0;
	JTextField rtf,gtf,btf;
	PickColors parent1=null;
	TwoColorPicker parent2=null; //one of these will be non-null and will have a method
	// called getAllColors() which is called when the "done" button is clicked
	boolean iffirst=false;
	int nomWidth = 180;
	PainterColorPickCanvas mycanvas=null;
	
	public PainterColorPicker(int xorg, int yorg, String title,PickColors parent,
		int r, int g, int b,JTextField rtf, JTextField gtf, JTextField btf)  { 
        super(title);
        parent1=parent;
        setup( xorg,  yorg, r,  g,  b, rtf,  gtf,  btf);
        }
   
	public PainterColorPicker(int xorg, int yorg, String title,TwoColorPicker parent,
		boolean iffirst,
		int r, int g, int b,JTextField rtf, JTextField gtf, JTextField btf)  { 
        super(title);
        parent2=parent;
        this.iffirst=iffirst;
        setup( xorg,  yorg, r,  g,  b, rtf,  gtf,  btf);
        }
   
   private void setup(int xorg, int yorg,int r, int g, int b,JTextField rtf, 
   					JTextField gtf, JTextField btf)  {
		this.rtf=rtf;
		this.gtf=gtf;
		this.btf=btf;
		red = r; green=g; blue=b;  
        getContentPane().setLayout(new BorderLayout());    
		JPanel reportPanel = new JPanel();
		reportPanel.setLayout(new GridLayout(0,1));
		WidgetMaker wmk = new WidgetMaker(this);
		mycanvas = new PainterColorPickCanvas(this, r,g,b);
		getContentPane().add(mycanvas);
		setSize(nomWidth,nomWidth+25);
		setLocation(xorg,yorg);
		setVisible(true);		
		}
		
	private double [] rgbFracsFromHue(double h) {
		/* return an array of three doubles [redFrac,greenFrac,BlueFrac] given hue=h
		   between 0.0 and 1.0
			h==0. is pure blue 
		When h=hue is between  0  and 1/3, R = 3*h; G = 0; B = 1 -3*h
		When h=hue is between 1/3 and 2/3, R = 2 -3*h; G = 3*h-1; B = 0
		When h=hue is between 2/3 and 3/3, R = 0; G = 3 - 3*h; B = 3*h-2 */
		double [] rgbFracs = new double[3];
		if(h < oneThird) { // iwh==1 case (green missing)
			rgbFracs[0] = 3.0*h;              // red
			rgbFracs[1] = 0.0;                // green
			rgbFracs[2] = 1.0 - rgbFracs[0];  // blue
			maxOutFracs(rgbFracs);
			// tellFracs(h,rgbFracs);
			return rgbFracs;
			}
		if(h < twoThirds) { //  iwh==2 case (blue missing)
			rgbFracs[0] = 2. - 3.0*h;  // red
			rgbFracs[1] = 3.0*h - 1.0; // green
			rgbFracs[2] = 0.0;         // blue
			maxOutFracs(rgbFracs);
			// tellFracs(h,rgbFracs);
			return rgbFracs;
			}
	    //  iwh==3 case (red missing)
			rgbFracs[0] = 0.0;         // red
			rgbFracs[1] = 3. - 3.0*h;   // green
			rgbFracs[2] = 3.0*h - 2.0;         // blue
			maxOutFracs(rgbFracs);
			// tellFracs(h,rgbFracs);
			return rgbFracs;
		// always, rgbFracs[0] + rgbFracs[1] + rgbFracs[2] == 1.0
		}
	
	private void maxOutFracs(double [] fracs) {
		double fmax = fracs[0];
		if(fmax < fracs[1])
			fmax = fracs[1];
		if(fmax < fracs[2])
			fmax = fracs[2];
		for(int i=0;i<3;i++)
			fracs[i] = fracs[i]/fmax;
		}
		
	private void tellFracs(double h, double [] fracs) {
		talk("rgbFracsFromHue("+h+") returns redfrac="+fracs[0]+", greenfrac="+fracs[1]+
				", bluefrac="+fracs[2]);
		}
	
	public double [] hcwFromRGBvals(int r, int g, int b) {
		double [] rgbFracs = new double[3];
		rgbFracs[0] = (double)r/255.000;
		rgbFracs[1] = (double)g/255.000;
		rgbFracs[2] = (double)b/255.000;
		return hcwFromRGBfracs(rgbFracs);
		}
		
	public double [] hcwFromRGBfracs(double [] rgbFracs) {
	/* given rgbFracs[] whose values can range from 0.0 to 1.0, returns an array of
	 three doubles: [hue, chroma, whiteness] */
	 	double [] hcw = new double[3];
		double leastVal = rgbFracs[0];
		double sumfracs=0.;
		int i;
		double [] rgbF = new double[3];
		double chroma = 0.;
		int iwh = 3; // when red is absent from hue
		
		if(leastVal > rgbFracs[1]) {
			leastVal = rgbFracs[1];
			iwh = 1; // when green is absent from hue
			}
		if(leastVal > rgbFracs[2]) {
			leastVal = rgbFracs[2];
			iwh = 2; // when blue is absent from hue
			}
			
		for(i=0;i<3;i++) {
			rgbF[i] = rgbFracs[i] - leastVal; // at least one of these will be zero;
			sumfracs += rgbF[i];
			}
			
		if(sumfracs < .0001) { // r==g==b so we have no hue at all
			hcw[0]=0.5;
			hcw[1] = 0;
			hcw[2] = leastVal;
			return hcw;
			}
					
		switch (iwh) { // get the hue variable, which is hcw[0]...
			case 1: /* green is absent from hue, hue is between 0 and 1/3, and
							R = 3*h; G = 0; B = 1 -3*h */
				hcw[0] = oneThird*rgbF[0]/sumfracs;	
				break; 
			case 2: /* blue is absent from hue, hue is between 1/3 and 2/3, and
							R = (2 -3*h); G = (3*h-1); B = 0 */
				hcw[0] = oneThird*(2.0 - rgbF[0]/sumfracs);
				break; 
			case 3: /* red is absent from hue, hue is between 2/3 and 3/3, and
						     R = 0; G = (3 - 3*h); B = (3*h-2) */
				hcw[0] =  oneThird*(3.0 - rgbF[1]/sumfracs);
				break; 
			}
	/* Ok now we have the hue stored in hcw[0]. 
	r/255 = (1-chroma)*whiteness + chroma*redness(hue)
	g/255 = (1-chroma)*whiteness + chroma*greeness(hue)
	b/255 = (1-chroma)*whiteness + chroma*blueness(hue)
	If the primary with the least value is red, then r/255 = (1-chroma)*whiteness
	If the primary with the highest value were green, g/255 = (1-chroma)*whiteness + chroma*1.0
	thus highest-lowest = chroma*1.0, so chroma = highest-lowest = g - r in the example.
	
	Then whiteness = (highest-chroma)/(1 - chroma)
					*/
	double maxVal = rgbFracs[0];
	if(maxVal < rgbFracs[1])
		maxVal = rgbFracs[1];
	if(maxVal < rgbFracs[2])
		maxVal = rgbFracs[2];
	hcw[1] = maxVal - leastVal;
	hcw[2] = Math.abs(1. - hcw[1]) > .0001? (maxVal - hcw[1])/(1. - hcw[1]) : 1.0;
	return hcw;
	}
	
	public int [] RGBfromHCW(double hue, double chroma, double whiteness) {
		int [] rgb = new int[3];
		int gray = (int)(.499 + 255.0*whiteness); // hbw[2] is whiteness
		if(chroma < .00001) { // it's just gray
			rgb[0]=gray;
			rgb[1]=gray;
			rgb[2]=gray;
			return rgb;
			}
		// nope, it actually has some color	
		double [] rgbFracs = rgbFracsFromHue(hue);
				
		for(int i=0;i<3;i++) {
			rgb[i] = (int)(.499+255.*((1.-chroma)*whiteness + chroma*rgbFracs[i]));
			if(rgb[i]>255)
				rgb[i]=255;
			}
		return rgb;
		}
		
	private void report() {
		rtf.setText(""+red);
		gtf.setText(""+green);
		btf.setText(""+blue);
		}
		
	public void setNewColorFromHCW(double hue,double chroma,double whiteness) {
	    HUE=hue;
	    CHROMA=chroma;
	    WHITENESS=whiteness;
	    int [] newrgb =  RGBfromHCW(hue, chroma, whiteness);
	    red = newrgb[0];
	    green = newrgb[1];
	    blue = newrgb[2];
	    rtf.setText(""+red);
	    gtf.setText(""+green);
	    btf.setText(""+blue);
		}
					
	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
		
	public void pickCurrentColor() {
		if(parent1 != null) parent1.getAllColors();
		if(parent2 != null && iffirst) 
				parent2.getFirstColors();
		if(parent2 != null && !iffirst) 
				parent2.getSecondColors();
		dispose();
		}
		
	public void stateChanged(javax.swing.event.ChangeEvent e) {
		talk("StateChanged... shouldn't happen");
		report();
		} 
}
