package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.*;
import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;


class RasterCache  {
	String [] frameKeys = null;
	int Nras=0;
	Hashtable rases;
	int [] age = null;
	int maxCache=10; // the max number of rasters to store
	int nCache = 0; // the number of rasters currently cached...
	boolean ifdiags=false;
	
	public void sepuku() {
		empty();
		age=null;
		rases=null;
		}
		
	public RasterCache(int nframes, int maxCache) {
		frameKeys = new String[nframes];
		age  = new int[nframes];
		Nras=nframes;
		nCache=0;
		this.maxCache = maxCache;
		for(int i = 0;i<nframes;i++) {
			frameKeys[i]=""+i;
			age[i]=-1;
			}
		rases = new Hashtable(nframes,(float)(.75));
		}
	
	public void finalize() {
		empty();
		talk("                                                                           RasterCache snuffs");
		}
		
	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}

	byte [] fetch(int ipic) { // all frames get older each time fetch is called
		for(int i=0;i<Nras;i++)
			if(age[i]>=0) ++age[i];
			
		if(rases.containsKey(frameKeys[ipic])) {
			//talk("frame "+ipic+" was in the cache & had age ="+age[ipic]);
			age[ipic]=0; // whenever a raster is fetched, it's reborn anew
			return((byte [])(rases.get(frameKeys[ipic])));
			}
		else {
			if(ifdiags) talk("no frame "+ipic+" in the cache");
			return null;
			}
		}
	
	public int adjustCacheSize(int nc) {
	    if(ifdiags)talk("adjust cache size from "+maxCache+" to "+nc+". current size="+nCache);
		while(nc < maxCache)
			cutCacheSize();
		while (nc > maxCache)
			boostCacheSize();
		if(ifdiags)
		talk("this cache now holds "+nCache+" byte rasters and has max size "+maxCache);
		return(maxCache);
		} 
	
	private void boostCacheSize() {
		if(++maxCache > Nras) Nras = maxCache;
		if(ifdiags) talk("max cache size boosted to "+maxCache);
		}
		
	private void cutCacheSize() {
		if(--maxCache<1) maxCache=1;
		while(nCache > maxCache)
			removeOldest();
		}
		
	private void removeOldest() {
		int ioldest=0,old=-1,i;
		for(i=0;i<Nras;i++)
			if(age[i]>old) {
				ioldest=i;
				old = age[i];
				}
		if(old>=0) {// there was something older...
			rases.remove(frameKeys[ioldest]);
			age[ioldest]=-1;// flags this frame as absent
			--nCache;
	if(ifdiags)
		talk("uncached oldest section ="+ioldest+" whose age="+old+" nCache="+nCache);
			}
		}
	
	public void empty() {
		while(nCache>0) removeOldest();
		}
		
	public void add(int ipic,byte [] ras) {
		if(nCache >= maxCache)   // must remove oldest raster to make room for new raster
			removeOldest();
		rases.put(frameKeys[ipic],ras);
		age[ipic]=0;
		++nCache;
		if(ifdiags) talk("Add section "+ipic+" after which cache has "+nCache+" entries");
		}

}
