package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;


import java.io.*;

class RasterWriter extends JFrame implements java.awt.event.ActionListener,java.awt.event.ItemListener,java.awt.event.WindowListener {
	int width=100,height=190,whichStack=-1;
	int num_frames=0, sectNumToWrite=0;
	String fileName;
	ZseriesMainWindow parent;
	BioradFile myfile=null;
	int nomwidth=90, nomheight=150;
	JButton doit;
	JComboBox whichColors;
	JTextField frameNum,numberOfJFrames;	

	public RasterWriter(ZseriesMainWindow parent, String myfilesName) {
		this.parent = parent;
		fileName=myfilesName;
		setBackground(Color.gray);
		getContentPane().setLayout(new GridLayout(0,1));
		setSize(nomwidth,nomheight);
		setBackground(Color.gray);
		WidgetMaker wmk = new WidgetMaker(this);

		doit = wmk.makeAJButton("Do it",Color.gray);
		num_frames=-1;
		try {
			talk("Open .PIC file called <"+myfilesName+">");
			myfile=new BioradFile(myfilesName,true);
			}
		catch (IOException e) {
			talk("Oops. Failed to open a valid Biorad File");
			}
		if(myfile != null) {
			num_frames=myfile.num_pics;
			talk("The .PIC file has "+num_frames+" sections.");
			}
		setBackground(Color.gray);
		frameNum = wmk.makeAJTextField("0",6);
		frameNum.setBackground(Color.gray);
		int islash = myfilesName.lastIndexOf("/");
		String stg;
		if(islash<0) stg = myfilesName;
		else stg = myfilesName.substring(islash+1);
		getContentPane().add(wmk.makeAJLabel("file name is")); 
		getContentPane().add(wmk.makeAJLabel(stg)); 
		getContentPane().add(wmk.makeAJLabel(""+myfile.width+" = width"));
		getContentPane().add(wmk.makeAJLabel(""+myfile.height+" = height"));
		getContentPane().add(wmk.makeAJLabel("write section")); 
		getContentPane().add(frameNum);
		numberOfJFrames = wmk.makeAJTextField("of "+(num_frames-1), 6,Color.gray);
		getContentPane().add(numberOfJFrames);
		whichColors = wmk.makeAJComboBox(new String [] {"redChan","greenChan","blueChan"});
		whichStack=0;
		if(whichStack >= 0) 
			whichColors.setSelectedIndex(whichStack);
		setBackgroundOfwhichColors();
		getContentPane().add(whichColors);
		getContentPane().add(doit);	
		
		setJTextFields();
		addWindowListener(this);
		 }
		 
	private void setAllBacks(Color col) {
		setBackground(Color.gray);
		doit.setForeground(col);
		whichColors.setForeground(col);
		frameNum.setForeground(col);
		numberOfJFrames.setForeground(col);
		}
		

	private void setBackgroundOfwhichColors() {
		switch (whichStack) {
			case 0:
				setAllBacks(Color.red);
				break;
			case 1:
				setAllBacks(Color.green);
				break;
			case 2:
				setAllBacks(Color.blue);
				break;
			default :
				setAllBacks(Color.gray);
				break;
			}
		}
		
//catch Choice events...	
	public void itemStateChanged(java.awt.event.ItemEvent event)
		{
		String arg = (String)(((JComboBox)(event.getSource())).getSelectedItem());
      	if(arg.equals("redChan")) {
      		whichStack=0;
      		talk("operate on red channel");
			setBackgroundOfwhichColors();
	    	}
      	if(arg.equals("greenChan")) {
      		whichStack=1;
      		talk("operate on green channel");
			setBackgroundOfwhichColors();
	    	}
      	if(arg.equals("blue channel")) {
      		talk("operate on blueChan");
      		whichStack=2;
			setBackgroundOfwhichColors();
	    	}
		}


// catch button events...
	public void actionPerformed(java.awt.event.ActionEvent event) {
    	byte [] raster = null;
    	byte [] header = null;
		Object object = event.getSource();
		String butonlobel = ((JButton)object).getLabel();
		if (object == doit) {
	    		getSectNum();
	    		RandomAccessFile outfile = null;
	    		outfile = parent.openWritingFile(this,"Pick the file to write");
	    		if(outfile != null && myfile != null) {
	    			raster = myfile.readShiftedBioradRaster(sectNumToWrite);
	    			header = new byte[myfile.bioradHeaderLength];
	    			for(int i = 0; i<myfile.bioradHeaderLength;i++)
	    				header[i]=myfile.header[i];
	    			header[4]=1;
	    			header[5]=0; // these two bytes encode the number of pictures in the stack.
	    			if(raster != null) {
	    				try { 
	    					outfile.write(header);
	    					outfile.write(raster);
	    					outfile.close();
	    					}
	    				catch(IOException ioe) {
	    					talk("write operation failed??? "+ioe);
	    					}
	    				outfile=null;
 	    				raster=null;
	    				}
	    			else talk("OOps could not read the biorad file raster. ");
	    			}
	    			
	    		else talk("Couldn't open the file to write... and myfile = "+myfile);
				return;
    			}
		}
	
// catch WindowListener events  
	public void windowActivated(java.awt.event.WindowEvent event) {
		// talk("window activated");
		}
	public void windowDeactivated(java.awt.event.WindowEvent event) {
		//  talk("window Deactivated");
		}
	public void windowOpened(java.awt.event.WindowEvent event) {}
	public void windowIconified(java.awt.event.WindowEvent event) {}
	public void windowDeiconified(java.awt.event.WindowEvent event) {}
	public void windowClosing(java.awt.event.WindowEvent event) {
		//  talk("window closing");
		if(myfile != null)
			myfile.sepuku();
		myfile=null;
	    setVisible(false);
        dispose();
		}
	public void windowClosed(java.awt.event.WindowEvent event) {
		}
	
				   	
	public void finalize() {
		talk("                                                                 RasterWriter for file= "+fileName+" snuffs");
		}
		
   private int getSectNum() {
		sectNumToWrite = (int)(Double.valueOf(frameNum.getText()).doubleValue());
		talk("Write out section "+sectNumToWrite);
		if(sectNumToWrite<0) sectNumToWrite=0;
		if(sectNumToWrite>=num_frames)
			sectNumToWrite = num_frames-1;
		setJTextFields();
		return sectNumToWrite;
		}

		
    private void setJTextFields() {
	numberOfJFrames.setText("of "+num_frames);
	frameNum.setText(""+sectNumToWrite);
	setTitle("picWriter");
	}	

	public void changeJPanelSize(double ratio) {
		int pwidth = getSize().width;
		int pheight = getSize().height;
		pwidth = (int)(ratio*pwidth +.4999);
		pheight = (int)(ratio*pheight +.4999);
		setSize(pwidth,pheight);
		show();
		}

static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}

