package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;

class RidgesFltr {
	SegmenterNode nd=null;
	int [] ext,updown,gval;
double peaktol,peakthresh;
int pixthresh;

	public RidgesFltr(SegmenterNode nd) {
		this.nd=nd;
		} 
		
 public boolean doit(short [] red, int NCOLS, int NROWS, int pixthresh, double peakthresh, 
 	double peaktol, short [] mask) { 
	boolean ifverbose = true;
	this.pixthresh = pixthresh;
	this.peakthresh = peakthresh;
	this.peaktol = peaktol;
	if(nd.diags) nd.talk("RidgesFltr with pixthresh= "+pixthresh+", peakthresh= "+this.peakthresh+", peaktol= "+peaktol);

	if(pixthresh <0 || pixthresh >=255) {
		nd.talk("ridges(): Bad value of pixthresh = "+pixthresh);
		tellWhatYouDo();
		return false;
		}

	if(peakthresh<1.0001) {
		nd.talk("ridges(): Bad value of peakthresh = "+peakthresh);
		tellWhatYouDo();
		return false;
		}

	if(peaktol <0.) {
		nd.talk("ridges(): Bad value of peaktol = "+peaktol);
		tellWhatYouDo();
		return false;
		}
	walkridges(red,NCOLS,NROWS,mask);
	return true;
	}

	public void tellWhatYouDo() {
		nd.talk("");
		nd.talk("The RidgesFltr has a single grayscale input raster and three parameters,");
		nd.talk("'pixthresh' is an integer-valued parameter. 'peakthresh' and'peaktol' are");
		nd.talk("floating point parameters.");
		nd.talk("This filter won't work on black/white input rasters; Use only grayscale inputs.");		
		nd.talk("The RidgesFltr finds peaks where the pixel value exceeds pixthresh, and where");
		nd.talk("the ratio of the peak's height to each adjacent valley's height exceeds peakthresh.");
		nd.talk("peakthresh must exceed 1. Values of peakthresh close to 1.0 pick out peaks that are");
		nd.talk("only a little brighter than their surroundings. Larger values of peakthresh confine");
		nd.talk("attention to brighter ridges and/or peaks."); 
		nd.talk("");
		nd.talk("When peaktol is between .99 and 1.01, you get the exact single-pixel peak locations found.");
		nd.talk("");
		nd.talk("When peaktol<.99, the peak locations are the single-pixel wide centers of zones (adjacent ");
		nd.talk("to the exact peak) wherein the pixel values exceed the peak value times peaktol. In this case,");
		nd.talk("The smaller peaktol is, the wider is the zone whose centerline is this filter's output.");
		nd.talk("");
		nd.talk("When peaktol>1.01 then you get widened ridges showing all pixels near");
		nd.talk("the true peak where the peaktol*pixel value exceeds the value at the peak.");
		nd.talk("In this case, the larger peaktol is, the wider the rigdes are.");
		}		

/*************** walkridges() ******************/
	private void walkridges(short [] red,int NCOLS,int NROWS,short [] mask) {
	int i,j,npix,offset,k,l,x,y;
	int [] linein, lineout;

	npix=NCOLS*NROWS;
	for(i=0;i<npix;i++) 
		mask[i]=0;
	
	int nn = (int )(1.0 + Math.sqrt(NCOLS*NCOLS + NROWS*NROWS));
	// ridges uses line buffers of length nn long enough to span the diagonal..
	linein = new int[nn];
	lineout = new int[nn];
	ext = new int[nn];
	updown = new int[nn];
	gval = new int[nn];

	/* do the horizontal strips */
	for(i=0;i<NROWS;i++) {
		offset = i*NCOLS;
		for(j=0;j<NCOLS;j++) {
			linein[j] = red[offset+j];
			lineout[j]=0;
			}
 	  	ridgeline(linein,NCOLS,pixthresh,lineout);
		for(j=0;j<NCOLS;j++) 
			mask[offset+j] = (short)lineout[j];
		} 

	/* do the vertical strips */
	for(j=0;j<NCOLS;j++) { 
		for(i=0;i<NROWS;i++) {
			linein[i] = red[i*NCOLS+j];
			lineout[i]=0;
			}	
		ridgeline(linein,NROWS,pixthresh,lineout);
		for(i=0;i<NROWS;i++) 
			if(lineout[i]==255) // this implements an 'or' into the previous filled mask[] array.
				mask[i*NCOLS+j] = (short)lineout[i];
		} 
	
	/* do the 45 degree strips from lower left to upper right
	sweep down the right edge of the raster  */
	l=0;
	for(k=4; k<NROWS; k++) {
		y=k; x=0; l=0;
		while(y>=0 && x < NCOLS) {
			linein[l] = red[y*NCOLS+x];
			lineout[l] = 0;
			++l;
			++x; --y;  // this walks up the diagonal towards the upper right
			}
		if(l>5) {
			ridgeline(linein,l,pixthresh,lineout);
			y=k; x=0; l=0;
			while(y>=0 && x < NCOLS) {
				if(lineout[l]==255)
					mask[y*NCOLS+x] = (short )lineout[l];
				++l;
				++x; --y;
				}
			}
		} 
	/*  do the 45 degree strips from lower left to upper right
	sweep across the bottom edge of the raster  */
	for(k=0; k<NCOLS; k++) {
		y=NROWS-1; x=k; l=0;
		while(y>=0 && x < NCOLS) {
			linein[l] = red[y*NCOLS+x];
			lineout[l] = 0;
			++l;
			++x; --y;  // this walks up the diagonal towards the upper right
			}
		if(l>5) {
			ridgeline(linein,l,pixthresh,lineout);
			y=NROWS-1; x=k; l=0;
			while(y>=0 && x < NCOLS) {
				if(lineout[l]==255)
					mask[y*NCOLS+x] = (short )lineout[l];
				++l;
				++x; --y;  // this walks up the diagonal towards the upper right
				}
			}
		} 
		
	
	/* do the 45 degree strips from upper left to lower right
	sweep down the right edge of the raster  */
	for(k=0; k<NROWS-4; k++) {
		y=k; x=0; l=0;
		while(y<NROWS && x < NCOLS) {
			linein[l] = red[y*NCOLS+x];
			lineout[l] = 0;
			++l;
			++x; ++y;  // this walks up the diagonal towards the upper right
			}
		if(l>5) {
			ridgeline(linein,l,pixthresh,lineout);
			y=k; x=0; l=0;
			while(y<NROWS && x < NCOLS) {
				if(lineout[l]==255)
					mask[y*NCOLS+x] = (short )lineout[l];
				++l;
				++x; ++y;
				}
			}
		}
	/*  do the 45 degree strips from upper left to lower right
	sweep across the top edge of the raster  */
	for(k=0; k<NROWS-4; k++) {
		y=0; x=k; l=0;
		while(y<NROWS && x < NCOLS) {
			linein[l] = red[y*NCOLS+x];
			lineout[l] = 0;
			++l;
			++x; ++y;  // this walks up the diagonal towards the upper right
			}
		if(l>5) {
			ridgeline(linein,l,pixthresh,lineout);
			y=0; x=k; l=0;
			while(y < NROWS && x < NCOLS) {
				if(lineout[l]==255)
					mask[y*NCOLS+x] = (short )lineout[l];
				++l;
				++x; ++y;  // this walks up the diagonal towards the upper right
				}
			}
		}
	}

/********************** ridgeline() **************/
private void ridgeline(int [] g, int nline,int graythresh,int [] peakline) {
int going,i,j,next,gm,gi,gp,vim,vip,pip,pim,pi,ncuts,cut,up,pkthresh,
	ctrpeak,delta;
boolean FINDNEXTCUT=false,ifwidepeaks,ifexactpeaks;
double peaktoldiff=1.0;

up=255*255;
pkthresh = (int)(up*peakthresh);
// .99 < peaktol < 1.01 means return the exact peaks.
ifexactpeaks = (peaktol >= 0.99 && peaktol <= 1.01);
ifwidepeaks = !ifexactpeaks && peaktol>1.01 ;

peaktoldiff = ifexactpeaks? 0 : Math.abs(1.-peaktol);

for(j=0;j<nline;j++) {
    ext[j]=0;
    updown[j]=0;
    gval[j]=0;
    }
/* get raw peaks and valleys */
next=0;
i=1;
while(g[i]==g[i-1] && i<nline) ++i;
if(i>= nline) i = nline-1;
/* just found first change in gray level */
going = g[i]>g[i-1]? 1 : (-1);

while(i<nline) {
	if((delta=g[i]-g[i-1]) != 0) { /* gray level changed */
		if(going>0 && delta<0) { /* we were going up, but now we're 
				going down, so we just passed a local maximum */
			ext[next] = i-1;
			gval[next] = g[i-1];
			updown[next] = 1;
			going=(-1);
			++next;
			}

		if(going<0 && delta>0) { /* we were going down, but now we're 
				going up, so we just passed a local minimum */
			ext[next]=i-1;
			gval[next]=g[i-1];
			updown[next]=(-1);
			going=1;
			++next;
			}
		}
	++i;
	} /* end while(i<nline) loop */
/* now cut whimpy peaks until no wimpy peaks are left */
ncuts=1;
while(ncuts>0) {
  ncuts=0;
  FINDNEXTCUT=false;
  while(i>1 && !FINDNEXTCUT) {
	if(updown[i] > 0) { // at a peak... 
		pi=gval[i];
		vip=gval[i+1];
		vim=gval[i-1];
		if(pi<graythresh) {  /* peak is too low to be interesting */
			cut = vip<vim? i-1 : i;
			// CUT: 
			++ncuts;
			for(j=cut; j<=next-3; j++) {
				ext[j]=ext[j+2];
				gval[j]=gval[j+2];
				updown[j]=updown[j+2];
				}	
			next-=2;
			i=cut+2;
			if(i>next-3) i=next-3;
			// nd.talk("DDD and i="+i);   
			FINDNEXTCUT=true;
			}
		if(!FINDNEXTCUT) {
		  pip=gval[i+2];
		  pim=gval[i-2];

		  if(vim>vip) { /*  this peak must exceed highest 
				nearby valley (which is vim) by enough */
			if(up*pi < pkthresh*vim) {
				if(pim<pi) cut=i-2;
				else cut = i-1;
			// CUT: 
			++ncuts;
			for(j=cut; j<=next-3; j++) {
				ext[j]=ext[j+2];
				gval[j]=gval[j+2];
				updown[j]=updown[j+2];
				}	
			next-=2;
			i=cut+2;
			if(i>next-3) i=next-3;
			FINDNEXTCUT=true;
				}
			}
		  else if(!FINDNEXTCUT) { /*  this peak must exceed highest 
				nearby valley (which is vip) by enough */
			if(up*pi<pkthresh*vip) {
				if(pip<pi) cut = i+1;
				else cut=i;
			// CUT: 
			++ncuts;
			for(j=cut; j<=next-3; j++) {
				ext[j]=ext[j+2];
				gval[j]=gval[j+2];
				updown[j]=updown[j+2];
				}	
			next-=2;
			i=cut+2;
			if(i>next-3) i=next-3;
			FINDNEXTCUT=true;
			}
		   }
		  } // end of first if(!FINDNEXTCUT) 
		} // end of if(updown[i]>0) ...
	  if(!FINDNEXTCUT) {
	  	--i;
	  	}
	  } /* end of while (i>1) loop  */
	}  /* end of while(ncuts>0) loop ... keep going through the whimpy eak cutting
	loop until there are no whimpy peaks left to cut... */

/* OK. All the peaks are now valid by being sufficiently higher than
their adjacent valleys... */
//nd.talk("GGG ");   

if(updown[0]>0) { /* first extremum is a peak */
	pi=gval[0];
	if(up*pi>=pkthresh*gval[1] && pi>= graythresh) {
		if(ifwidepeaks) {
			  paintwidepeak(g,ext[0],   ext[0],ext[1], (int )(1024*peaktoldiff*pi),peakline,graythresh);
        //    paintwidepeak(g,ipeakLoc, imin,  imax,         tol,           int [] peakline,graythresh)
			}
		else {
		  	ctrpeak = findtruepeak(g,ext[0],ext[0],ext[1],
			    (int )(peaktoldiff*pi),up,pkthresh);
			peakline[ctrpeak]=255;
			}
		}
	}
if(next>=nline) next=nline; 
if(next>0 && updown[next-1]>0) { /* last extremum is a peak */
	pi=gval[next-1];
	if(next<2) next=2;
	if(up*pi>=pkthresh*gval[next-2] && pi>= graythresh) {
		if(ifwidepeaks) {
			  paintwidepeak(g,ext[next-1],ext[next-2],ext[next-1], (int )(1024*peaktoldiff*pi),peakline,graythresh);
        //    paintwidepeak(g,ipeakLoc,     imin,       imax,               tol,        int [] peakline,graythresh)
			}
		else {
		  ctrpeak = findtruepeak(g,ext[next-1],ext[next-2],ext[next-1],
				(int )(peaktoldiff*pi),up,pkthresh);
		  peakline[ctrpeak]=255;
		  }
		}
	}

if(next>1 && ifwidepeaks) {
   for(i=1;i<next-1;i++) 
	if(updown[i]>0) {
		pi=up*gval[i];
		if(pi > pkthresh*gval[i-1] &&
			pi > pkthresh*gval[i+1]) {
			  paintwidepeak(g, ext[i],   ext[i-1],ext[i+1], (int )(1024*peaktoldiff*gval[i]),peakline,graythresh);
        //    paintwidepeak(g,ipeakLoc,    imin,   imax,                      tol,    int [] peakline,graythresh)
			    }
		}
	}

else {
  if(next>1) {
   for(i=1;i<next-1;i++) 
	if(updown[i]>0) {
		pi=up*gval[i];
		if(pi > pkthresh*gval[i-1] &&
			pi > pkthresh*gval[i+1]) {
			    ctrpeak = findtruepeak(g,ext[i],ext[i-1],ext[i+1],
				(int )(peaktoldiff*gval[i]),up,pkthresh);
			    peakline[ctrpeak]=255;
			    }
		}
    }
   }
}

/********************** findtruepeak() **************/
private int findtruepeak(int [] g, int i,int imin,int imax, int tol,int up, int pkthresh) {
int avg,j,navg,diff;
boolean UPSWEEP=false;
if(tol==0) return (i);

avg=i;
navg=1;
j=i;
while(j>imin && !UPSWEEP) {
	--j;
	if(up*g[j] < pkthresh*g[imin]) UPSWEEP=true;
	else {if((diff=g[i]-g[j]) <0) diff = (-diff);
		if(diff>tol) 
			UPSWEEP=true;
		}
	if(!UPSWEEP) {
		avg +=j;
		navg++;
		}
	}
// UPSWEEP:
j=i;
while(j<imax) {
	++j;
	if(up*g[j] < pkthresh*g[imax]) {
		i=avg/navg;
		return(i);
		}
	if((diff=g[i]-g[j]) <0) diff = (-diff);
	if(diff>tol) {
		i=avg/navg;
		return(i);
		}
	avg +=j;
	navg++;
	}
i=avg/navg;
return(i);
}


/********************** paintwidepeak() **************
                                                                 tol=1024*|1-peaktol|*g[i]        */
private void paintwidepeak(int [] g, int i,int imin,int imax,int tol, int [] peakline,int graythresh) {
/*	                           g ext[i]  ext[i-1] ext[i+1] 		                    
put a 255 in peakline[j] for all  imin <= j <= imax at which
g[j] differs from g[i] by less than tol. Work j out from i in
increasing then decreasing directions  */

int avg,j,navg,diff;
boolean UPSWEEP=false;
peakline[i]=255;
avg=i;
navg=1;
j=i;
while(j>imin && !UPSWEEP) {  // move j to the left
	--j;
	if((diff=1024*(g[i]-g[j])) > tol ) 
	    UPSWEEP=true;
	else if(g[j]>graythresh) 
		peakline[j]= 255 ;	
	}

j=i;
while(j<imax) {  // move j to the right
	++j;
	if((diff=1024*(g[i]-g[j]))>tol) return;
	if(g[j]>graythresh) 
	    peakline[j]=255;
	}
}

}