package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;

class SegPar extends JPanel implements java.awt.event.ActionListener {
int intVal=0;
double doubleVal=0.;
boolean ifDoublePar=false;
boolean ifVary=false;
String parName,varyParName;
JButton nameTogglesVary;
JTextField parText;
SegmenterNode parent=null;
int minVal=0, maxVal=255;
WidgetMaker wmk;


public SegPar(int iparVal, int minV, int maxV, String name, SegmenterNode parent) {
	this.parent = parent;
	intVal=iparVal;
	ifDoublePar=false;
	// this is an integer parameter
	wmk = new WidgetMaker(this);
	parText = wmk.makeAJTextField((""+intVal),4);
	minVal=minV;
	maxVal=maxV;
	commonSetup(name);
	}

public SegPar(double dparVal, double minV, double maxV, String name, SegmenterNode parent) {
	this.parent = parent;
	ifDoublePar=true;
	intVal=(int)(10000*dparVal);
	// this is a double precision parameter
	wmk = new WidgetMaker(this);
	parText = parText = wmk.makeAJTextField((""+(.0001*intVal)),6);
	minVal=(int)(10000*minV);
	maxVal=(int)(10000*maxV);
	commonSetup(name);
	}

	private void commonSetup(String name) {
		this.parName=name;
		varyParName = "*"+parName+"*";
		setLayout(new GridLayout(1,0));
		nameTogglesVary = wmk.makeAJButton(parName);
		add(nameTogglesVary); add(parText);
		wmk=null;
		}
	
	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
	
	private String shortNumString(double val) {
		String stg = ""+val;
		if(stg.length() <= 8)
			return stg;
		return stg.substring(0,8);
		}
		
	public void setVal(double dval) {
		intVal = (int)(10000*dval);
		if(intVal<minVal) intVal=minVal;
		if(intVal>maxVal) intVal=maxVal;
		doubleVal=.0001*intVal;
		parText.setText(shortNumString(dval));
		}
	
	public void setVal(int ival) {
		intVal = ival;
		if(intVal<minVal) intVal=minVal;
		if(intVal>maxVal) intVal=maxVal;
		parText.setText(""+ival);
		}
			
	public int getVal() {  
		doubleVal = Double.valueOf(parText.getText()).doubleValue();
		int ipar;
		if(ifDoublePar) {
			ipar=(int)(10000*doubleVal);
			if(ipar<minVal) ipar=minVal;
			if(ipar>maxVal) ipar=maxVal;
			}
		else {
			ipar = (int)doubleVal;
			if(ipar<minVal) ipar=minVal;
			if(ipar>maxVal) ipar=maxVal;
			}
		if(ipar != intVal) { 
			intVal=ipar;   // new parameter value
			parent.someParChanged();
			}
		if(ifDoublePar) parText.setText(shortNumString(doubleVal));
		else  parText.setText(""+intVal);
		return intVal;
		}
	
	public int getMutantVal() {
		int mutantVal,ival = getVal(); // this is the curent actual value
		// perturb it
		double rnd = 2.*(Math.random()-.5)*parent.parent.mutateFrac;
		if((mutantVal = (int)((1.0+rnd)*ival)) == ival) {
			if(ival>minVal && rnd<0) mutantVal=ival-1;
			else mutantVal=ival+1;
			}
		if(mutantVal>maxVal)
			mutantVal=maxVal;
		if(mutantVal<minVal)
			mutantVal=minVal;
		if(ifDoublePar) 
			setVal(.0001*mutantVal);
		else 
			setVal(mutantVal);
		parent.someParChanged();
		return mutantVal;
		}
		
	public void actionPerformed(java.awt.event.ActionEvent event) {
		String arg = event.getActionCommand();
	    if(arg.equals(parName) || arg.equals(varyParName)){
	    	ifVary = !ifVary;
	    	if(ifVary) 
	    		nameTogglesVary.setLabel(varyParName);
	    	else
	    		nameTogglesVary.setText(parName);
    		}
    	}
    	
}