package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;

class Segmenter extends JFrame implements Runnable,java.awt.event.ActionListener, java.awt.event.WindowListener{
	SegmenterCanvas canvas;
	BestSegJFrame showBestPic=null;
	SegTower [] towers = null;
	SegmenterNode [][] nodes;
	ShowPicJPanel parent;
	SegPic theBestPic=null;
	SegPic currentPic=null; // most recent pic made by runTheChain();
	int maskMakerXorYorZ=-1; /* 1 means XY section stack is loaded currently, 
								2 means XZ section stack is loaded currently
								3 means the ZY section stack is loaded currently
								anything else means no particular maskmaking stack is loaded  */
	

	int making3Dmask=0; // set to 1, 2, or 3 when making a 3D mask...
	BitMask a3dmask=null;
	String [] maskKeyWord = new String[] {"","mask1","mask2","mask3","mask1","mask2","mask3"};					
	int numXYmaskSects=-1;
	int numXZmaskSects=-1;
	int numZYmaskSects=-1;
	MakeA3Dmask maskMaker=null; // used only to restore visibity
	static final int XYSECT=1, XZSECT=2, ZYSECT=3;
	boolean maskMakingPaused=false;
	SegmenterNode the3DmaskNode=null;
	int mask3DNodesInChan=-1;
	int maskPasses = 20;

	boolean ifVaryingParameters=false;
	int lastClickedPicTower = -1, lastClickedPicNumber = -1;
	int curJFrameDepth=1;
	int curSect=0;
	boolean ifXYsect=true;
	int [][] maskrgbs = new int[4][6];
	int nTowerCanvases=1;
	
	boolean ifWritePostscriptFile=false;
	int npaths=2; // path1 always holds special nodes like chan1In, chan2In, chan3In,show,maskOut
	int nsteps=10;
	int xinc = 100;
	int yinc = 100;
	int width=(nsteps+1)*xinc;
	int height=(npaths)*xinc;
	int nomwidth=width, nomheight=height+40;
	Thread runChainThread = null;
	JButton varyParsJButton,pauseButton;
	
	int displayRow=-1,	displayCol=-1;
	static int possibilities=26;
	int fillHeight = 170;
	int nullID = 18;
	int showRow=-1,showCol=-1;
	long calcGen=0; // counter that advances every time we run some part of the filter chain
	double mutateFrac=.07;
	int stampHeight=150;
	int stampWidth=(4*stampHeight)/3;
	int stampXorig=0,stampYorig=0;
	boolean ifShowingBestPic = false;
	boolean ifUseFilterChainsInFrontOfMainDisplay=false;
	JButton filterUseToggleButton;
	
	public void PutDataThroughFilterChain(boolean ifput) {
		ifUseFilterChainsInFrontOfMainDisplay=ifput;
		if(ifUseFilterChainsInFrontOfMainDisplay)
			filterUseToggleButton.setText("(toggle)useFilters3Drecons");
		else
			filterUseToggleButton.setText("(toggle) training filters");
		Stereo3DReconstruction.PutDataThroughFilterChain(ifput);
		}
	
	public boolean ifPutDataThroughFilterChain() {
		return ifUseFilterChainsInFrontOfMainDisplay;
		}
	
	public void newStampLocation(int stampXorig,int stampYorig) {
		this.stampXorig=stampXorig;
		this.stampYorig=stampYorig;
		for(int i=0;i<nTowerCanvases;i++) {
			towers[i].newStampLocation(stampXorig,stampYorig);
			towers[i].canvas.repaint();
			}
		}
	
	public void  dismissAllBut(int survivingStatus) {
		for(int i=0;i<nTowerCanvases;i++)
			towers[i].dismissAllBut(survivingStatus);
		}
			
	public void noOneIsBest() {
		if(towers != null) 
			for(int id=0;id<nTowerCanvases; id++)
				towers[id].noOneIsBest();
		theBestPic=null;
		}
	
	public void setBestYet(SegPic theBestPic) {
		this.theBestPic = theBestPic;
		if(theBestPic != null && 
				(showBestPic == null || !showBestPic.confirmSize(theBestPic.wholeCols,theBestPic.wholeRows)) ) {
			talk("Segmenter.setBestYet");
			if(showBestPic != null) 
				showBestPic.sepuku();
			talk("Segmenter....    MAKING NEW BestSegJFrame");
			showBestPic = new BestSegJFrame(stampWidth,stampHeight,stampXorig,stampYorig, 
				theBestPic, this);
			}
		avoidGiantStamps(); // makes sure stamps are not taller than the raw data
		talk("Segmenter....    showing new BestSegJFrame");
		showBestPic.newBestPic(theBestPic);
		}
	
	private void makeShowTowerCanvasses() {
		if(towers != null) return;
		talk("Segmenter... making new SegTower");
		towers = new SegTower[nTowerCanvases];
		for(int id=0;id<nTowerCanvases; id++)
			towers[id] = new SegTower(id, stampWidth, stampHeight,this);
		}
	
	private void avoidGiantStamps() {
		if(stampHeight > stampHeightBound(stampHeight))
			newStampSizes();			
		}
		
	private int stampHeightBound(int aheight) {
		if(theBestPic==null || aheight < theBestPic.wholeRows) 
			return aheight;
		else
			return theBestPic.wholeRows - 1;
		}
		
	private void newStampSizes() {
		stampHeight = stampHeightBound(stampHeight);
		stampWidth=(4*stampHeight)/3;
		if(showBestPic != null ) {
			showBestPic.changeStampSizes(stampWidth,stampHeight);
			for(int id=0;id<nTowerCanvases;id++)
				towers[id].changeStampSizes(stampWidth,stampHeight);
			newStampLocation(showBestPic.canvas.stampXorig, showBestPic.canvas.stampYorig);
			}
		}
			
	public void changeStampSizes (boolean ifLarger) {
		unRegisterLastClickedPic(1);
		if(ifLarger) {
			if(stampHeight >= 250) {
				talk("the stamps are already at their maximim size");
				return;
				}
			stampHeight=(7*stampHeight)/6;
			if(stampHeight > 250)
				stampHeight=250;
			}
		else {
			if(stampHeight<=50) {
				talk("the stamps are already at their smallest");
				return;
				}
			stampHeight=(5*stampHeight)/6;
			if(stampHeight < 100)
				stampHeight=100;
			}
		newStampSizes();
		}
		
	public void makeAnotherSegTower() {
			SegTower []  newTowers = new SegTower[nTowerCanvases+1];
			unRegisterLastClickedPic(2);
			avoidGiantStamps();
			for(int id=0;id<nTowerCanvases; id++)
				newTowers[id]=towers[id]; // copy old towers into new place holder
			newTowers[nTowerCanvases] = new SegTower(nTowerCanvases, stampWidth, stampHeight,this);
			towers=null;
			towers = newTowers;
			++nTowerCanvases;
			for(int id=0;id<nTowerCanvases; id++)
				towers[id].canvas.repaint();
			}
			
	public void eliminateLastSegTower() { // only if it has only empty pics
		unRegisterLastClickedPic(3);
		if(nTowerCanvases<=1) {
			talk("You cannot eliminate the only remaining stamp tower");
			return;
			}
		int lastTow = nTowerCanvases-1, nhigh = towers[0].canvas.nhigh;
		boolean lastTowerIsEmpty=true;
		for(int j=0;j<nhigh;j++)
			if(towers[lastTow].canvas.thePics[j].status >= 0 ) 
				lastTowerIsEmpty=false;
		
		if(!lastTowerIsEmpty) {
			reOrderStampsInTowers();
			for(int j=0;j<nhigh;j++)
				if(towers[lastTow].canvas.thePics[j].status >= 0 ) {
					talk("You must empty all the pics in stamp tower "+lastTow+" before you can");
					talk("eliminate it");
					return;
					}
			}

		// OK cut last tower.
		--nTowerCanvases;
		towers[lastTow].sepuku();
		towers[lastTow]=null;
		}	
			
	public void reOrderStampsInTowers() { // bring highest status stamps up and to the left;
		int curtow,curpic,i,j,iswap,jswap,curStat,highestStat,nhigh = towers[0].canvas.nhigh;
		SegPic tempPic;
		SegTowerCanvas tempParent;
		unRegisterLastClickedPic(4);
		for(curtow=0;curtow<nTowerCanvases;curtow++) 
			for(curpic=0;curpic<nhigh;curpic++) {
				curStat =sortStatus(curtow,curpic);
				highestStat = curStat;
				iswap=curtow; jswap = curpic;
				for(j=curpic+1;j<nhigh;j++) // find higher status pics in this tower.
					if(sortStatus(curtow,j) > highestStat) {
						highestStat = sortStatus(curtow,j);
						iswap=curtow;
						jswap =j;
						}
				for(i=curtow+1;i<nTowerCanvases;i++) // seek higher status pics on other towers.
					for(j=0;j<nhigh;j++) 
						if(sortStatus(i,j) > highestStat) {
							highestStat = sortStatus(i,j);
							iswap=i;
							jswap =j;
							}

				if(highestStat>curStat) { // if there's a higher status stamp, swap it
					tempPic = towers[iswap].canvas.thePics[jswap];
					towers[iswap].canvas.thePics[jswap] = towers[curtow].canvas.thePics[curpic];
					towers[curtow].canvas.thePics[curpic] = tempPic;
					// swap their parents so they inform the right tower when their status changes
					tempParent = towers[curtow].canvas.thePics[curpic].parent;
					towers[curtow].canvas.thePics[curpic].parent = 
							towers[iswap].canvas.thePics[jswap].parent;
					towers[iswap].canvas.thePics[jswap].parent = tempParent;					
					}
				}
		for(i=0;i<nTowerCanvases;i++) 
			for(j=0;j<nhigh;j++)  {
				towers[i].canvas.thePics[j].isStampCurrent = false;
				}
		for(i=0;i<nTowerCanvases;i++) 
			towers[i].canvas.repaint();
		}
		
	private int sortStatus(int i, int j) {
		int stat =	towers[i].canvas.thePics[j].status;
		if(stat==3)  // it's the bestYet pic. Assign it the highest winning status
			return 7777777;
		if(stat>=0) 
			return towers[i].canvas.thePics[j].nPixOn;
		else return stat;
		}
				
	public void invalidateShowTowerPics() {
		unRegisterLastClickedPic(5);
		if(towers != null)
		   for(int id=0;id<nTowerCanvases; id++)
			  towers[id].emptyPics();
		}
	
	public int leftmostTowerWithNonContender() {
		if(towers == null)
			return -1;

		for(int id=0;id<nTowerCanvases; id++)
			if(towers[id].firstNonContender() >=0)
				return id;
		return -1;// means there are no spaces holding non-contenders available
		}	

	public int leftmostTowerWithEmptyPic() {
		for(int id=0;id<nTowerCanvases; id++)
			if(towers[id].firstEmptyPic() >=0)
				return id;
		return -1; // means there are no empty pic spaces available
		}	
		
	public void tellAllColors() {
		talk("The red chan colors are: ("+maskrgbs[0][0]+", "+maskrgbs[0][1]+", "+maskrgbs[0][2]+") and ("+
		maskrgbs[0][3]+", "+maskrgbs[0][4]+", "+maskrgbs[0][5]+")");

		talk("The green chan colors are: ("+maskrgbs[1][0]+", "+maskrgbs[1][1]+", "+maskrgbs[1][2]+") and ("+
		maskrgbs[1][3]+", "+maskrgbs[1][4]+", "+maskrgbs[1][5]+")");

		talk("The blue chan colors are: ("+maskrgbs[2][0]+", "+maskrgbs[2][1]+", "+maskrgbs[2][2]+") and ("+
		maskrgbs[2][3]+", "+maskrgbs[2][4]+", "+maskrgbs[2][5]+")");
		}

	public Segmenter(ShowPicJPanel parent) {
		super("Data Segmentation          processing flows left to right");
		this.parent=parent;
		setBackground(new Color(200,200,200));
		getContentPane().setLayout(new BorderLayout());
		canvas = new SegmenterCanvas(this);
		WidgetMaker wmk = new WidgetMaker(this);
		JPanel midpan,toppan,botpan;
		maskrgbs[0][0] = 0;   // markred for the red channel mask display
		maskrgbs[0][1] = 255; // markgreen for the red channel mask display
		maskrgbs[0][2] = 255; // markblue for the red channel mask display
		maskrgbs[0][3] = 255; // dimred for the red channel mask display
		maskrgbs[0][4] = 0;   // dimgreen for the red channel mask display
		maskrgbs[0][5] = 0;   // dimblue for the red channel mask display
		
		maskrgbs[1][0] = 255; // markred for the green channel mask display
		maskrgbs[1][1] = 0;   // markgreen for the green channel mask display
		maskrgbs[1][2] = 255; // markblue for the green channel mask display
		maskrgbs[1][3] = 0;   // dimred for the green channel mask display
		maskrgbs[1][4] = 255; // dimgreen for the green channel mask display
		maskrgbs[1][5] = 0;   // dimbluefor the green channel mask display
		
		maskrgbs[2][0] = 255; // markred for the blue channel mask display
		maskrgbs[2][1] = 255; // markgreen for the blue channel mask display
		maskrgbs[2][2] = 0;   // markblue for the blue channel mask display
		maskrgbs[2][3] = 0;   // dimred for the blue channel mask display
		maskrgbs[2][4] = 120; // dimgreen for the blue channel mask display
		maskrgbs[2][5] = 255; // dimbluefor the blue channel mask display
		
		maskrgbs[3][0] = 128; // markred for the other channel mask display
		maskrgbs[3][1] = 180; // markgreen for the other channel mask display
		maskrgbs[3][2] = 255; // markblue for the other channel mask display
		maskrgbs[3][3] = 255; // dimred for the other channel mask display
		maskrgbs[3][4] = 180; // dimgreen for the other channel mask display
		maskrgbs[3][5] = 128; // dimbluefor the other channel mask display
		

		toppan = new JPanel();
		toppan.setLayout(new GridLayout(1,0));
		toppan.add(wmk.makeAJButton("re-order postage stamps"));	
		toppan.add(wmk.makeAJButton("more postage stamps"));	
		toppan.add(wmk.makeAJButton("fewer postage stamps"));	
		toppan.add(wmk.makeAJButton("write segmentation file"));
		toppan.add(wmk.makeAJButton("read segmentation file"));	
		pauseButton = wmk.makeAJButton("make a 3D mask");
		toppan.add(pauseButton);		
		toppan.add(wmk.makeAJButton("stop Making 3D  mask"));		

		midpan = new JPanel();
		midpan.setLayout(new GridLayout(1,0));
		midpan.add(wmk.makeAJButton("Run the filter chain"));
		midpan.add(wmk.makeAJButton("add new path (row)"));
		midpan.add(wmk.makeAJButton("new sect; rerun all pics"));
		midpan.add(wmk.makeAJButton("emptyAllButBestYet"));
		midpan.add(wmk.makeAJButton("make stamps larger"));
		midpan.add(wmk.makeAJButton("click help"));
		midpan.add(wmk.makeAJButton("--------------"));		

		botpan = new JPanel();
		botpan.setLayout(new GridLayout(1,0));
		botpan.add(wmk.makeAJButton("re-order stamps"));
		botpan.add(wmk.makeAJButton("edit all parameters"));
		varyParsJButton=wmk.makeAJButton("vary parameters");   botpan.add(varyParsJButton);
		botpan.add(wmk.makeAJButton("emptyAllButContenders"));
		botpan.add(wmk.makeAJButton("make stamps smaller"));
		filterUseToggleButton = wmk.makeAJButton("(toggle) training filters");
		botpan.add(filterUseToggleButton);		
		botpan.add(wmk.makeAJButton("available  "));		
		
		JPanel buttonPan = new JPanel();
		buttonPan.setLayout(new GridLayout(0,1));
		buttonPan.add(toppan);
		buttonPan.add(midpan);
		buttonPan.add(botpan);
		setWinSize();
		buttonPan.setSize(width,40);
		getContentPane().add("North",buttonPan);
		getContentPane().add("Center",canvas);
		addWindowListener(this);
		nodes =  new SegmenterNode[npaths][nsteps];
		int whatAmI;
		for(int col=0;col<nsteps;col++) {
			whatAmI = 0;
			if(col==0)
				 whatAmI = 15; // zGrad starts the chain 
			if(col==nsteps-1)
				 whatAmI = 5; // BlkWht filter ends the chain 
			nodes[1][col] = new SegmenterNode(this,1,col,whatAmI,(1+col)*xinc, yinc/2, 
						(4*xinc)/5, (4*yinc)/5,  -1, -1, -2, -2, false);
			if(col>0) { // established straight wire links
				nodes[1][col].inPath1 = 1;
				nodes[1][col].inStep1 = col-1;
				}
			}
//                     0          1        2       3          4          5          6      7      8           9       10      11       12       13       14       15       16      17       18     19         20        21       22       23     24      25
//String [] whats = {"wire","threshold","erode","thicken","ridges","BlkWht","lowpass","NiceBalls","NiceTubes","invert","skel","undust","ThreshAnd","ROImask","DimEdges","Zgrad","gapFill","avail6","null","chan1","chan2","chan3","mask1","mask2","mask3","mask4"};
		nodes[0][1] = new SegmenterNode(this,0,1,possibilities-7,-1,-1,(4*xinc)/5, (4*yinc)/5,  -1, -1, -2, -2, false);
			// nodes[0][1] is chan1In  
		nodes[0][2] = new SegmenterNode(this,0,2,possibilities-6,-1,-1,(4*xinc)/5, (4*yinc)/5,  -1, -1, -2, -2, false);
			// ch2in = nodes[0][2]
		nodes[0][3] = new SegmenterNode(this,0,3,possibilities-5,-1,-1,(4*xinc)/5, (4*yinc)/5,  -1, -1, -2, -2, false);
			// ch3in = nodes[0][3]
		nodes[0][4] = new SegmenterNode(this,0,4,possibilities-4,-1,-1,(4*xinc)/5, (4*yinc)/5,  -1, -1, 0, 1, false);
			// mask1  = nodes[0][4] 
		nodes[0][5] = new SegmenterNode(this,0,5,possibilities-3,-1,-1,(4*xinc)/5, (4*yinc)/5,  -1, -1, 0, 2, false);
		    // mask2 nodes[0][5]
		nodes[0][6] = new SegmenterNode(this,0,6, possibilities-2,-1,-1,(4*xinc)/5, (4*yinc)/5,  -1, -1, 0, 3, false);
			// mask3  = nodes[0][6] 
		nodes[0][0] = new SegmenterNode(this,0,0, possibilities-1,-1,-1,(4*xinc)/5, (4*yinc)/5,  -1, -1, -1, -1, false);
			// mask4  = nodes[0][0] 
		for(int i=7;i<nsteps;i++) // fill out the rest of row 0 with null nodes
			nodes[0][i] = new SegmenterNode(this,0, i, -1,-1,-1,(4*xinc)/5, (4*yinc)/5,  -1, -1, -2, -2, false);
		setWinSize();
		newConfig();
		}
		
	public void changeJPanelSize(double ratio) {
		int pwidth = getSize().width;
		int pheight = getSize().height;
		pwidth = (int)(ratio*pwidth +.4999);
		pheight = (int)(ratio*pheight +.4999);
		setSize(pwidth,pheight);
		show();
		}

	public void repaintAll() {
		// talk("repaintAll()");
		canvas.repaint();
		// repaint();
		}
		
	private void setWinSize() {
		width = xinc*(1+nsteps);
		height = yinc*npaths;
		}
		
	private void showMaskNode(int iwh) {
		if(iwh<1 || iwh>4) {
			talk("argument to showMaskNode() must be a mask number between 1 and 4");
			return;
			}
		int i,maskrow=0,maskcol=4,inchanrow=0,inchancol=1,whichInChan=-1;
		switch (iwh) {
			default :
			case 1 :
				// mask1  = nodes[0][4] 
				maskcol=4;
				// parent.redOnly(); removed july 2001
				break;
			case 2 :
				// mask2  = nodes[0][5] 
				maskcol=5;
				// parent.greenOnly(); removed july 2001
				break;
			case 3 :
				// mask3  = nodes[0][6] 
				maskcol=6;
				// parent.blueOnly();   removed july 2001
				break;
			case 4 :
				// mask4  = nodes[0][0] 
				maskcol=4;
				break;				
			}
		talk("mask is in row "+maskrow+", and col "+maskcol);
		setShowNode(maskrow,maskcol);
		the3DmaskNode=nodes[showRow][showCol];
		mask3DNodesInChan=-1;
		// find out which input channel this mask is for
		talk("the3DmaskNode.inStep2 = "+the3DmaskNode.inStep2);
		if(the3DmaskNode.inPath2 == 0) {
			if(the3DmaskNode.inStep2 == 1) {
				mask3DNodesInChan=1; // red chan in
				talk("Mask "+iwh+" operates on red input channel ");
				// parent.redOnly();  removed july 2001
				}
			if(the3DmaskNode.inStep2 == 2) {
				mask3DNodesInChan=2; // green chan in
				talk("Mask "+iwh+" operates on green input channel ");
				// parent.greenOnly();  removed july 2001
				}
			if(the3DmaskNode.inStep2 == 3) {
				mask3DNodesInChan=3; // red chan in
				talk("Mask "+iwh+" operates on blue input channel ");
				// parent.blueOnly();  removed july 2001
				}
			}
		}
		
	public void setShowNode(int row, int col){
		talk("*****************************************setShowNode("+row+", "+col+")");
		if(showRow != row || showCol != col) {
			// invalidate all the SegTower Pics
			talk("*************************************setShowNode("+row+", "+col+") actually set");
			showRow=row;
			showCol=col;
			}
		markActiveNodes();
		}
					
	public void newConfig() {
		setWinSize();
		setFillHeight();
		int row,col;
		setSize(width,fillHeight);
		for(row=1;row<npaths;row++) 
			for(col=0;col<nsteps;col++) 
			   if(nodes[row][col].whatAmI >=0)  
				   nodes[row][col].newGeometry(row,col,(1+col)*xinc, yinc/2+(row-1)*yinc, (4*xinc)/5, (4*yinc)/5);
				
		for(col=0;col<nsteps;col++) // do the special nodes in path 0
			if(nodes[0][col].whatAmI >=0)  {
				nodes[0][col].newGeometry(0,col,-1,-1, (4*xinc)/5, (4*yinc)/5);
				}
		markActiveNodes();
		}
	
	public int howManyParametersToVary() {
		int nvary=0;
		for(int row=1;row<npaths;row++)
			for(int col=0;col<nsteps;col++)
				if(nodes[row][col] != null)
					nvary += nodes[row][col].howManyParametersToVary();
		return nvary;
		}
		
	public void hideAllParamJPanels() {
	    //talk("hide all filter parameter panels to become invisible");
		  for(int row=0;row<npaths;row++) 
			for(int col=0;col<nsteps;col++) 
					if(nodes[row][col] != null)
						nodes[row][col].disappear();
	    repaintAll();
		}
		
	public void markActiveNodes() {
		int row,col;
		for(row=0;row<npaths;row++) // first mark everything inactive 
			for(col=0;col<nsteps;col++) 
				if(nodes[row][col] != null) {
					nodes[row][col].ifActive=false;
					}
		// now flag as active only the nodes needed to compute what's shown
			if(showRow<0 || showCol<0) return;
			 // at least this node is active.
			nodes[showRow][showCol].ifActive=true;
			boolean somethingChanged=true;
			while(somethingChanged) {
			  somethingChanged=false;
			  for(col=nsteps-1;col>=0;col--)
				for(row=0;row<npaths;row++)
					if(nodes[row][col] != null && nodes[row][col].backPropActiveStatus())
							somethingChanged=true;
					// backPropActiveStatus() returns true only if a node was flagged as active.
			  }
			}
				
	public void registerLastClickedPic(int towerNum, int picNum) {
		lastClickedPicTower = towerNum;
		lastClickedPicNumber = picNum;
		}
		
	public boolean lastClickedPicClickedAgainQ(int towerNum, int picNum) {
		if(lastClickedPicTower != towerNum) return false;
		if(lastClickedPicNumber != picNum) return false;
		return true;
		}
		
	public void unRegisterLastClickedPic(int iwh) {
		lastClickedPicTower = -1;
		lastClickedPicNumber = -1;
		}
		
	public void saveTheMaskFIle() {
	    if(a3dmask != null && making3Dmask > 0) {
	    	talk("saving the  "+maskKeyWord[making3Dmask]+" mask.");
	    	a3dmask.writeBitMaskToDisk();
	    	}
	    else talk("Segmenter.saveTheMaskFIle() told to save null BitMask");
		}
		
	private String XYZswapName(String currentFileName,String newXYZ) {
		if(currentFileName==null) return "";
		int swapIndex = currentFileName.lastIndexOf("XY");		
	    if(swapIndex<0)
	    	swapIndex = currentFileName.lastIndexOf("XZ");		
	    if(swapIndex<0)
	    	swapIndex = currentFileName.lastIndexOf("ZY");
	    if(swapIndex<0) 
	    	return currentFileName;
	    return (new String((currentFileName.substring(0,swapIndex)+newXYZ+
	    		currentFileName.substring(swapIndex+2))));
	    }

	public boolean loadNewVantageStack(int whichOfXorYoZ) {
		String currentRedFileName=null,currentGreenFileName=null,currentBlueFileName=null;
		String newXYZ=null,newRedFileName=null,newGreenFileName=null,newBlueFileName=null;
		talk("loadNewVantageStack("+whichOfXorYoZ+") with current maskMakerXorYorZ= "+
					maskMakerXorYorZ);
		for(int i=0;i<3;i++)
			parent.invalidateRasterCaches(i);
	    if(parent.curStack.redfile != null) 
	    	currentRedFileName = new String(parent.curStack.redfile.filename);
	    if(parent.curStack.greenfile != null) 
	    	currentGreenFileName = new String(parent.curStack.greenfile.filename);
	    if(parent.curStack.bluefile != null) 
	    	currentBlueFileName = new String(parent.curStack.bluefile.filename);
	    if(currentRedFileName == null && currentGreenFileName == null && currentBlueFileName == null) {
			maskMakerXorYorZ = -1;
			talk("================= currentFIleName==null=============");
			return false;
	    	}
	    switch (whichOfXorYoZ) { 
	    	case 1:
	    		newXYZ = "XY";
	    		break;
	    	case 2:
	    		newXYZ = "XZ";
	    		break;
	    	case 3:
	    		newXYZ = "ZY";
	    		break;
	    	default :
	    		newXYZ = null;
	    		break;
	    	}
	    if(newXYZ==null) {
	    	talk("impossible error in loadNewVantageStack()");
	    	return false;
	    	}
	    // insert newXYZ string into the current file name.
	    newRedFileName = XYZswapName(currentRedFileName,newXYZ);
	    newGreenFileName = XYZswapName(currentGreenFileName,newXYZ);
	    newBlueFileName = XYZswapName(currentBlueFileName,newXYZ);
	    if(newRedFileName == null && newGreenFileName == null && newBlueFileName == null) {
	    	talk("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$");
	    	talk("$$$$$$$$$$$$$$$$$$$$$$in loadNewVantageStack(), currentFIleNames = "+currentRedFileName+
	    	", "+currentGreenFileName+", "+currentBlueFileName);
	    	talk("in which there is no occurance of XY or XZ or ZY");
	    	talk("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$");
			maskMakerXorYorZ = -1;
			return false;
	    	}
		maskMakerXorYorZ = whichOfXorYoZ;
	    talk("In loadNewVantageStack, newFileNames = "+currentRedFileName+
	    	", "+currentGreenFileName+", "+currentBlueFileName+", and now maskMakerXorYorZ="+
	    			maskMakerXorYorZ);		
	    switch (making3Dmask) { // get the current file name
	    	case (1) : // doing the red channel
	    	case (4) : // doing the red channel
	    		talk("Segmenter.loadNewVantageStack(), making3Dmask="+making3Dmask+", whichOfXorYoZ="+
	    			whichOfXorYoZ+" loads newRedFileName="+newRedFileName);
				parent.getNewStacks(newRedFileName,newGreenFileName,newBlueFileName,
					parent.parent.navigatorJPanel,making3Dmask-1,newRedFileName);
 				break;
	    	case (2) : // doing the green channel
	    	case (5) : // doing the green channel
	    		talk("Segmenter.loadNewVantageStack(), making3Dmask="+making3Dmask+", whichOfXorYoZ="+
	    			whichOfXorYoZ+" loads newGreenFileName="+newGreenFileName);
			parent.getNewStacks(newRedFileName,newGreenFileName,newBlueFileName,
					parent.parent.navigatorJPanel,making3Dmask-1,newGreenFileName);
 				break;
	    	case (3) : // doing the blue channel
	    	case (6) : // doing the blue channel
	    		talk("Segmenter.loadNewVantageStack(), making3Dmask="+making3Dmask+", whichOfXorYoZ="+
	    			whichOfXorYoZ+" loads newBlueFileName="+newBlueFileName);
				parent.getNewStacks(newRedFileName,newGreenFileName,newBlueFileName,
					parent.parent.navigatorJPanel,making3Dmask-1,newBlueFileName);
 				break;
 			}
 		parent.gotosection(0);
 		try{Thread.sleep(200);}
 		catch(InterruptedException e) {
 			;}
 					
  		return true;
		}
		
	public boolean startMaking3Dmask(MakeA3Dmask maskMakerTrigger) { 
		/* returns true if mask making was actually initiated. This is called only by a
		   MakeA3Dmask() object when the "go button" is pushed. */
		int i;
		int iwh = maskMakerTrigger.making3Dmask;
		// talk("in Segmenter.startMaking3Dmask(), maskMakerTrigger.making3Dmask = "+iwh);
	    if(making3Dmask > 0) {
	    	talk("We're currently "+maskMakerTrigger.maskChoices[making3Dmask-1]+". Stop that before making another");
	    	return false;
	    	}
	    talk("Segementer starts "+maskMakerTrigger.maskChoices[iwh-1]+".");
		numXYmaskSects=-1;
		numXZmaskSects=-1;
		numZYmaskSects=-1;
		switch (iwh) {
			default :
				break;
			case (1) :
			case (4) :
				if(parent.curStack.redfile != null) {
					numXYmaskSects=(parent.curStack.redfile).num_pics;
					numXZmaskSects=(parent.curStack.redfile).height;
					numZYmaskSects=(parent.curStack.redfile).width;
					showMaskNode(1);
					}
				break;				
			case (2) :
			case (5) :
				if(parent.curStack.greenfile != null) {
					numXYmaskSects=(parent.curStack.greenfile).num_pics;
					numXZmaskSects=(parent.curStack.greenfile).height;
					numZYmaskSects=(parent.curStack.greenfile).width;
					showMaskNode(2);
					}
				break;				
			case (3) :
			case (6) :
				if(parent.curStack.bluefile != null) {
					numXYmaskSects=(parent.curStack.bluefile).num_pics;
					numXZmaskSects=(parent.curStack.bluefile).height;
					numZYmaskSects=(parent.curStack.bluefile).width;
					showMaskNode(3);
					}
				break;
			}
			
		if(numXYmaskSects < 0) {
			talk("Biorad input file for "+maskMakerTrigger.maskChoices[iwh-1]+" is null. Mask making canceled!!!");
			return false;
			}
		stopVaryingParams();
		if(a3dmask != null) {
			talk("Segmenter purging old a3dmask");
			a3dmask.sepuku();
			}
		if(iwh>3) // modify existing mask from disk
			a3dmask = new BitMask(numZYmaskSects, numXZmaskSects,  numXYmaskSects, true);
		else 	// make a brand new mask		
			a3dmask = new BitMask(numZYmaskSects, numXZmaskSects,  numXYmaskSects, false);
		if(a3dmask==null)
			talk("OOPS> Segmenter failed to make new BitMask");
		if(!maskMakerTrigger.ifMakeXYsects) numXYmaskSects=-1;
		if(!maskMakerTrigger.ifMakeXZsects) numXZmaskSects=-1;
		if(!maskMakerTrigger.ifMakeZYsects) numZYmaskSects=-1;
	    making3Dmask = 0; // it gets set to iwh only if some sections are to be processed
		if(numXYmaskSects>0 || numXZmaskSects>0 || numZYmaskSects>0) 
	    	making3Dmask = iwh;
			maskMaker=null;
		if(making3Dmask > 0) {
			parent.ifMakingMask=true;
			setPauseButton(0);
			maskPasses = 20;
			if(runChainThread == null) 
				runChainThread = new Thread(this);
			runChainThread.start();
	    	return true;
			}
		else {
			talk("You must process either the XY, and/or the XZ, and/or the ZY sections,");
			talk("else no mask is made");
			setPauseButton(-1);
			return false;
			}
		}
	
	private void setPauseButton(int setTo) {
		switch (setTo) {
			case 1 :
				pauseButton.setText("continue 3D mask making");
				maskMakingPaused=true;
				break;
				
			case 0 :
				pauseButton.setText("pause 3D mask making");
				maskMakingPaused=false;
				break;
			
			default :
				pauseButton.setText("make a 3D mask");
				maskMakingPaused=false;
				break;
			}				
		}
			
	public void stopMaking3Dmask() {
		talk("Save the mask file");
	    saveTheMaskFIle();
		talk("restore the XY sections");
		loadNewVantageStack(1);
	    making3Dmask = 0;
		parent.ifMakingMask=false;
	    talk("Stopped making the 3D mask and saved the file");		
		a3dmask.sepuku();
		a3dmask=null;
		numXYmaskSects=-1;
		numXZmaskSects=-1;
		numZYmaskSects=-1;
		maskMaker=null;
		setPauseButton(-1);
		talk("Kill the run thread");
		if(runChainThread != null) {
			runChainThread.stop();
			runChainThread=null;
			}
		}		
			
	public void actionPerformed(java.awt.event.ActionEvent event) {
		String arg = event.getActionCommand();
		
	    if(arg.equals("make a 3D mask")) {
	    	if(making3Dmask > 0) {
	    		talk("We're currently making the "+maskKeyWord[making3Dmask]+" mask. Stop that before making another");
	    		return;
	    		}
	    	talk("Start making a 3D mask.");
			if(maskMaker !=null)
				maskMaker.getAttention();
	    	else 
	    		maskMaker = new MakeA3Dmask(this); // when user pushes the "go" button, this calls startMaking3Dmask
	    							// and then mask making actually starts.
			}
			
	    if(arg.equals("stop Making 3D  mask")) {
	    	if(making3Dmask > 0) 
	    		stopMaking3Dmask();
	    	else
	    		talk("No 3D mask making currently in progress.");
	    	return;
			}
		
	    if(arg.equals("continue 3D mask making")) {
			setPauseButton(0);
	    	return;
	    	}
		
	    if(arg.equals("pause 3D mask making")) {
			setPauseButton(1);
			saveTheMaskFIle();
	    	return;
	    	}
		
	    if(making3Dmask > 0) {
	    	talk("Bulletin!");
	    	talk("You are currently making a 3D mask. All other buttons are deactivated");
	    	talk("until the 3D mask is finished or until you stop that process");
	    	return;
	    	}
	    	
	    if(arg.equals("write segmentation file")) {
	    	ArchiveMarks disker = new ArchiveMarks();
			disker.readWriteSegmenterFile(true,this);
			disker=null; 
		  	return;
		  	}

	    if(arg.equals("read segmentation file")) {
	    	ArchiveMarks disker = new ArchiveMarks();
			disker.readWriteSegmenterFile(false,this);
			disker=null; 
		  	return;
		  	}

	    if(arg.equals("more postage stamps")) {
	    	makeAnotherSegTower();   
		  	return;
			}

	    if(arg.equals("fewer postage stamps")) {
	    	eliminateLastSegTower();   
		  	return;
			}

	    if(arg.equals("make stamps larger")) {
	    	changeStampSizes(true);   
		  	return;
			}

	    if(arg.equals("make stamps smaller")) {
	    	changeStampSizes(false);   
		  	return;
			}
		
	    if(arg.equals("(toggle) training filters")) {
	    	PutDataThroughFilterChain(true);
		  	return;
			}
		
	    if(arg.equals("(toggle)segmentFor3DRecon")) {
	    	PutDataThroughFilterChain(false);
		  	return;
			}
		
	    if(arg.equals("re-order postage stamps")) {
	    	reOrderStampsInTowers();   
			unRegisterLastClickedPic(6);
		  	return;
			}


	    if(arg.equals("new sect; rerun all pics")) {
			unRegisterLastClickedPic(7);
			recomputeAllPics();
    		}
    		
	    	if(arg.equals("Run the filter chain")) {
	    		// temporarily turn off ifVaryingParameters;
	    		boolean tempIfVary = ifVaryingParameters;
				ifVaryingParameters=false;
	    		runTheChain();
				if(!ifUseFilterChainsInFrontOfMainDisplay)
					unRegisterLastClickedPic(8);
				//restore the ifParamVary state
				ifVaryingParameters=tempIfVary;
				// this is not a perfect way to do this because the run thread may be making
				// ongoing calls to runTheChain() and these will do good only when
				// ifVaryingParameters == true;
				}
    			
    			
	    	if(arg.equals("edit all parameters")) {
		  		for(int row=0;row<npaths;row++) 
					for(int col=0;col<nsteps;col++) 
					if(nodes[row][col] != null && nodes[row][col].whatAmI>0 && 
							nodes[row][col].whatAmI != nodes[row][col].nullID)
						nodes[row][col].appear();
				unRegisterLastClickedPic(9);
	    		repaintAll();
    			}
    			
	    	if(arg.equals("add new path (row)")) {
	    		talk("add new path (row)");
	    		addNewNodePath();
				repaintAll();
				unRegisterLastClickedPic(10);
    			}
    			
	    	if(arg.equals("vary parameters")) {
				if(runChainThread == null) 
					runChainThread = new Thread(this);
				ifVaryingParameters=true;
				runChainThread.start();
				varyParsJButton.setText("stop varying Parameters");
				unRegisterLastClickedPic(11);
    			}
	    	if(arg.equals("stop varying Parameters")) {
	    		stopVaryingParams();
    			}
    			
 		if(arg.equals("re-order stamps")) {
 			 reOrderStampsInTowers();
 			 return;
 			 }   

	    if(arg.equals("click help")) {
			talk(""); 
			talk("Double click in upper left of any stamp to promote to bestYet.");
			talk("Double click in upper right corner of any stamp to promote to contender");
			talk("Double click in lower left corner of any stamp to demote an existing contender");
			talk("Double click in lower right corner of any stamp to dismiss the picture.");
			talk("");
	    	return;
	    	}

	    if(arg.equals("becomes contender")) {
	    	if(lastClickedPicTower >= 0 && lastClickedPicTower < nTowerCanvases  && lastClickedPicNumber >=0) 
	    		towers[lastClickedPicTower].canvas.promoteContender(lastClickedPicNumber);
	    	return;
	    	}
	    	
	    if(arg.equals("emptyAllButBestYet")) {
	    	dismissAllBut(3);
	    	reOrderStampsInTowers();   
	    	return;
	    	}

	    if(arg.equals("emptyAllButContenders")) {
	    	dismissAllBut(2);
	    	reOrderStampsInTowers();   
	    	return;
	    	}

   		return;
    	}

    private void stopVaryingParams() {
		ifVaryingParameters=false;
		if(runChainThread != null) {
			runChainThread.stop();
			runChainThread=null;
			}
		varyParsJButton.setText("vary parameters");
		unRegisterLastClickedPic(12);
    	}
    		
	private void addNewNodePath() {
		int row,col,whatAmI;
		SegmenterNode [][] oldNodes =  nodes;
		nodes = new SegmenterNode[npaths+1][nsteps];
		for(row=0;row<npaths;row++) // copy the old array into the new array
			for(col=0;col<nsteps;col++) 
				nodes[row][col] = oldNodes[row][col];				
		++npaths;
		setWinSize();
		row=npaths-1;
		// talk("when creating new path, xinc="+xinc+" and yinc="+yinc);
		for(col=0;col<nsteps;col++) { // add the new row of nodes...
			whatAmI = 0;
			if(col==0)
				 whatAmI = 15; // zGrad starts the chain 
			if(col==nsteps-1)
				 whatAmI = 5; // BlkWht filter ends the chain 
			nodes[row][col] = new SegmenterNode(this,row,col,whatAmI,(1+col)*xinc, yinc/2+(row-1)*yinc, 
						(4*xinc)/5, (4*yinc)/5, -1, -1, -2, -2, false);
			}
		oldNodes=null;
		newConfig();
		hideAllParamJPanels();
		}

	public void replaceNode(SegmenterNode nd, int whatKind) {
		int row = nd.myPath, col=nd.myStep;
		if(row>0) {
		  nodes[row][col] = new SegmenterNode(this, row, col, whatKind, (1+col)*xinc, yinc/2+(row-1)*yinc, 
						(4*xinc)/5, (4*yinc)/5, nd.inPath1, nd.inStep1, nd.inPath2, nd.inStep2, true);
		  nd.sepuku();
		  repaintAll();
		  }
		}
		
static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
		
	// catch WindowListener events  
	public void windowActivated(java.awt.event.WindowEvent event) {
		// talk("window activated");
		}
	public void windowDeactivated(java.awt.event.WindowEvent event) {
		//  talk("window Deactivated");
		}
	public void windowOpened(java.awt.event.WindowEvent event) {}
	public void windowIconified(java.awt.event.WindowEvent event) {}
	public void windowDeiconified(java.awt.event.WindowEvent event) {}
	public void windowClosing(java.awt.event.WindowEvent event) {
	    /* setVisible(false);
	    dispose(); */
		}
	public void windowClosed(java.awt.event.WindowEvent event) {
		}

	public void update(Graphics2D g) {
		paint(g);
		}
		
	public void markAllNodesChanged() {
		for(int row=0;row<npaths;row++) 
			for(int col=0;col<nsteps;col++)
				if(nodes[row][col] != null)
					nodes[row][col].anyChangeSinceLastRun = true;
		}
			
	public void drawNodes(GraphicsPs gg) {
		// draw the connections between nodes
		gg.newColor(255,255,0);
		int torow,tocol;
		markActiveNodes();
		for(int row=0;row<npaths;row++) 
			for(int col=0;col<nsteps;col++) {
				if(nodes[row][col].inPath1 >= 0 && nodes[row][col].inStep1 >= 0)  {
					nodes[row][col].setTextColor(gg);
					torow=nodes[row][col].inPath1;
					tocol=nodes[row][col].inStep1;
					gg.drawLine(nodes[row][col].myXcen,nodes[row][col].myYcen,
						nodes[torow][tocol].myXcen,nodes[torow][tocol].myYcen);
					}
				if(nodes[row][col].inPath2 >= 0 && nodes[row][col].inStep2 >= 0)  {
					nodes[row][col].setTextColorDim(gg);
					torow=nodes[row][col].inPath2;
					tocol=nodes[row][col].inStep2;
					gg.drawLine(nodes[row][col].myXcen,nodes[row][col].myYcen,
						nodes[torow][tocol].myXcen,nodes[torow][tocol].myYcen);
					}
				}
		// draw the nodes (which erase some parts of the connections just drawn
		for(int row=0;row<npaths;row++) 
			for(int col=0;col<nsteps;col++) {
				if(nodes[row][col].whatAmI>=0) 
					nodes[row][col].drawNode(gg);
				}
		if(showRow>=0 && showCol>=0 && nodes[showRow][showCol] != null)
			nodes[showRow][showCol].flagShowNode(gg); 
		}
	
	public void cutLinksTo(int thePath,int theStep) {
		int row,col;
		if(showRow==thePath && showCol==theStep) setShowNode(-1,-1);
		for(row=0;row<npaths;row++) 
			for(col=0;col<nsteps;col++) { 
			 	if(nodes[row][col].inPath1==thePath && nodes[row][col].inStep1 == theStep) 
			 		nodes[row][col].eraseInput1Links();
			 	if(nodes[row][col].inPath2== thePath && nodes[row][col].inStep2 == theStep) 
			 		nodes[row][col].eraseInput2Links();
			 	}			   
		}
		
	public void whatDidMouseHit(int xm, int ym, int [] nodeRef) {
		/* on return, 
		  nodeRef[0] = 1 means (xm,ym) is in the chan1InputImage 
					 = 2 means (xm,ym) is in the chan2InputImage 
					 = 3 means (xm,ym) is in the chan3InputImage 
					 = 4 means (xm,ym) is in the maskOutput node
					 = 5 means (xm,ym) is in the "show" box at upper right
					 = 0 means (xm,ym) is in one of the filter nodes
	When nodeRef[0]==0, then nodeRef[1] has the row=path of the node and 
			nodeRef[2] has the col=step of the node  */ 
			return;
		}
		
	public boolean linkTwoNodesQ(int startx, int starty, int endx, int endy) {
		// if (endx,endy) hits a node to the left of a node hit by (startx,starty) then link them
		// provided the right node has a currently available input.
		int row,col,startPath=-1, startStep=-1, endPath=-1,endStep=-1,input;
		for(row=0;row<npaths;row++)
			for(col=0;col<nsteps;col++)
				if(nodes[row][col] != null) {
					if(nodes[row][col].mouseHitMeQ(startx,starty)) {
						startPath=row;
						startStep=col;
						}
					if(nodes[row][col].mouseHitMeQ(endx,endy)) {
						endPath=row;
						endStep=col;
						}
					}
					
		if(startPath<0 || startStep<0) {
			talk("the link did not start in any node");
				return false;
				}
			
		if(endPath<0 || endStep<0) {
			talk("the link did not end in any node");
				return false;
				}
				
		if(startPath==endPath && startStep==endStep) { // link began and ended in the same node
			if(startPath==0 && (startStep==1 || startStep==2 || startStep==3)) { 
				showNodeParameters(startPath,startStep);
				return false;
				}
			
			else if((Math.abs(startx-endx) + Math.abs(starty-endy)) > 10) {
				talk("erasing input links because you dragged across a single node");
				nodes[startPath][startStep].eraseInputLinks();// erase input links
				repaintAll();
				return false;
				}
			showNodeParameters(startPath,startStep);
			return false;
			}

		if(nodes[startPath][startStep].whatAmI==nullID) { // link started in a null node
			talk("path started in a null node. No link possible");
			return false;
			}
			
		if(nodes[endPath][endStep].whatAmI==nullID) { // ended in a null node
			talk("path ended in a null node. No link possible");
			return false;
			}
			
		if(startPath==0) { // mouse down was in one of the special nodes and the link ended in another node
			if(startStep==1 || startStep==2 || startStep==3) { 
				talk("the input to the channel input is fixed.");
				return false;
				}
			if(nodes[endPath][endStep].whatAmI == nullID) { // ended in a null node
					talk("path ended in a null node. No link possible");
					return false;
					}
			if((input=nodes[startPath][startStep].availableInput()) <0) {
				talk("mouseDrag started in a node with no more inputs available");
				return false;
				}				
			// OK we have a new link from a node with available input:
			nodes[startPath][startStep].anyChangeSinceLastRun=true;
			if(input==1) {
				nodes[startPath][startStep].inPath1 = endPath;
				nodes[startPath][startStep].inStep1 = endStep;
				}
			else {
				nodes[startPath][startStep].inPath2 = endPath;
				nodes[startPath][startStep].inStep2 = endStep;
				}
			repaintAll();
			return true;
			}
			
		// if we fall through to here the start node is a filter node and the end node is another node
		// which is not a null node		
		if(endPath==0 && (endStep==0 || endStep>3)) {// disallow normal nodes from using the mask nodes as input
			talk("You cannot use the mask nodes as input to anything except the show node. endStep="+endStep);
			talk("endPath="+endPath+"  whatAmI="+nodes[endPath][endStep].whatAmI);
			return false;
			}

		if((input=nodes[startPath][startStep].availableInput()) <0) {
				talk("mouseDrag started in a node without available inputs");
				return false;
				}
				
		// OK (startx,starty) was in a node with an available input...
		// enforce left-to-right flow
		if(endStep >= startStep && nodes[endPath][endStep].imageInputNode < 0) {
			talk("flow must be from left to right");
				return false;
			}
					
  		// OK make a new link... We know if we get to here, input==1 or input==2
		nodes[startPath][startStep].anyChangeSinceLastRun=true;
		if(input==1) {
			nodes[startPath][startStep].inPath1 = endPath;
			nodes[startPath][startStep].inStep1 = endStep;
			}
		else {
			nodes[startPath][startStep].inPath2 = endPath;
			nodes[startPath][startStep].inStep2 = endStep;
			}
		repaintAll();
		return true;
		}
		
	public void showNodeParameters(int thePath,int theStep)	 {
		if(nodes[thePath][theStep] != null) 
			nodes[thePath][theStep].appear();
		repaintAll();
		}
		
	public void setFillHeight() {
		if(nodes[npaths-1][0] != null)
			fillHeight=nodes[npaths-1][0].myYcen + 160;
		}
		
	public void displayParJPanels() {
		// talk("Segmenter.displayParJPanels()");
		int myx=getLocation().x;
		int myy=getLocation().y;
		for(int row=0;row<npaths;row++) 
			for(int col=0;col<nsteps;col++) 
				if(nodes[row][col] != null) {
					nodes[row][col].display(myx,myy);
					nodes[row][col].repaint();
					}
			}	
	
	public void archiveSegmentationData(ArchiveMarks disker, boolean ifWrite) {
		int i,whatAmI,row=1,col=0,j;
		double badval = 7777.777;
		int [] ints12 = new int[12];
		String nodeName;
		if(ifWrite) { // write segmentation info to disk
			disker.writeNum(npaths," = npaths");
			disker.writeNum(nsteps," = nsteps");
			readAllParameterValues();
			for(row=0;row<npaths;row++)
				for(col=0;col<nsteps;col++) {
						nodeName=nodes[row][col].encodeInto12Ints(ints12);
						disker.write12Ints(ints12,nodeName);
						}
			disker.writeAString("path step whatAmI inPath1 inStep1 inPath2 inStep2 par0 par1 par2 par3");
			return;
			}
		else { // recover segmentation data from disk and rebuild segmentation chain 
			int np = (int)disker.getNextFileDouble(badval);
			int ns = (int)disker.getNextFileDouble(badval);
			talk("Rebuilding a filter chain with "+npaths+" paths and "+nsteps+" steps");
			if(np<2 || np>12) {
				talk("archiveSegmentationData() read npaths="+np+", which is outside legal bounds");
				return;
				}
			if(ns<7 || ns>12) {
				talk("archiveSegmentationData() read nsteps="+ns+", which is outside legal bounds");
				return;
				}
			int [][][] allThePars = new int[np][ns][12];
			for(i=0;i<np;i++)
				for(j=0;j<ns;j++)
					allThePars[i][j][2]=-1; // this is the whatAmI and a negative parameter flags a non-existent node
			// there should be a line in the file for each of the 7 + (npaths-1)*nsteps nodes to be reconstituted
			int n2read = np*ns;
			for(j=0;j<n2read;j++) {	
			  disker.read12Ints(ints12);
			  if(!valid12ints(ints12,np,ns)) {
				talk("got bad line of input in archiveSegmentationData. No changes made");
				return;
				}
			  else {
			  	// talk12ints(ints12);
				row=ints12[0];
				col=ints12[1];
				for(i=0;i<12;i++)
					allThePars[row][col][i] = ints12[i];
				}
			  }
			rebuildWholeChain(np,ns,allThePars);
			}
		}
		
	public boolean compatibleChainTopology(int np,int ns,int [][][] allPars)  {
		int row,col;
		if(np > npaths || ns > nsteps)  return false;
		for(row=0;row<np;row++)
			for(col=0;col<ns;col++) {
				if(nodes[row][col].whatAmI != allPars[row][col][2])  {
					//talk("compatibleChainTopology returns false because of a whatAmI mismatch.");
					return false;
					}					
				if(nodes[row][col].inPath1 != allPars[row][col][3]) {
					//talk("compatibleChainTopology returns false because of an inPath1 mismatch.");
					return false;
					}					
				if(nodes[row][col].inStep1 != allPars[row][col][4]) {
					//talk("compatibleChainTopology returns false because of an inStep1 mismatch.");
					return false;
					}									
				if(nodes[row][col].inPath2 != allPars[row][col][5]) {
					//talk("compatibleChainTopology returns false because of an inPath2 mismatch.");
					return false;
					}					
				if(nodes[row][col].inStep2 != allPars[row][col][6]) {
					//talk("compatibleChainTopology returns false because of an inStep2 mismatch.");
					return false;
					}									
//  0     1     2       3      4       5       6         7      8       9      10     11
// path step whatAmI inPath1 inStep1 inPath2 inStep2   par0    par1    par2    par3  extra
				} 
		//talk("compatibleChainTopology returns true");
		return true;
		}
				
	public void rebuildWholeChain(int np, int ns, int [][][]allPars) {
		int i,j,whatAmI;
		int MaskColorsData=2;
		if(showRow >= 0 && showCol >= 0)
			MaskColorsData=nodes[showRow][showCol].MaskColorsData;
		if(!compatibleChainTopology(np,ns,allPars)) { // make a new chain
			talk("make a whole new chain");
			SegmenterNode [][] oldNodes = nodes;
			npaths = np;
			nsteps = ns;
			nodes = new SegmenterNode[npaths][nsteps];
			// there should be a line in the file for each of the 7 + (npaths-1)*nsteps nodes to be reconstituted
			for(i=0;i<npaths;i++) 
				for(j=0;j<nsteps;j++) {				
					if(i != allPars[i][j][0]  || j != allPars[i][j][1])
						talk("warning, bad input file with rows and columns out of place...");
					nodes[i][j] = new SegmenterNode(this,i,j,allPars[i][j][2],(1+j)*xinc, yinc/2+(i-1)*yinc, 
						(4*xinc)/5, (4*yinc)/5, allPars[i][j][3], allPars[i][j][4], allPars[i][j][5], allPars[i][j][6], false);
				nodes[i][j].decodeParamValues(allPars[i][j][7],allPars[i][j][8],allPars[i][j][9],allPars[i][j][10]);
				}
			oldNodes=null;
			  }
		else { // load new parameter values into the existing nodes...
			for(i=0;i<npaths;i++) 
				for(j=0;j<nsteps;j++) 	
					nodes[i][j].decodeParamValues(allPars[i][j][7],allPars[i][j][8],allPars[i][j][9],allPars[i][j][10]);
				}

		if(showRow>=0 && showCol>=0 && showCol<nsteps && showRow<npaths) 
			nodes[showRow][showCol].MaskColorsData = MaskColorsData;
		else {showRow=-1; showCol=-1;}
							
		setWinSize();
		newConfig();
		hideAllParamJPanels();
		}
				
	private void talk12ints(int [] aray) {
		String stg = "";
		for(int i=0;i<12;i++)
			stg = stg +" "+aray[i];
		talk(stg);
		}
		
	private void readAllParameterValues() {
		for(int row=1;row<npaths;row++)
			for(int col=0;col<nsteps;col++)
				nodes[row][col].readParameterValues();
		}
		
	private boolean valid12ints(int [] aray, int nrows, int ncols) {
//  0     1     2       3      4       5       6         7      8       9      10     11
// path step whatAmI inPath1 inStep1 inPath2 inStep2   par0    par1    par2    par3  extra
		if(aray[0] <0 || aray[0] >= ncols) return false;
		if(aray[1] <0 || aray[1] >= ncols) return false;
		if(aray[2] >= possibilities) return false;
		if(aray[3] >= nrows ) return false;
		if(aray[4] >= ncols ) return false;
		if(aray[5] >= nrows ) return false;
		if(aray[6] >= ncols ) return false;
		return true;
		}
		
	private void blank(short [] aras) {
		if(aras==null) return;
		int np = aras.length;
		for(int i=0;i<np;i++) 
			aras[i]=0;
		}	

	private void segmentNewPicToMask() {
		int XYZSECT=-1,sectNum,i;
		sectNum=-1;
		talk("segmentNewPicToMask()");
		if(numXYmaskSects > 0) { // do the XY sectons first
			i=0;
			while(XYZSECT<0 && i<numXYmaskSects) {
				if(a3dmask.toDoXYsects[i]) {
					sectNum = i;
					XYZSECT=XYSECT;
					}
				i++;
				}
			}
			
		if(sectNum <0 && numXZmaskSects > 0) {  // do the XZ sections second
			i=0;
			while(XYZSECT<0 && i<numXZmaskSects) {
				if(a3dmask.toDoXZsects[i]) {
					sectNum = i;
					XYZSECT=XZSECT;
					}
				i++;
				}
			}
			
		if(sectNum <0 && numZYmaskSects > 0) { // do the ZY sections third
			i=0;
			while(XYZSECT<0 && i<numZYmaskSects) {
				if(a3dmask.toDoZYsects[i]) {
					sectNum = i;
					XYZSECT=ZYSECT;
					}
				i++;
				}
			}

		if(XYZSECT < 0) { // found no section that needed processing
				talk("finished making the 3D "+maskKeyWord[making3Dmask]);
				talk("Now restore the XY section view...");
				stopMaking3Dmask();
				return;
				}
		// if fall through, there is a section to process	
		talk("*****************segmentNewPicToMask: XYZSECT="+XYZSECT+", sectNum="+sectNum);

		markAllNodesChanged();
		switch (XYZSECT) {
			case (XYSECT) :
				if(maskMakerXorYorZ != 1)
					loadNewVantageStack(1);
				talk("process XY section "+sectNum);
				parent.gotosection(sectNum);
				blank(the3DmaskNode.shortRas);	
		talk("segmentNewPicToMask: runTheChain for an XY section");
						
				runTheChain();  		
				noOneIsBest(); // this causes the SegPic just made by										
				currentPic.promoteToBestYet(); // runTheChain() to be
				setBestYet(currentPic); // promoted to best yet pic.
	    		dismissAllBut(3); // dismiss all but best yet. Keep only the current section 
	    				// in view.
	    		reOrderStampsInTowers(); 
	    		// or the 2D mask just made into the 3D mask
	    		 // this is the mask
				if(a3dmask != null)
					a3dmask.orShortTwoDimMask(the3DmaskNode.shortRas, 0, sectNum);
				else
					talk("Segmenter segmenting XYsection found a3dmask==null");
				a3dmask.toDoXYsects[sectNum] = false; // mark XY section as completed   		  
				break;
			case (XZSECT) :
				if(maskMakerXorYorZ != 2)
					loadNewVantageStack(2);
				talk("process XZ section "+sectNum);
				parent.gotosection(sectNum);
				blank(the3DmaskNode.shortRas);			
				runTheChain();  		
				noOneIsBest(); // this causes the SegPic just made by										
				currentPic.promoteToBestYet(); // runTheChain() to be
				setBestYet(currentPic); // promoted to best yet pic.
	    		dismissAllBut(3); // dismiss all but best yet. Keep only the current section 
	    				// in view.
	    		reOrderStampsInTowers(); 
	    		// or the 2D mask just made into the 3D mask
	    		 // this is the mask
				if(a3dmask != null)
					a3dmask.orShortTwoDimMask(the3DmaskNode.shortRas, 1, sectNum);
				else
					talk("Segmenter segmenting XYsection found a3dmask==null");
				a3dmask.toDoXZsects[sectNum] = false; // mark XZ section as completed   		  
				break;
			case (ZYSECT) :
				if(maskMakerXorYorZ != 3)
					loadNewVantageStack(3);
				talk("process ZY section "+sectNum);
				parent.gotosection(sectNum);
				blank(the3DmaskNode.shortRas);			
				runTheChain();  		
				noOneIsBest(); // this causes the SegPic just made by										
				currentPic.promoteToBestYet(); // runTheChain() to be
				setBestYet(currentPic); // promoted to best yet pic.
	    		dismissAllBut(3); // dismiss all but best yet. Keep only the current section 
	    				// in view.
	    		reOrderStampsInTowers(); 
	    		// or the 2D mask just made into the 3D mask
	    		 // this is the mask
				if(a3dmask != null)
					a3dmask.orShortTwoDimMask(the3DmaskNode.shortRas, 2, sectNum);
				else
					talk("Segmenter segmenting XYsection found a3dmask==null");
				a3dmask.toDoZYsects[sectNum] = false; // mark ZY section as completed   		  
				break;
			}
			
		if(--maskPasses <=0) {
			saveTheMaskFIle();
			maskPasses = 20;
			}
		}		
	
	public void run() { // this runsTheChain when the varyPars switch is set
				// or when we are making a 3-D mask.
		int sleepTime = 100;
	  while(true) {
		if(making3Dmask <= 0 && showRow>=0 && showCol>=0) { // show the pic in the next avaliable picture
	    	markActiveNodes();
			if(ifVaryingParameters && howManyParametersToVary()>0 &&
					(towers == null || leftmostTowerWithEmptyPic() >= 0) ) 
	    			runTheChain();
			sleepTime = 100;
	    	}

		if(making3Dmask > 0 ) { // making 3D mask...
			if(!maskMakingPaused) {
				sleepTime = 20;
				segmentNewPicToMask();
				}
			else {
				talk("3D mask making paused...");
				sleepTime = 4000;
				}
			}
			
		try {
			Thread.sleep(sleepTime);
			}
		catch (InterruptedException e) {
			;
			} 
	  }
	 }
	
	public void getReadyToSegmentNewSection(int sectionNumToProcess) {
		if(BioradCanvas.curJFrame != sectionNumToProcess)
			parent.gotosection(sectionNumToProcess);
		System.out.println("processPrimaryChannelData, sectNum="+sectionNumToProcess);	
		markAllNodesChanged();
		}
	
	public byte [] processPrimaryChannelData(int chanNumToProcess, double nonMaskFrac)  {
		if(!ifUseFilterChainsInFrontOfMainDisplay)
			return null;
		System.out.println("processPrimaryChannelData channel="+chanNumToProcess);
		
		int oldShowRow = showRow, oldShowCol = showCol;	
		stopVaryingParams();
		
		showRow = 0;
		/* red chan mask1 in row 0, col 4
		 * green chan mask2 in row 0, col 5
		 * blue chan mask3 in row 0, col 6   */
		 
		switch (chanNumToProcess) {
			case 0:
				showCol=4; // red chan
				break;
	
			case 1:
				showCol=5; // red chan
				break;
	
			case 2:
				showCol=6; // red chan
				break;
			}
		
		SegmenterNode showNode = nodes[showRow][showCol];
		runTheChain();
		byte [] retn = currentPic.getMaskedDataForOneChannel(showNode, nonMaskFrac);
		currentPic.status = -1;
		showRow = oldShowRow;
		showCol = oldShowCol;
		return retn;
		}

	public void recomputeAllPics() {
	if(showRow<0 || showCol<0 || towers == null || theBestPic==null) {// there's no pic to show
		talk("Select some node to show a picture of before doing this");
		return;
		}
	SegmenterNode showNode = nodes[showRow][showCol];
	stopVaryingParams();
	int row,col,i,j,nhigh = towers[0].canvas.nhigh,npth, nstp,itsStatus=0;
	SegPic aPic=null;
	int besti=-1,bestj=-1;
	for(i=0;i<nTowerCanvases;i++) 
		for(j=0;j<nhigh;j++) {
			aPic = towers[i].canvas.thePics[j];
			itsStatus = aPic.status;
			if(aPic != null && itsStatus>=0) { // it's a valid set of parameters.
				if(itsStatus ==3) {// it's the bestYet Pic
					besti=i;
					bestj=j;
					}
				npth = aPic.numPaths;
				nstp = aPic.numSteps;
				rebuildWholeChain(npth, nstp, aPic.allNodePars);
				runChainPart1();
				// collect all the parameters and topological structure of the filter chain just run.				
				int [][][] allNodePars = new int[npth][nstp][12];
	    		for(row=0;row<npth;row++)
	    			for(col=0;col<nstp;col++) 
	    				if(nodes[row][col] != null)
	    					nodes[row][col].encodeInto12Ints(allNodePars[row][col]);	
	    		showNode = nodes[showRow][showCol];		
				aPic.acceptPic(showNode,allNodePars,npth,nstp);
				if(itsStatus == 3)
					aPic.promoteToBestYet();
				else 
					aPic.status = itsStatus;						
				}
			}
				    
	if(besti>-0 && bestj>=0) {
			aPic = towers[besti].canvas.thePics[bestj];
			aPic.promoteToBestYet();
			}
	}
		
	private void runChainPart1() {
	    markActiveNodes();
	    ++calcGen;
	    int row,col;
		if(!ifUseFilterChainsInFrontOfMainDisplay && ifVaryingParameters && theBestPic != null) {
			rebuildWholeChain(theBestPic.numPaths, theBestPic.numSteps, 
							theBestPic.allNodePars);
			talk("varying "+howManyParametersToVary()+" parameters");
			}
								    			
	    for(row=0;row<npaths;row++) // unset all justRun flags.
	    	for(col=0;col<nsteps;col++)
	    			nodes[row][col].justRun=false;
	    for(col=1;col<4;col++)
	    	nodes[0][col].runTheFilter(calcGen); // make the input channels current
	    for(col=0;col<nsteps;col++) // flow is from left to right
	    	for(row=1;row<npaths;row++)
	    		nodes[row][col].runTheFilter(calcGen);
	    nodes[0][4].runTheFilter(calcGen); // make mask1 nodes current
	    nodes[0][5].runTheFilter(calcGen); // make mask2 nodes current
	    nodes[0][6].runTheFilter(calcGen); // make mask3 nodes current
	    nodes[0][0].runTheFilter(calcGen); // make mask4 nodes current
		}
	 		
	public boolean runTheChain() {
		runChainPart1();
	    int row,col;
	    int howManyParsVary = 	!ifUseFilterChainsInFrontOfMainDisplay? 0 : howManyParametersToVary();
	    talk("There are "+howManyParsVary+" parameters scheduled to be varied.");	    
		boolean ifMadeNewPic=false;
		if(showRow>=0 && showCol>=0) { // show the pic in the next available picture
			if(towers == null)
				makeShowTowerCanvasses();
			int id=leftmostTowerWithEmptyPic();
			SegmenterNode showNode = nodes[showRow][showCol];
			if(id>=0) { // there is an empty pic
				int ipic = towers[id].firstEmptyPic();
				currentPic = towers[id].canvas.thePics[ipic];
// collect all the parameters and topological structure of the filter chain just run.				
				int [][][] allNodePars = new int[npaths][nsteps][12];
	    		for(row=0;row<npaths;row++)
	    			for(col=0;col<nsteps;col++) 
	    				if(nodes[row][col] != null)
	    					nodes[row][col].encodeInto12Ints(allNodePars[row][col]);			
				currentPic.acceptPic(showNode,allNodePars,npaths,nsteps);
				ifMadeNewPic=true;
				if(!ifShowingBestPic) {
					setBestYet(currentPic);
					currentPic.promoteToBestYet();
					ifShowingBestPic=true;
					}
				}	
			}	    
	    repaintAll();
	    if(howManyParsVary<=0)
	    	ifMadeNewPic=false;
	    return ifMadeNewPic;
		}

}
