package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.awt.*;  
import javax.swing.*;

import java.awt.image.*;
import java.util.*;
import java.lang.*;
import java.io.*;

class Stereo3DReconstruction extends JFrame  implements Runnable, java.awt.event.ActionListener,java.awt.event.WindowListener {
	ShowPicJPanel picJPanel=null;
	int nomwidth=275, nomheight=400;
	JTextField fromSection, toSection, stereoSeparationField,kevinBaconField,numMovieFramesField,angleBetweenMovieFramesField,
		movieFrameScaleFactorField, progressReportField, blendFractionField, redTranspFIeld, greenTranspFIeld, blueTranspFIeld,
		   depthCueCloseField,depthCueFarField,maskFracRField, maskFracGField, maskFracBField;
	
	JTextField  [][] ChannelBoostFields = new JTextField[3][4];
	double [][] ChannelBoosts = new double[3][4];
	int stereoSeparation = 6,nmovieframes=30;
	double angleBetweenMovieFrames=1.25,movieFrameScaleFactor=.5,redNonMaskFrac=.15, greenNonMaskFrac=.15, blueNonMaskFrac=.15;
	JButton stopButton,doRedButton, doGreenButton, doBlueButton, blenderButton, crossEyeButton,
		ifSaveReverseOrderFramesButton;
	static JButton useSegmentationFiltersToggleButton;
	JButton [] doSegmentsButtons = new JButton[4];
	String [] doSegmentButtonsLabels = {"do seg  1","do seg 2","do seg 3","do seg4"};
	String [] omitSegmentButtonsLabels = {"omit seg  1","omit seg 2","omit seg 3","omit seg4"};
	boolean [] ifDoSegments = {true,false,false,false};
	int whichBlend = 0; // 0 means blend the max projection with the ray-traced projection
						// 1 means do only a maximum projection
						// 2 means do only a ray-traced projection
	int currentSegment = 0; // tells which of r possible movie segments we are currently making the movie or single stereo pair overlay for
	int from=0,  to=20;
	int oldFrom = -1;
	int oldTo = -1;
	boolean overlayIsValid=false;
	int [] overlayRas=null;
	int curWidth=0,  curHeight=0;
	double scaleFactor = 0.5;

	// needed by the run() method to make the frames of movie overlays	    	  
	boolean ifdoraytrace=true, ifdomaxproj=true, ifdored=true, ifdogreen=true, ifdoblue=true, ifSaveReverseOrderFrames = true;
	int depthCueBrightnessOfFromSection=255, depthCueBrightnessOfToSection=220;

	int [] redtransp = new int[256], greentransp = new int[256], bluetransp = new int[256]; 
	double redChannelBoost,greenChannelBoost,blueChannelBoost;
	int redSatur8Transp=230, greenSatur8Transp=230, blueSatur8Transp=230;
	double stereoSeparationPerSection;
	StereoPairFrame spf,spfPrevious=null;
	int [] rayTraceRaster,maxProjectionRaster;
	int frameNumber, framehi=0,frameOffset=0;
	String movieFileBaseName = "nullo";
	int stereoWidth;
	int [] leftMargins,rightMargins;
	boolean movieInProgress=false,ifMovieIsStereo=true, ifCrossEyes=true;
	String frameTitle;
	Thread movieThread=null; 
	double sixDegreesFromKevinBacon = .05, blendFractionOfMaxProjection=.4;
	static int addGap=20;
	int doingWhichSegment=0;
	Color segmentIsCurrentColor = new Color(255,255,128);
	Color brownback = new Color(225,150,75);

	public static void PutDataThroughFilterChain(boolean ifso) {
		if(ifso)
			useSegmentationFiltersToggleButton.setText("(toggle) 3D recons from segmentated data");
		else
			useSegmentationFiltersToggleButton.setText("(toggle) no segmentation in 3D recons");
		}


	public void archiveParameters(ArchiveMarks disker,boolean ifWrite) {
		int i,chan,segment;
		double badval = 7777.777;
		if(ifWrite) { // write parameters to disk
			getBoosts();
			getOtherParameters();
			// first do booleans
			disker.writeNum(ifdoraytrace,"  ifdoraytrace");
			disker.writeNum(ifdomaxproj, "  ifdomaxproj");
			disker.writeNum(ifdored, "  ifdored");
			disker.writeNum(ifdogreen, "  ifdogreen");
			disker.writeNum(ifdoblue, "  ifdoblue");
			disker.writeNum(ifSaveReverseOrderFrames,"  ifSaveReverseOrderFrames");
			for(segment=0;segment<4;segment++)
				disker.writeNum(ifDoSegments[segment], ("  ifDoSegments["+segment+"]"));
			// now do individual ints
			disker.writeNum(whichBlend, "  whichBlen=");
			disker.writeNum(currentSegment, " currentSegment ");
			disker.writeNum(from, "  from");
			disker.writeNum(to, "  to");
			disker.writeNum(stereoSeparation, "  stereoSeparation");
			disker.writeNum(nmovieframes, "  nmovieframes");
			disker.writeNum(depthCueBrightnessOfFromSection, "  depthCueBrightnessOfFromSection");
			disker.writeNum(depthCueBrightnessOfToSection, "  depthCueBrightnessOfToSection");
			disker.writeNum(redSatur8Transp, "  redSatur8Transp");
			disker.writeNum(greenSatur8Transp, "  greenSatur8Transp");
			disker.writeNum(blueSatur8Transp, "  blueSatur8Transp");

			disker.writeNum(redNonMaskFrac, "  redNonMaskFrac");
			disker.writeNum(greenNonMaskFrac, "  greenNonMaskFrac");
			disker.writeNum(blueNonMaskFrac, "  blueNonMaskFrac");
			
			// now do individual doubles
			disker.writeNum(scaleFactor, "  scaleFactor");
			disker.writeNum(angleBetweenMovieFrames, "  angleBetweenMovieFrames");
			disker.writeNum(movieFrameScaleFactor, "  movieFrameScaleFactor");
			disker.writeNum(sixDegreesFromKevinBacon, "  sixDegreesFromKevinBacon");
			disker.writeNum(blendFractionOfMaxProjection, "  blendFractionOfMaxProjection");
			// now write the array-> double [3][4] ChannelBoosts;			
			for(chan=0;chan<3;chan++)
				for(segment=0;segment<4;segment++)
					disker.writeNum(ChannelBoosts[chan][segment],("   ChannelBoosts["+chan+"]["+segment+"]"));					
			return;
			}
		else { // recover parameters from disk and set text fields 
			// now fetch booleans
			ifdoraytrace =(disker.getNextFileDouble(badval) > 0.5);     // retrieve a boolean
			ifdomaxproj = (disker.getNextFileDouble(badval) > 0.5);     // retrieve a boolean
			ifdored = (disker.getNextFileDouble(badval) > 0.5);     // retrieve a boolean
			ifdogreen = (disker.getNextFileDouble(badval) > 0.5);     // retrieve a boolean
			ifdoblue = (disker.getNextFileDouble(badval) > 0.5);     // retrieve a boolean
			ifSaveReverseOrderFrames = (disker.getNextFileDouble(badval) > 0.5);     // retrieve a boolean
			for(segment=0;segment<4;segment++)
				ifDoSegments[segment] = (disker.getNextFileDouble(badval) > 0.5);     // retrieve a boolean

			// now fetch individual ints
			whichBlend = (int)disker.getNextFileDouble(badval); // retrieve an int
			currentSegment = (int)disker.getNextFileDouble(badval); // retrieve an int
			from = (int)disker.getNextFileDouble(badval); // retrieve an int
			to = (int)disker.getNextFileDouble(badval); // retrieve an int
			stereoSeparation = (int)disker.getNextFileDouble(badval); // retrieve an int
			nmovieframes = (int)disker.getNextFileDouble(badval); // retrieve an int
			depthCueBrightnessOfFromSection = (int)disker.getNextFileDouble(badval); // retrieve an int
			depthCueBrightnessOfToSection = (int)disker.getNextFileDouble(badval); // retrieve an int
			redSatur8Transp = (int)disker.getNextFileDouble(badval); // retrieve an int
			greenSatur8Transp = (int)disker.getNextFileDouble(badval); // retrieve an int
			blueSatur8Transp = (int)disker.getNextFileDouble(badval); // retrieve an int
			redNonMaskFrac = disker.getNextFileDouble(badval); // retrieve a double
			greenNonMaskFrac = disker.getNextFileDouble(badval); // retrieve a double
			blueNonMaskFrac = disker.getNextFileDouble(badval); // retrieve a double
			// now fetch individual doubles
			scaleFactor = disker.getNextFileDouble(badval);     // retrieve a double
			angleBetweenMovieFrames = disker.getNextFileDouble(badval);     // retrieve a double
			movieFrameScaleFactor = disker.getNextFileDouble(badval);     // retrieve a double
			sixDegreesFromKevinBacon = disker.getNextFileDouble(badval);     // retrieve a double
			blendFractionOfMaxProjection = disker.getNextFileDouble(badval);     // retrieve a double

			// now fetch the array-> double [3][4] ChannelBoosts;			
			for(chan=0;chan<3;chan++)
				for(segment=0;segment<4;segment++)
					ChannelBoosts[chan][segment] = disker.getNextFileDouble(badval);  // retrieve a double
			setBoosts();
			setParameters(scaleFactor, blendFractionOfMaxProjection, ChannelBoosts[0][currentSegment], ChannelBoosts[1][currentSegment], ChannelBoosts[2][currentSegment]);
			adjustButtons();
			setOtherParameters();
			return;
			}					
		}
	
	private void setOtherParameters() {
		depthCueCloseField.setText(""+depthCueBrightnessOfFromSection);
		depthCueFarField.setText(""+depthCueBrightnessOfToSection);	   
		redTranspFIeld.setText(""+redSatur8Transp);	   
		greenTranspFIeld.setText(""+greenSatur8Transp);	   
		blueTranspFIeld.setText(""+blueSatur8Transp);	   
		maskFracRField.setText(""+redNonMaskFrac);
		maskFracGField.setText(""+greenNonMaskFrac);
		maskFracBField.setText(""+blueNonMaskFrac);
		angleBetweenMovieFramesField.setText(dbleSixString(angleBetweenMovieFrames));	   
		numMovieFramesField.setText(""+nmovieframes);	   
		stereoSeparationField.setText(""+stereoSeparation);	 
		kevinBaconField.setText(""+sixDegreesFromKevinBacon);
		toSection.setText(""+to);	   
		fromSection.setText(""+from);	   
		}		   
	
	private void getOtherParameters() {
		depthCueBrightnessOfToSection  = (int)(Double.valueOf(depthCueFarField.getText()).doubleValue());
		depthCueBrightnessOfFromSection  = (int)(Double.valueOf(depthCueCloseField.getText()).doubleValue());
		depthCueBrightnessOfToSection = inlimits(0,255,depthCueBrightnessOfToSection);
		depthCueBrightnessOfFromSection = inlimits(0,255,depthCueBrightnessOfFromSection);
		depthCueFarField.setText(""+depthCueBrightnessOfToSection);
		depthCueCloseField.setText(""+depthCueBrightnessOfFromSection);
	    from = inlimits(0,picJPanel.curStack.NJFrames-1,from);	
	    to = inlimits(0,picJPanel.curStack.NJFrames-1,to);
	    stereoSeparation = (int)(Double.valueOf(stereoSeparationField.getText()).doubleValue());
	    stereoSeparation = Math.abs(stereoSeparation);
		stereoSeparationField.setText(""+stereoSeparation);
		sixDegreesFromKevinBacon = Double.valueOf(kevinBaconField.getText()).doubleValue();
		kevinBaconField.setText(""+sixDegreesFromKevinBacon);
	    nmovieframes = (int)(Double.valueOf(numMovieFramesField.getText()).doubleValue());
	    angleBetweenMovieFrames = Double.valueOf(angleBetweenMovieFramesField.getText()).doubleValue();	    
		redSatur8Transp = (int)(Double.valueOf(redTranspFIeld.getText()).doubleValue());
		greenSatur8Transp = (int)(Double.valueOf(greenTranspFIeld.getText()).doubleValue());
		blueSatur8Transp = (int)(Double.valueOf(blueTranspFIeld.getText()).doubleValue());
		redNonMaskFrac = Double.valueOf(maskFracRField.getText()).doubleValue();
		greenNonMaskFrac = Double.valueOf(maskFracGField.getText()).doubleValue();
		blueNonMaskFrac = Double.valueOf(maskFracBField.getText()).doubleValue();
		}		   
	
	public void goToNextSegment() {
		if(++currentSegment > 3)
			currentSegment=0;
		getBoosts();
		adjustButtons();
		}
		
	public void progressReport(String stg) {
		if(stg != null) {
			progressReportField.setText(stg);
			return;
			}
		int i = (int)(4*Math.random()); 
		switch (i) {
			case 0:
			default:
				progressReportField.setText("Every colored button is a toggle. Try them");
				break;
			case 1:
				progressReportField.setText("Max depth cueing brightness = 255; min=0");
				break;
			case 2:
				progressReportField.setText("Max transparency = 255; min=0->complete opacity");
				break;
			case 3:
				progressReportField.setText(" rayTrace sects back2front if startSect>stopSect");
				break;
			}			
		}
	 
	public void setCrossEyedStereo(boolean ifCrossEm) {
		ifCrossEyes = ifCrossEm;
		adjustButtons();
		}
				
// catch button and menu events...
	public void actionPerformed(java.awt.event.ActionEvent event) {
		String arg = event.getActionCommand();
		int segment;
	progressReport(null);		
	if(arg.equals("cancel movie")) {
		movieInProgress=false;
		talk("Movie cancelled!");
		talk("Movie making will cease after current frame is finished");
		return;
		}
		
	if(movieInProgress) {
		talk("While movie making is in progress, only the 'cancel movie' button works");
		progressReport("making movie. Only 'cancel movie' button works");
		return;
		}

	if(arg.equals("(toggle) no segmentation in 3D recons")) {
		Stax.hub.segmentationJPanel.PutDataThroughFilterChain(true);
		Stax.hub.segmentationJPanel.setVisible(true);
		return;
		}
		
	if(arg.equals("(toggle) 3D recons from segmentated data")) {
		Stax.hub.segmentationJPanel.PutDataThroughFilterChain(false);
		return;
		}
		
	if(arg.equals("read parameters from disk")) {
		(new ArchiveMarks()).readWriteProjectionsPanelData(false, this);
		return;
		}
		
	if(arg.equals("WRITE PARAMETERS TO DISK")) {
		(new ArchiveMarks()).readWriteProjectionsPanelData(true, this);
		return;
		}
		
	if(arg.equals("next seg")) {
		goToNextSegment();
		return;
		}
		
	if(arg.equals("save reverse order movie frames")) {
		ifSaveReverseOrderFrames = false;
		ifSaveReverseOrderFramesButton.setText("Omit reverse-order movie frames");
		return;
		}
		
	if(arg.equals("Omit reverse-order movie frames")) {
		ifSaveReverseOrderFrames = true;
		ifSaveReverseOrderFramesButton.setText("save reverse order movie frames");
		return;
		}
		
	if(arg.equals("blend max projection with ray-traced projection")) {
		whichBlend=1;
		adjustButtons();
		return;
		}
		
	if(arg.equals("do maximum projection only")) {
		whichBlend=2;
		adjustButtons();
		return;
		}
		
	if(arg.equals("do ray-traced projection only")) {
		whichBlend=0;
		adjustButtons();
		return;
		}
		
	if(arg.equals("cross eyes")) {
		ifCrossEyes=false;
		adjustButtons();
		return;
		}
			
	if(arg.equals("diverge eyes")) {
		ifCrossEyes=true;
		adjustButtons();
		return;
		}

	if(arg.equals("include red channel")) {
		ifdored=false;
		adjustButtons();
		return;
		}
	if(arg.equals("omit red channel")) {
		ifdored=true;
		adjustButtons();
		return;
		}
		
	if(arg.equals("include green channel")) {
		ifdogreen=false;
		adjustButtons();
		return;
		}
	if(arg.equals("omit green channel")) {
		ifdogreen=true;
		adjustButtons();
		return;
		}
		
	if(arg.equals("include blue channel")) {
		ifdoblue=false;
		adjustButtons();
		return;
		}
	if(arg.equals("omit blue channel")) {
		ifdoblue=true;
		adjustButtons();
		return;
		}
	
	if(arg.equals("movie rotates stereo pairs")) {	
		ifMovieIsStereo = false;
		stopButton.setText("movie rotates single view");
		return;
		}

	if(arg.equals("movie rotates single view")) {	
		ifMovieIsStereo = true;
		stopButton.setText("movie rotates stereo pairs");
		return;
		}
		
	if(arg.equals("compute overlay") || arg.equals("make single stereo pair") || arg.equals("makeStereoPairMovie")) {	
		makeMovie(arg);
		return;
		}
		    
	for(segment=0;segment<4;segment++) {
		if(arg.equals(doSegmentButtonsLabels[segment])) {
			ifDoSegments[segment] = false;
			adjustButtons();
			return;
			}
		if(arg.equals(omitSegmentButtonsLabels[segment])) {
			ifDoSegments[segment] = true;
			adjustButtons();
			return;
			}
		}

   	}
	
	public double [] getBoosts() {
		for(int chan=0;chan<3;chan++)
			for(int segment=0;segment<4;segment++) 
					ChannelBoosts[chan][segment] = Double.valueOf(ChannelBoostFields[chan][segment].getText()).doubleValue();
		redChannelBoost = ChannelBoosts[0][currentSegment];
		greenChannelBoost = ChannelBoosts[1][currentSegment];
		blueChannelBoost = ChannelBoosts[2][currentSegment];		
		setBoosts();
		getScaleAndBlendFrac(); 	    
		double [] retn = new double[5];
		retn[0]=redChannelBoost;
		retn[1]=greenChannelBoost;
		retn[2]=blueChannelBoost;
		retn[3]=blendFractionOfMaxProjection;
		retn[4]=movieFrameScaleFactor;
		return retn;
		}
			
	private void setBoosts() {
		for(int chan=0;chan<3;chan++)
			for(int segment=0;segment<4;segment++) 
				ChannelBoostFields[chan][segment].setText(dbleSixString(ChannelBoosts[chan][segment]));
		}		

	public void setParameters(double scaleFactor, double blendFrac, double redboost, double greenboost, double blueboost) {
		getBoosts();
		ChannelBoosts[0][currentSegment] = redboost;
		ChannelBoosts[1][currentSegment] = greenboost;
		ChannelBoosts[2][currentSegment] = blueboost;
		setBoosts();
		movieFrameScaleFactor=scaleFactor;
		movieFrameScaleFactorField.setText(dbleSixString(movieFrameScaleFactor));
		blendFractionOfMaxProjection = blendFrac;
		blendFractionField.setText(dbleSixString(blendFractionOfMaxProjection));	
		}
			
	private void adjustButtons() {
		if(picJPanel.curStack.redfile==null) { // there is no red channel
			ifdored=false;
			doRedButton.setText("    ");
			}
		else {
			if(ifdored) 
				doRedButton.setText("include red channel");
			else
				doRedButton.setText("omit red channel");
			}
			
		if(picJPanel.curStack.greenfile==null) { // there is no green channel
			ifdogreen=false;
			doGreenButton.setText("    ");
			}
		else {
			if(ifdogreen) 
				doGreenButton.setText("include green channel");
			else
				doGreenButton.setText("omit green channel");
			}
			
		if(picJPanel.curStack.bluefile==null) { // there is no blue channel
			ifdoblue=false;
			doBlueButton.setText("    ");
			}
		else {
			if(ifdoblue) 
				doBlueButton.setText("include blue channel");
			else
				doBlueButton.setText("omit blue channel");
			}
			
		if(ifCrossEyes)
			crossEyeButton.setText("cross eyes");
		else 
			crossEyeButton.setText("diverge eyes");
			
		switch (whichBlend) {
			case 0:
			default :
				whichBlend=0;
				ifdoraytrace=true;
				ifdomaxproj=true;
				blenderButton.setText("blend max projection with ray-traced projection");
				break;
			case 1:
				ifdoraytrace=false;
				ifdomaxproj=true;
				blenderButton.setText("do maximum projection only");
				break;
			case 2:
				ifdoraytrace=true;
				ifdomaxproj=false;
				blenderButton.setText("do ray-traced projection only");
				break;
			}
			
		for(int segment=0;segment<4;segment++) {
			if(ifDoSegments[segment]) 
				doSegmentsButtons[segment].setText(doSegmentButtonsLabels[segment]);	
			
			else	 		
				doSegmentsButtons[segment].setText(omitSegmentButtonsLabels[segment]);	
				
			if(currentSegment==segment)
				doSegmentsButtons[segment].setBackground(segmentIsCurrentColor);
			else
				doSegmentsButtons[segment].setBackground(brownback);
			}

		}

	private static final String dbleSixString(double anum) {
		String stg = ""+anum;
		if(stg.length() < 6)
			return stg;
		else
			return stg.substring(0,6);
		}
				
	private void makeTransparencyLUTS() {
		getBoosts(); 
		redSatur8Transp = (int)(Double.valueOf(redTranspFIeld.getText()).doubleValue());
		greenSatur8Transp = (int)(Double.valueOf(greenTranspFIeld.getText()).doubleValue());
		blueSatur8Transp = (int)(Double.valueOf(blueTranspFIeld.getText()).doubleValue());
		if(redSatur8Transp<0) 
			redSatur8Transp=0;
		if(redSatur8Transp>254)
			redSatur8Transp=254;
		redTranspFIeld.setText(""+redSatur8Transp);
		if(greenSatur8Transp<0) 
			greenSatur8Transp=0;
		if(greenSatur8Transp>254)
			greenSatur8Transp=254;
		greenTranspFIeld.setText(""+greenSatur8Transp);
		if(blueSatur8Transp<0) 
			blueSatur8Transp=0;
		if(blueSatur8Transp>254)
			blueSatur8Transp=254;
		blueTranspFIeld.setText(""+blueSatur8Transp);
		for(int i=0; i<256; i++) {
			redtransp[i] = ((255-i)*255 + i*redSatur8Transp)/255;
			greentransp[i] = ((255-i)*255 + i*greenSatur8Transp)/255;
			bluetransp[i] = ((255-i)*255 + i*blueSatur8Transp)/255;
			}
		}
	
	public void run() {
		// talk("Start run thread to make movie frames. framehi = "+framehi);
		boolean ifMakePair = true;
		double blendFrac;
		int currentSegmentHold=currentSegment;
		if(framehi>0 && !ifMovieIsStereo)
			ifMakePair=false;
		int stereoSep = ifMakePair? stereoSeparation : 0;
		int segment,frameFileNumberBase,tellframe=-1;
		if(!ifCrossEyes) 
			stereoSep = -stereoSep;

		getMarginsOfStereoPairs(sixDegreesFromKevinBacon*stereoSep); 
		if(ifdoraytrace)
			rayTraceRaster = new int[stereoWidth*curHeight];
		if(ifdomaxproj)
			maxProjectionRaster = new int[stereoWidth*curHeight];
		makeTransparencyLUTS();
		picJPanel.curStack.initializeByteChanStack();	
		spfPrevious = null;
		while(movieInProgress && frameNumber<=framehi) {
			if(framehi>0)
				tellframe=frameNumber;
			picJPanel.curStack.makeStereoPairOverlay(rayTraceRaster, ifdoraytrace, maxProjectionRaster, ifdomaxproj, 
				stereoWidth, leftMargins[frameNumber], rightMargins[frameNumber], from, to, depthCueBrightnessOfFromSection, depthCueBrightnessOfToSection, 
					ifdored, ifdogreen, ifdoblue, redtransp, greentransp, bluetransp,  redNonMaskFrac,greenNonMaskFrac, blueNonMaskFrac, sixDegreesFromKevinBacon*stereoSep,  
						sixDegreesFromKevinBacon*angleBetweenMovieFrames*(frameOffset+frameNumber),ifMakePair, tellframe, progressReportField);
			frameTitle = ("stereoPair from sect "+from+" to "+to+" with "+stereoSep+" degree eye separation");
			if(framehi>0)
				frameTitle = frameTitle+" (movieFrame "+frameNumber+")";
			blendFrac = blendFractionOfMaxProjection;
			if(!ifdomaxproj)
				blendFrac=0;
			if(!ifdoraytrace)
				blendFrac=1;
			spf = new StereoPairFrame(this, frameTitle, rayTraceRaster, maxProjectionRaster, blendFrac, stereoWidth, curHeight, scaleFactor, 
				stereoSep, ifMakePair, redChannelBoost, greenChannelBoost, blueChannelBoost);

			if(framehi>0 && frameNumber == 0) {
			  	movieFileBaseName = spf.scaledPic2jpg(null);
			  	// delete file called (movieFileBaseName)
			  	movieFileBaseName = removeEnd(movieFileBaseName, ".jpg");
			  	movieFileBaseName = removeEnd(movieFileBaseName, ".jpeg");
			  	}
			if(spfPrevious != null) 
				spfPrevious.stopPainting();
			if(framehi>0) {
				spf.stopResponding();
				for(currentSegment=0;currentSegment<4;currentSegment++) 
					if(ifDoSegments[currentSegment])  {
						frameFileNumberBase = ifSaveReverseOrderFrames? 2*(framehi+1)*currentSegment : (framehi+1)*currentSegment;
						getBoosts();
						spf.takeParamsFromParent();
						spf.scaledPic2jpg((movieFileBaseName+fourDigitNum(frameFileNumberBase+frameNumber)+".jpg"));
						if(ifSaveReverseOrderFrames) 
							spf.scaledPic2jpg((movieFileBaseName+fourDigitNum(frameFileNumberBase + 2*framehi+1 - frameNumber)+".jpg"));
						}
				try{
					Thread.sleep(200);
					if(spfPrevious != null) 
			  		spfPrevious.sepuku();
					}
				catch(InterruptedException e) {
					talk("run thread got exception: "+e);
					}
				}
			spfPrevious=spf;	
			++frameNumber;
			}
		if(ifMovieIsStereo)
			stopButton.setText("movie rotates stereo pairs");
		else
			stopButton.setText("movie rotates single view");			
		movieInProgress=false;
		picJPanel.curStack.disposeOfByteChanStack();	

	    rayTraceRaster=null;
	   	currentSegment = currentSegmentHold;
		spf.startResponding();
		}
		
	private static final int inlimits(int low,int hi,int i) {
		if(i<low)
			i=low;
		if(i>hi)
			i=hi;
		return i;
		}
		
	private void getMarginsOfStereoPairs(double colShiftPerSection) {
		int tces; 
	    if(colShiftPerSection >= 0) 
	    	tces = from;
	    else {
	    	tces = to;
	    	colShiftPerSection = -colShiftPerSection;
	    	}
	    int ifrinc = (to >= from)? 1 : -1;
		int mvyFrmNumber,leftcolmin,leftcolmax,rightcolmin,rightcolmax,acol,col,ifr,leftMargin,rightMargin,lwidth=0,rwidth=0;
		double swingOffset;
		leftMargins = new int[1+framehi];
		rightMargins = new int[1+framehi];
		int [] leftLeft = new int[1+framehi],leftRight = new int[1+framehi],rightLeft = new int[1+framehi],rightRight = new int[1+framehi];
		stereoWidth = 2*curWidth;
		for(mvyFrmNumber=0; mvyFrmNumber <= framehi; ++mvyFrmNumber) {
			swingOffset= sixDegreesFromKevinBacon*angleBetweenMovieFrames*(frameOffset+mvyFrmNumber);
			leftcolmin = 10000;
			leftcolmax = -10000;
			rightcolmin = 10000;
			rightcolmax = -10000;
			ifr=from;	
	   	 	while(ifr != (to+ifrinc)) {
	    		for(col=0;col<=curWidth;col+=curWidth) {
					acol = col + (int)(.49 + (swingOffset + colShiftPerSection)*(ifr - tces));
					if(acol<leftcolmin)
						leftcolmin=acol;
					if(acol>leftcolmax)
						leftcolmax=acol;
					acol = col + (int)(.49 +  (swingOffset - colShiftPerSection)*(ifr - tces));
					if(acol<rightcolmin)
						rightcolmin=acol;
					if(acol>rightcolmax)
						rightcolmax=acol;
	    			}
	    		ifr += ifrinc;
	    		}
	    	leftLeft[mvyFrmNumber] = leftcolmin;
	    	leftRight[mvyFrmNumber] = leftcolmax;
	    	rightLeft[mvyFrmNumber] = rightcolmin;
	    	rightRight[mvyFrmNumber] = rightcolmax;
	    	if(lwidth < leftcolmax - leftcolmin)
	    		lwidth = leftcolmax - leftcolmin;
	    	if(rwidth < rightcolmax - rightcolmin)
	    		rwidth = rightcolmax - rightcolmin;
	    	}
	    	stereoWidth = lwidth+rwidth;
	    	talk("lwidth="+lwidth+" and rwidth="+rwidth+" and so stereoWidth="+stereoWidth);
	    	
	    	/* center each view in its panel. For each mvyFrmNumber, want
	    	leftMargins[mvyFrmNumber] + (leftRight[mvyFrmNumber] + leftLeft[mvyFrmNumber])/2 == lwidth/2
	    	rightMargins[mvyFrmNumber] + (rightRight[mvyFrmNumber] + rightLeft[mvyFrmNumber])/2 == rwidth/2
	    	  but we also need leftMargins[mvyFrmNumber] + leftLeft[mvyFrmNumber] >= 0
	    	  and we need rightMargins[mvyFrmNumber] + rightLeft[mvyFrmNumber] >= 0
	    	       */
		for(mvyFrmNumber=0; mvyFrmNumber <= framehi; ++mvyFrmNumber) {
	    	leftMargins[mvyFrmNumber] = (lwidth - (leftRight[mvyFrmNumber] + leftLeft[mvyFrmNumber]))/2;
	    	if(leftMargins[mvyFrmNumber] + leftLeft[mvyFrmNumber] < 0) {
	    		talk("For frame number "+mvyFrmNumber+", had to adjust leftMargins from "+leftMargins[mvyFrmNumber]);
	    		leftMargins[mvyFrmNumber] = -leftLeft[mvyFrmNumber];
	    		talk("                                                      to "+leftMargins[mvyFrmNumber]);
	    		} 

	    	rightMargins[mvyFrmNumber] = (rwidth - (rightRight[mvyFrmNumber] + rightLeft[mvyFrmNumber]))/2;
	    	if(rightMargins[mvyFrmNumber] + rightLeft[mvyFrmNumber] < 0) {
	    		talk("For frame number "+mvyFrmNumber+", had to adjust rightMargins from "+rightMargins[mvyFrmNumber]);
	    		rightMargins[mvyFrmNumber] = -rightLeft[mvyFrmNumber];
	    		talk("                                                      to "+rightMargins[mvyFrmNumber]);
	    		} 
	    	rightMargins[mvyFrmNumber] += lwidth;
	    	}
	    // now find the least gap between the leftView's right edge and the rightView's left edge
	    int leastgap = stereoWidth;
		for(mvyFrmNumber=0; mvyFrmNumber <= framehi; ++mvyFrmNumber) 
			if(leastgap >  (rightMargins[mvyFrmNumber] + rightLeft[mvyFrmNumber]) - (leftMargins[mvyFrmNumber] + leftRight[mvyFrmNumber]) )
				leastgap =  (rightMargins[mvyFrmNumber] + rightLeft[mvyFrmNumber]) - (leftMargins[mvyFrmNumber] + leftRight[mvyFrmNumber]);
		talk("leastgap = "+leastgap+" when stereoWidth == "+stereoWidth+" and lwidth=="+lwidth+" and rwidth=="+rwidth);
		stereoWidth -= leastgap;
	    talk("Revision: lwidth="+lwidth+" and rwidth="+rwidth+" and so stereoWidth="+stereoWidth);
	    /* recompute the right margins so as to move the rightView leastgap pixels to the left
	    	This time, for each mvyFrmNumber, want
	    	rightMargins[mvyFrmNumber] + (rightRight[mvyFrmNumber] + rightLeft[mvyFrmNumber])/2 == rwidth/2 - leastgap */
		for(mvyFrmNumber=0; mvyFrmNumber <= framehi; ++mvyFrmNumber) 
	    	rightMargins[mvyFrmNumber] -= leastgap;
		}
	
	private void getScaleAndBlendFrac() {
	    if(blendFractionField != null)
	    	blendFractionOfMaxProjection = (Double.valueOf(blendFractionField.getText()).doubleValue());	
	    if(movieFrameScaleFactorField != null)
	    	movieFrameScaleFactor = Double.valueOf(movieFrameScaleFactorField.getText()).doubleValue();
		}	
			
	private void makeMovie(String arg) {	
		boolean ifMakingStereoPair = arg.equals("make single stereo pair") ||  arg.equals("makeStereoPairMovie");
		if(ifMakingStereoPair && movieInProgress) {
			talk("Stereo movie already in progress. Use 'cancel movie' button to stop it");
			return;
			}
		movieInProgress = arg.equals("makeStereoPairMovie");	
	    // fetch the new from and to limits...
	    from = (int)(Double.valueOf(fromSection.getText()).doubleValue());	
	    to = (int)(Double.valueOf(toSection.getText()).doubleValue());
	    getScaleAndBlendFrac();
	    if(blendFractionOfMaxProjection<=0) {
	    	blendFractionOfMaxProjection=0;
	    	blendFractionField.setText("0");
	    	whichBlend=2;
	    	adjustButtons();
	    	}	    	
	    if(blendFractionOfMaxProjection>=1) {
	    	blendFractionOfMaxProjection=1;
	    	blendFractionField.setText("1");
	    	whichBlend=1;
	    	adjustButtons();
	    	}	    	
		getOtherParameters();
	    fromSection.setText((""+from));	
	    toSection.setText((""+to));	
	    if(ifMakingStereoPair)
	    	talk("making stereo pair 3D reconstruction from sections "+from+" --> "+to);
	    else
	    	talk("computing overlay of sections "+from+" to "+to);
	    // make sure the raster is the right size
	    if(!ifMakingStereoPair) {
	    	if(!overlayIsValid || overlayRas==null || curWidth != picJPanel.curWidth || 
		    	curHeight != picJPanel.curHeight) {
				curWidth = picJPanel.curWidth;
				curHeight = picJPanel.curHeight;
				overlayRas = new int[curWidth*curHeight];
				talk("width="+curWidth+",  curHeight="+curHeight);
				}
	    	picJPanel.curStack.overlaySections(overlayRas, from, to, whichBlend, redNonMaskFrac,greenNonMaskFrac, blueNonMaskFrac, progressReportField);
	    	oldFrom=from; oldTo=to; overlayIsValid=true;
			new StereoPairFrame(this, ("Overlay (sects "+from+" to "+to+")"), overlayRas, null,0.,
				curWidth, curHeight, scaleFactor, 0,true, redChannelBoost, greenChannelBoost, blueChannelBoost);
			return;
	    	}

	    if(ifMakingStereoPair) {
			curWidth = picJPanel.curWidth;
			curHeight = picJPanel.curHeight;
	    	stereoSeparation = Math.abs(stereoSeparation);
			stereoSeparationField.setText(""+stereoSeparation);
			spfPrevious=null;
			stereoSeparationPerSection = sixDegreesFromKevinBacon*stereoSeparation;

		
	    	frameNumber=0;
			talk("stereo pair width = "+stereoWidth+",  stereo pair height="+curHeight);
			if(movieInProgress) { // set up for many-frame movie
				frameOffset = -nmovieframes/2; //sweeps from -nmovieframes/2  to nmovieframes/2
				framehi=nmovieframes;
				stopButton.setText("cancel movie");
				scaleFactor = movieFrameScaleFactor;
				}
			else {
				frameOffset = 0; //no sweep at all
				framehi = 0;
				}
			movieInProgress=true;
			movieThread = new Thread(this);
			movieThread.start();
			}
	    }
		
	private static String fourDigitNum(int i) {
		if(i<10)
			return ("000"+i);
		else if(i<100)	
			return ("00"+i);
		else if(i<1000)	
			return ("0"+i);
		else 
			return (""+i);	
		}
		
	private static String removeEnd(String stg, String tag) {
		if(stg==null || tag == null)
			return stg;
		int i;
		while((i = stg.indexOf(tag)) > 0)
			stg = stg.substring(0,i);
		return stg;
		}
		
   public Stereo3DReconstruction(ShowPicJPanel picJPanel)   {   
        super("Stereo3DReconstructions");
	this.picJPanel = picJPanel;
	getContentPane().setLayout(new GridLayout(0,1));    // NEED to specify a layout manager!!
	WidgetMaker wmk = new WidgetMaker(this);
	Color redback = new Color(255,100,100);
	Color greenback = new Color(100,255,100);
	Color blueback = new Color(100,100,255);
	Color yellowback = new Color(225,255,120);
	JPanel [] chanpans = new JPanel[3];
	ChannelBoostFields = new JTextField[3][4];
	ChannelBoosts = new double[3][4];
	int chan,segment;
	
	JPanel chanButtonsPan = new JPanel();
	chanButtonsPan.setLayout(new GridLayout(1,5));
	chanButtonsPan.add(wmk.makeAJButton("next seg"));
	for(segment=0;segment<4;segment++) {
		doSegmentsButtons[segment] = wmk.makeAJButton(omitSegmentButtonsLabels[segment],brownback);
		chanButtonsPan.add(doSegmentsButtons[segment]);
		}

	for(segment=0;segment<4;segment++) {
		for(chan=0; chan<3; chan++) 
			ChannelBoosts[chan][segment]=1.0;			
		ChannelBoostFields[0][segment] = wmk.makeAJTextField(dbleSixString(1.0),6, redback);
		ChannelBoostFields[1][segment] = wmk.makeAJTextField(dbleSixString(1.0),6, greenback);
		ChannelBoostFields[2][segment] = wmk.makeAJTextField(dbleSixString(1.0),6, blueback);
		}
	
	for(chan=0; chan<3; chan++)  {
		chanpans[chan]=new JPanel();
		chanpans[chan].setLayout(new GridLayout(1,5));
		switch (chan) {
			case 0:
				chanpans[chan].add(wmk.makeAJLabel("redBoost"));
				break;
			case 1:
				chanpans[chan].add(wmk.makeAJLabel("greenBoost"));
				break;
			case 2:
				chanpans[chan].add(wmk.makeAJLabel("blueBoost"));
				break;
			}
		for(segment=0;segment<4;segment++) 
			chanpans[chan].add(ChannelBoostFields[chan][segment]);
		currentSegment=0;
		getBoosts();
		}
		
	JPanel panA = new JPanel();
	panA.setLayout(new GridLayout(1,4));
	panA.add(wmk.makeAJLabel("     "));
	panA.add(wmk.makeAJLabel("     "));
	panA.add(wmk.makeAJLabel("start section="));
	fromSection = wmk.makeAJTextField("10", 4);
	panA.add(fromSection);

	JPanel panB = new JPanel();
	panB.setLayout(new GridLayout(1,4)); 
	panB.add(wmk.makeAJLabel("      "));
	panB.add(wmk.makeAJLabel("      "));
	panB.add(wmk.makeAJLabel("stop section="));
	toSection = wmk.makeAJTextField("12", 4);
	panB.add(toSection);
	
	JPanel panBB = new JPanel();
	panBB.setLayout(new GridLayout(1,4)); 
	blendFractionField = wmk.makeAJTextField((""+blendFractionOfMaxProjection),8);
	panBB.add(wmk.makeAJLabel("blendFraction="));
	panBB.add(blendFractionField);
	panBB.add(wmk.makeAJLabel("for maximum"));
	panBB.add(wmk.makeAJLabel("projection"));

	/*JPanel panlab  = new JPanel();
	panlab.setLayout(new GridLayout(1,2)); 
	panlab.add(wmk.makeAJLabel("  channel transparencies"));
	panlab.add(wmk.makeAJLabel("  channel boost Factors")); */
	
	JPanel panTR = new JPanel();
	panTR.setLayout(new GridLayout(1,2)); 
	panTR.add(wmk.makeAJLabel("red channel transparency="));
	redTranspFIeld = wmk.makeAJTextField((""+redSatur8Transp),4,redback);
	panTR.add(redTranspFIeld);

	JPanel panTG = new JPanel();
	panTG.setLayout(new GridLayout(1,2)); 
	panTG.add(wmk.makeAJLabel("green channel transparency="));
	greenTranspFIeld = wmk.makeAJTextField((""+greenSatur8Transp),4,greenback);
	panTG.add(greenTranspFIeld);

	JPanel panTB = new JPanel();
	panTB.setLayout(new GridLayout(1,2)); 
	panTB.add(wmk.makeAJLabel("blue channel transparency="));
	blueTranspFIeld = wmk.makeAJTextField((""+blueSatur8Transp),4,blueback);
	panTB.add(blueTranspFIeld);
	
	JPanel  segFltrTogPan = new JPanel();
	segFltrTogPan.setLayout(new GridLayout(1,1)); 
	useSegmentationFiltersToggleButton = wmk.makeAJButton("(toggle) no segmentation in 3D recons");
	segFltrTogPan.add(useSegmentationFiltersToggleButton);

	// start new stuff added February 2008
	JPanel panMaskFracR = new JPanel();
	panMaskFracR.setLayout(new GridLayout(1,2)); 
	panMaskFracR.add(wmk.makeAJLabel("red nonMaskFrac"));
	maskFracRField = wmk.makeAJTextField((""+redNonMaskFrac),4,redback);
	panMaskFracR.add(maskFracRField);

	JPanel panMaskFracG = new JPanel();
	panMaskFracG.setLayout(new GridLayout(1,2)); 
	panMaskFracG.add(wmk.makeAJLabel("green nonMaskFrac"));
	maskFracGField = wmk.makeAJTextField((""+greenNonMaskFrac),4,redback);
	panMaskFracG.add(maskFracGField);

	JPanel panMaskFracB = new JPanel();
	panMaskFracB.setLayout(new GridLayout(1,2)); 
	panMaskFracB.add(wmk.makeAJLabel("blue nonMaskFrac"));
	maskFracBField = wmk.makeAJTextField((""+blueNonMaskFrac),4,redback);
	panMaskFracB.add(maskFracBField);

	// end new stuff added February 2008
	
	JPanel panDCF = new JPanel();
	panDCF.setLayout(new GridLayout(1,4)); 
	depthCueFarField = wmk.makeAJTextField(""+depthCueBrightnessOfToSection,4);
	panDCF.add(depthCueFarField);
	panDCF.add(wmk.makeAJLabel("depth cueing"));
	panDCF.add(wmk.makeAJLabel("brightness of"));
	panDCF.add(wmk.makeAJLabel("stop section"));

	JPanel panDCC = new JPanel();
	panDCC.setLayout(new GridLayout(1,4)); 
	depthCueCloseField = wmk.makeAJTextField(""+depthCueBrightnessOfFromSection,4);
	panDCC.add(depthCueCloseField);
	panDCC.add(wmk.makeAJLabel("depth cueing"));
	panDCC.add(wmk.makeAJLabel("brightness of"));
	panDCC.add(wmk.makeAJLabel("start section"));
	
	JPanel panCCC = new JPanel();
	panCCC.setLayout(new GridLayout(1,1)); 
	blenderButton = wmk.makeAJButton("                                                      ",brownback);
	panCCC.add(blenderButton);

	JPanel panC = new JPanel();
	panC.setLayout(new GridLayout(1,2)); 
	crossEyeButton = wmk.makeAJButton("       ",brownback);
	panC.add(crossEyeButton);
	doRedButton = wmk.makeAJButton("                     ",redback);
	panC.add(doRedButton);
	
	JPanel panD = new JPanel();
	panD.setLayout(new GridLayout(1,2)); 
	panD.add(wmk.makeAJButton("compute overlay"));
	doGreenButton = wmk.makeAJButton("                     ",greenback);
	panD.add(doGreenButton);

	JPanel panCC = new JPanel();
	panCC.setLayout(new GridLayout(1,2)); 
	panCC.add(wmk.makeAJButton("make single stereo pair"));
	doBlueButton = wmk.makeAJButton("                     ",blueback);
	panCC.add(doBlueButton);

	
	JPanel panF = new JPanel();
	panF.setLayout(new GridLayout(1,2)); 
	stereoSeparationField = wmk.makeAJTextField(""+stereoSeparation, 8);
	panF.add(stereoSeparationField);
	panF.add(wmk.makeAJLabel("= Eye Separation Angle"));
	
	JPanel panFKB = new JPanel();
	panFKB.setLayout(new GridLayout(1,2)); 
	kevinBaconField = wmk.makeAJTextField(""+sixDegreesFromKevinBacon, 8);
	panFKB.add(kevinBaconField);
	panFKB.add(wmk.makeAJLabel("= DepthParam (larger seems deeper)"));
	
	JPanel panG = new JPanel();
	panG.setLayout(new GridLayout(1,2)); 
	numMovieFramesField = wmk.makeAJTextField(""+nmovieframes, 4);
	panG.add(numMovieFramesField);
	panG.add(wmk.makeAJLabel("= # movie frames"));
	
	JPanel panH = new JPanel();
	panH.setLayout(new GridLayout(1,2)); 
	angleBetweenMovieFramesField = wmk.makeAJTextField(""+angleBetweenMovieFrames, 4);
	panH.add(angleBetweenMovieFramesField);
	panH.add(wmk.makeAJLabel("=angleBetweenMovieFrames"));
	
	JPanel panI = new JPanel();
	panI.setLayout(new GridLayout(1,2)); 
	movieFrameScaleFactorField = wmk.makeAJTextField(""+movieFrameScaleFactor, 4);
	panI.add(movieFrameScaleFactorField);
	panI.add(wmk.makeAJLabel("=MovieFrameScaleFactor"));
	
	JPanel panJ = new JPanel();
	panJ.setLayout(new GridLayout(1,2)); 
	panJ.add(wmk.makeAJButton("makeStereoPairMovie"));
	stopButton=wmk.makeAJButton("movie rotates stereo pairs",brownback);
	panJ.add(stopButton);

	JPanel panZ = new JPanel();
	panZ.setLayout(new GridLayout(1,1)); 
	ifSaveReverseOrderFramesButton=wmk.makeAJButton("save reverse order movie frames",brownback);
	panZ.add(ifSaveReverseOrderFramesButton);

	JPanel panIO  = new JPanel();
	panIO.setLayout(new GridLayout(1,2)); 
	panIO.add(wmk.makeAJButton("read parameters from disk"));
	panIO.add(wmk.makeAJButton("WRITE PARAMETERS TO DISK"));

	getContentPane().add(panA); 
	getContentPane().add(panDCC); 
	getContentPane().add(panB); 
	getContentPane().add(panDCF); 	
	getContentPane().add(panBB); 
	
	getContentPane().add(panMaskFracR); 
	getContentPane().add(panMaskFracG); 
	getContentPane().add(panMaskFracB); 
	getContentPane().add(segFltrTogPan); 

	// getContentPane().add(panlab); 
	getContentPane().add(panTR); 
	getContentPane().add(panTG); 
	getContentPane().add(panTB); 
	getContentPane().add(chanButtonsPan);	
	for(chan=0; chan<3; chan++)  
		getContentPane().add(chanpans[chan]);
	getContentPane().add(panZ); 
	getContentPane().add(panCCC); 
	getContentPane().add(panC); 
	getContentPane().add(panD); 
	getContentPane().add(panCC); 
	getContentPane().add(panJ); 
	getContentPane().add(panF); 
	getContentPane().add(panFKB); 
	getContentPane().add(panG); 
	getContentPane().add(panH); 
	getContentPane().add(panI); 
	progressReportField = wmk.makeAJTextField("                           ",30,yellowback);
	getContentPane().add(progressReportField); 
	getContentPane().add(panIO); 
    setSize(nomwidth, nomheight);
    addWindowListener(this);
    adjustButtons();
	progressReport(null);
	}


// catch WindowListener events  
	public void windowActivated(java.awt.event.WindowEvent event) {
		// talk("window activated");
		}
	public void windowDeactivated(java.awt.event.WindowEvent event) {
		//  talk("window Deactivated");
		}
	public void windowOpened(java.awt.event.WindowEvent event) {}
	public void windowIconified(java.awt.event.WindowEvent event) {}
	public void windowDeiconified(java.awt.event.WindowEvent event) {}
	public void windowClosing(java.awt.event.WindowEvent event) {
	    setVisible(false);
		}
	public void windowClosed(java.awt.event.WindowEvent event) {
		}
	 
	public void changeJPanelSize(double ratio) {
		int pwidth = getSize().width;
		int pheight = getSize().height;
		pwidth = (int)(ratio*pwidth +.4999);
		pheight = (int)(ratio*pheight +.4999);
		setSize(pwidth,pheight);
		show();
		}

	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}
