package staxx;

/* Info ...  author, copyright, bundle-type, etc */
/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
import java.lang.*;
import java.util.*;
import java.awt.*;  
import javax.swing.*;

class StereoPairFrame extends JFrame implements java.awt.event.ActionListener,java.awt.event.WindowListener{
	StereoPairCanvas canvas=null;
	JTextField scaleField,blendField;
	Stereo3DReconstruction parent=null;
	boolean ifStereoPair;
	String title;
	JButton autoRecolorButton;
	ZseriesMainWindow ZSeriesAncestor;
	boolean ifAutoRecoloring = true, dontPaint=false, ifResponsive=true, ifMakePair=true;
	double redChannelBoost,greenChannelBoost,blueChannelBoost;
	
	public StereoPairFrame(Stereo3DReconstruction parent, String title,int [] rayTraceRaster, int [] maxProjectionRaster, double blendFrac,  int originalPicWidth, 
			int originalPicHeight, double scaleFactor, int stereoSeparation, boolean ifMakePair, double redChannelBoost, double greenChannelBoost,
					double blueChannelBoost) {
		this.ifMakePair = ifMakePair;
		ZSeriesAncestor = parent.picJPanel.parent; 
    	this.title = new String(title);
    	this.redChannelBoost = redChannelBoost;
    	this.greenChannelBoost = greenChannelBoost;
    	this.blueChannelBoost = blueChannelBoost;
    	ifStereoPair = (stereoSeparation != 0);
    	this.parent = parent;  // parent maintains sticky scaleFactor
		canvas = new StereoPairCanvas(this, originalPicWidth,  originalPicHeight, rayTraceRaster,
				maxProjectionRaster,  blendFrac,  scaleFactor, stereoSeparation);
    	getContentPane().setLayout(new BorderLayout()); 
    	WidgetMaker wmk = new WidgetMaker(this);
    	Panel jpan = new Panel();
    	jpan.setLayout(new GridLayout(1,0));
    	jpan.add(wmk.makeAJButton("Save scaled jpg"));
    		Panel kpan = new Panel();
    		kpan.setLayout(new GridLayout(1,2));
    		kpan.add(wmk.makeAJButton("recolor/rescale to ->"));
    		scaleField = wmk.makeAJTextField(""+scaleFactor,12);
    		kpan.add(scaleField);
    	jpan.add(kpan);
    		Panel qpan = new Panel();
    		qpan.setLayout(new GridLayout(1,2));
    		qpan.add(wmk.makeAJButton("newBlendFrac4MaxProj->"));
    		blendField = wmk.makeAJTextField(""+blendFrac,6);
    		qpan.add(blendField);
    	jpan.add(qpan);
    	if(ifStereoPair)
    		jpan.add(wmk.makeAJButton("swap eyes"));
    	getContentPane().add("North",jpan);
    	resize();
    	getContentPane().add("Center",canvas);

    	Panel bpanhi = new Panel();
    	bpanhi.setLayout(new GridLayout(1,0));
		
    	Panel bpanlo = new Panel();
    	bpanlo.setLayout(new GridLayout(1,0));
		
    	Panel bpan = new Panel();
    	bpan.setLayout(new GridLayout(2,1));
		
		autoRecolorButton =  wmk.makeAJButton("AutoRecoloring");
		ifAutoRecoloring=true;
    	bpanhi.add(autoRecolorButton);
    	bpanlo.add(wmk.makeAJButton("dispose"));

    	bpanhi.add(wmk.makeAJButton("++redChan"));
    	bpanlo.add(wmk.makeAJButton("--redChan"));
    	
    	bpanhi.add(wmk.makeAJButton("++greenChan"));
    	bpanlo.add(wmk.makeAJButton("--greenChan"));
    	
    	bpanhi.add(wmk.makeAJButton("++blueChan"));
    	bpanlo.add(wmk.makeAJButton("--blueChan"));
    	
    	bpanhi.add(wmk.makeAJButton("params to 'Projections'"));
    	bpanlo.add(wmk.makeAJButton("params from 'Projections'"));
    	bpan.add(bpanhi);
    	bpan.add(bpanlo);
    	getContentPane().add("South",bpan);

		repaint();
		adjustTitle();
		setVisible(true);
    	addWindowListener(this);
		IAmAHotPic(); // so that recoloring automatically updates this pic
		}
	
	public void adjustTitle() {
    	if(ifStereoPair) {
    		if(canvas.ifSwapEyes) 
    			setTitle("diverged-eye "+this.title);				
			else 
    			setTitle("cross-eyed "+this.title);
    			
    		}
    	else 
    		setTitle(this.title);
		}
	
	public void IAmAHotPic() {
		ZSeriesAncestor.setHotPic(this,ifAutoRecoloring);
		}
			
	private void resize() {
   		setSize(canvas.scaledPicWidth+1,canvas.scaledPicHeight+66);
   		setSize(canvas.scaledPicWidth,canvas.scaledPicHeight+65);
		}
	
	public void stopResponding() {
		ifResponsive=false;
		}
		
	public void startResponding() {
		ifResponsive=true;
		}
		
	public void stopPainting() {
		dontPaint=true;
		canvas.dontPaint=true;
		}
					
	public void sepuku() {
		stopPainting();
		ZSeriesAncestor.setHotPic(this,false);
		canvas.sepuku();
		canvas=null;
		dispose();
		ZSeriesAncestor=null;
		parent=null;
		}

	public void paint(Graphics g) {
		if(dontPaint) return;
		canvas.paint(g);
		}
	
	public void recolorRescale() {
		canvas.blendFrac = Double.valueOf(blendField.getText()).doubleValue();
		if(canvas.blendFrac < 0.)
			canvas.blendFrac=0;
		if(canvas.blendFrac > 1)
			canvas.blendFrac=1;
		blendField.setText(""+canvas.blendFrac);
		parent.blendFractionField.setText(""+canvas.blendFrac);
		canvas.scaleFactor = Double.valueOf(scaleField.getText()).doubleValue();
		canvas.forceRescaleImage();
		resize();
		}
		
	public void takeParamsFromParent() {
		double [] newboosts = parent.getBoosts();
		redChannelBoost = newboosts[0];
		greenChannelBoost = newboosts[1];
		blueChannelBoost = newboosts[2];
		canvas.blendFrac = newboosts[3];
		blendField.setText(""+canvas.blendFrac);
		canvas.scaleFactor = newboosts[4]; 
		scaleField.setText(""+canvas.scaleFactor);
		canvas.forceRescaleImage();
		resize();
		}
					
	public String scaledPic2jpg(String filename) {
		if(ifMakePair)
			return canvas.scaledPic2jpg(filename);
		else
			return canvas.scaledLeftPic2jpg(filename);
		}
							
	public void actionPerformed(java.awt.event.ActionEvent event) {
		if(!ifResponsive) {
			talk("this movie frame does not respond to buttons while movie making is in progress");
			return;
			}
			
		String arg = event.getActionCommand();
		if(arg.equals("dispose")) {
			sepuku();
			return;
			}
					
		if(ifStereoPair && arg.equals("swap eyes")) {
			canvas.ifSwapEyes = !(canvas.ifSwapEyes);
			parent.setCrossEyedStereo(!canvas.ifSwapEyes);
			canvas.flipEyes();
			adjustTitle();
			resize();
			return;
			}

		if(arg.equals("recolor/rescale to ->") || arg.equals("newBlendFrac4MaxProj->")) {
			recolorRescale();
			return;
			}

		if(arg.equals("Save scaled jpg")) {
			scaledPic2jpg(null);
			return;
			}
		
		if(arg.equals("++redChan")) {
			redChannelBoost *= 1.05;
			recolorRescale();
			return;
			}
		
		if(arg.equals("--redChan")) {
			redChannelBoost *= .94;
			recolorRescale();
			return;
			}
		
		if(arg.equals("++greenChan")) {
			greenChannelBoost *= 1.05;
			recolorRescale();
			return;
			}
		
		if(arg.equals("--greenChan")) {
			greenChannelBoost *= .94;
			recolorRescale();
			return;
			}
		
		if(arg.equals("++blueChan")) {
			blueChannelBoost *= 1.05;
			recolorRescale();
			return;
			}
				
		if(arg.equals("--blueChan")) {
			blueChannelBoost *= .94;
			recolorRescale();
			return;
			}

		if(arg.equals("params to 'Projections'")) {
			parent.setParameters(canvas.scaleFactor,canvas.blendFrac,redChannelBoost,greenChannelBoost,blueChannelBoost);
			return;
			}

		if(arg.equals("params from 'Projections'")) {
			takeParamsFromParent();
			return;
			}

		if(arg.equals("ManualRecoloring")) {
			autoRecolorButton.setText("AutoRecoloring");
			ifAutoRecoloring = true;
			IAmAHotPic();
			return;
			}
		
		if(arg.equals("AutoRecoloring")) {
			autoRecolorButton.setText("ManualRecoloring");
			ifAutoRecoloring = false;
			IAmAHotPic();
			return;
			}

		}
// catch WindowListener events  
	public void windowActivated(java.awt.event.WindowEvent event) {
		// talk("window activated");
		}
	public void windowDeactivated(java.awt.event.WindowEvent event) {
		//  talk("window Deactivated");
		}
	public void windowOpened(java.awt.event.WindowEvent event) {}
	public void windowIconified(java.awt.event.WindowEvent event) {}
	public void windowDeiconified(java.awt.event.WindowEvent event) {}
	public void windowClosing(java.awt.event.WindowEvent event) {
	    sepuku();
		}
	public void windowClosed(java.awt.event.WindowEvent event) {
		}
		
	static void talk(String stg) { // write a line on stdout
		System.out.println(stg);
		}
}

