package staxx;

/* Info ...  author, copyright, bundle-type, etc */

/*  <stax program for inspecting and processing stacks of confocal micrographs from Biorad *.pic format files>
    Copyright (C) <2008>  <Garrett Odell>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.lang.*;
import java.util.*;
import java.awt.*;  import javax.swing.*;

class ThickenFltr {
	SegmenterNode nd=null;
	short [] work=null;
	int workrows=-1, workcols=-1;
	
	public ThickenFltr(SegmenterNode nd) {
		this.nd=nd;
		}
	
	private void makeWork(int incols, int inrows) {
		if(work==null || incols != workrows || inrows != workrows) {
			workrows=inrows;
			workcols=incols;
			work = new short[inrows*incols];
			}
		}
		
	public void tellWhatYouDo() {
		nd.talk("");
		nd.talk("This filter thickens a mask by turning on pixels, previously off, which have sufficiently many ");
		nd.talk("neighbors in the on state.");
		nd.talk("The ThickenFltr has a single input raster which we assume is a black/white raaster.");
		nd.talk("It has three parameters, 'fill',  'Nnghb',  and  'iter8'.");
		nd.talk("It performs iter8 successive iterations of the following process.");
		nd.talk("Pixels with values > 127 are transferred unchanged to the output (i.e. 'on' pixels aren't changed).");
		nd.talk("For 'off' pixels < 127, the output pixel remains zero if that pixel has fewer than Nnghb neighbors whose");
		nd.talk("values are >= 127. Else (when there are at least Nnghb neighbors with 'on' values >= 127) the output");
		nd.talk("pixel is set to 255 (i.e. turned on). ");
		}
		


public boolean doit(short [] inras, int incols, int inrows, int fc, int minNeigh, int iter8, short [] outras) {
		int i,row, col, npix = incols*inrows,rcoff,j,neigh;
		int [] offsets = {-1-incols, -incols, -incols+1,
		                     -1,                 1,
		                    incols-1,  incols,  incols+1};
		short on=(short)255, fillColor=(short)fc, enough=(short)127;
		makeWork(incols,inrows);
		if(nd.diags) nd.talk("ThickenFltr with minNeigh="+minNeigh+" and iter8 = "+iter8);
		for(i=0;i<npix;i++)  // store input array in the work array
			work[i] = inras[i];	
		while(iter8>0) {	
		  for(row=1;row<inrows-1;row++)
    		for(col=1;col<incols-1;col++) {
				rcoff = row*incols + col;
				if(work[rcoff]<enough) { // only for off pixels do we count the neigbors
					j=0; neigh=0;
	    			while(j<8) {
						if(work[rcoff+offsets[j]] >=enough) 
							++neigh;
						if(neigh>=minNeigh) j=9;
		    			++j;
		    			}
		    		if(neigh>=minNeigh) outras[rcoff] = fillColor;
					else
	    				outras[rcoff]=work[rcoff]; // just copy non-blank input pixels
		    		}
				else
	    			outras[rcoff]=work[rcoff]; // just copy non-blank input pixels
	    		}
		    /* copy border pixels */
    		for(i=0; i<incols; i++) 
    			outras[i]= inras[i];          // top row
    		for(i=npix-incols; i<npix; i++) 
    			outras[i] =inras[i];         // bottom row
    		for(row=1; row<inrows; row++) {
				rcoff = row*incols;
				outras[rcoff]  =  inras[rcoff];
				outras[rcoff-1] = inras[rcoff-1];
				}
			--iter8;
			if(iter8>0) 
				for(i=0;i<npix;i++)
					work[i]=outras[i];
			}			
	return true;
	}

}